/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock settings
*/
#define OS_SYSTEM_FREQ  (41779200uL)

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ   (OS_SYSTEM_FREQ)
#define OS_TICK_FREQ   (1000u)
#define OS_INT_FREQ    (OS_TICK_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
//
// System, clock and standby control
//
#define POWKEY1                     (*(volatile OS_U32*)0xFFFF0404u)
#define POWCON                      (*(volatile OS_U32*)0xFFFF0408u)
#define POWKEY2                     (*(volatile OS_U32*)0xFFFF040Cu)
#define PLLKEY1                     (*(volatile OS_U32*)0xFFFF0410u)
#define PLLCON                      (*(volatile OS_U32*)0xFFFF0414u)
#define PLLKEY2                     (*(volatile OS_U32*)0xFFFF0418u)
//
// OS timer
//
#define T0LD                        (*(volatile OS_U32*)0xFFFF0300u)
#define T0VAL                       (*(volatile OS_U32*)0xFFFF0304u)
#define T0CON                       (*(volatile OS_U32*)0xFFFF0308u)
#define T0CLRI                      (*(volatile OS_U32*)0xFFFF030Cu)
//
// Interrupt controller
//
#define IRQSTA                      (*(volatile OS_U32*)0xFFFF0000u)
#define IRQSIG                      (*(volatile OS_U32*)0xFFFF0004u)
#define IRQEN                       (*(volatile OS_U32*)0xFFFF0008u)
#define IRQCLR                      (*(volatile OS_U32*)0xFFFF000Cu)
//
// Interrupt sources and peripheral IDs
//
#define SWI_ID_BIT                  ( 1u)
#define TIMER0_ID_BIT               ( 2u)
#define TIMER1_ID_BIT               ( 3u)
#define UART_ID_BIT                 (14u)

#define TIMER0_ID_MASK              (1u << TIMER0_ID_BIT)
#define OS_TIMER_TICK_PENDING_MASK  (1u << TIMER0_ID_BIT)
#define UART_ID_MASK                (1u << UART_ID_BIT)
#define OS_UART_INT_PENDING_MASK    (1u << UART_ID_BIT)

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _InitPLL()
*
*  Function description
*    Initialize PLL to speed up system
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*/
static void _InitPLL(void) {
  PLLKEY1 = 0xAAu;      // Unlock PLL configuration register
  PLLCON  = 0x01u;      // Normal mode, use ext. crystal
  PLLKEY2 = 0x55u;      // Write configuration
  POWKEY1 = 0x01u;      // Unlock power configuration register
  POWCON  = (0u << 0)   // Select div1 = full speed
          | (1u << 3)   // FINT allways at full speed
          | (0u << 4);  // Normal mode (not power down)
  POWKEY2 = 0xF4u;      // Write configuration
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return T0VAL;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return IRQSTA & OS_TIMER_TICK_PENDING_MASK;
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_DOWNCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  OS_INT_IncDI();
  IRQCLR = 0xFFFFFFFFu;                     // Initially disable all interrupts
  //
  // Initialize timer for embOS, assuming PLL is already initialized
  //
  T0CON  = 0u;                              // Initially disable timer
  T0LD   = (OS_TIMER_FREQ / OS_INT_FREQ);   // initialize time constant
  T0CON  = (1u << 7)                        // Enable Timer
         | (1u << 6)                        // in periodic mode
         | (0u << 4)                        // bits 5 and 4 reserved, write 0
         | (0u << 2);                       // bits 3 and 2: 00 => runs at core clock
  T0CLRI = 0u;                              // Clear timer interrupt
  IRQEN |= TIMER0_ID_MASK;                  // Enable timer interrupt
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       embOS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_U32 IntStat;

  OS_INT_Enter();
  IntStat = IRQSTA;
  //
  // Handle user specific IRQs at this point
  //
  if (IntStat & OS_TIMER_TICK_PENDING_MASK) {
    T0CLRI = 0u;                // clear timer interrupt
    OS_TICK_Handle();           // OS timer tick handler
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
    JLINKDCC_Process();
#endif
  }
  OS_INT_Leave();
}

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
OS_INTERWORK int __low_level_init(void) {
  _InitPLL();
  return 1;                               // Always return 1 to enable segment initilization !
}

/*************************** End of file ****************************/
