/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock settings
*/
#define OS_SYSTEM_FREQ   (32000000u)

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ    (OS_SYSTEM_FREQ / 2)
#define OS_TICK_FREQ     (1000u)
#define OS_INT_FREQ      (OS_TICK_FREQ)
#define OS_TIMER_RELOAD  (OS_TIMER_FREQ / OS_INT_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
#define USART0_BASE_ADDR   (0xfffc0000)
#define USART1_BASE_ADDR   (0xfffc4000)
#define USART2_BASE_ADDR   (0xfffc8000)
#define USART_CR_OFFS      (0x00)
#define USART_MR_OFFS      (0x04)
#define USART_IER_OFFS     (0x08)
#define USART_IDR_OFFS     (0x0C)
#define USART_IMR_OFFS     (0x10)
#define USART_CSR_OFFS     (0x14)
#define USART_RHR_OFFS     (0x18)
#define USART_THR_OFFS     (0x1C)
#define USART_BRGR_OFFS    (0x20)
#define USART_RTOR_OFFS    (0x24)
#define USART_TIGR_OFFS    (0x28)
//
//      PIO control register
//
#define PIO_PDR            (*(volatile OS_U32*)0xfffec004)
//
//      Timer 0
//
#define TC_CCRC0           (*(volatile OS_U32*)0xfffd0000)
#define TC_CMRC0           (*(volatile OS_U32*)0xfffd0004)
#define TC_CVC0            (*(volatile OS_U32*)0xfffd0010)
#define TC_RCC0            (*(volatile OS_U32*)0xfffd001c)
#define TC_SRC0            (*(volatile OS_U32*)0xfffd0020)
#define TC_IERC0           (*(volatile OS_U32*)0xfffd0024)
#define TC_IDRC0           (*(volatile OS_U32*)0xfffd0028)
#define TC_BMR             (*(volatile OS_U32*)0xfffd00c4)
//
//      Peripheral clock enable register
//
#define PCER               (*(volatile OS_U32*)0xffff4010)
//
//      Advanced interrupt controller (AIC)
//
#define AIC_SMR0           (*(volatile OS_U32*)0xfffff000)
#define AIC_SMR1           (*(volatile OS_U32*)0xfffff004)
#define AIC_SMR2           (*(volatile OS_U32*)0xfffff008)
#define AIC_SMR3           (*(volatile OS_U32*)0xfffff00c)
#define AIC_SMR4           (*(volatile OS_U32*)0xfffff010)
#define AIC_SMR6           (*(volatile OS_U32*)0xfffff018)
#define AIC_SVR0           (*(volatile OS_U32*)0xfffff080)
#define AIC_SVR1           (*(volatile OS_U32*)0xfffff084)
#define AIC_SVR2           (*(volatile OS_U32*)0xfffff088)
#define AIC_SVR3           (*(volatile OS_U32*)0xfffff08c)
#define AIC_SVR4           (*(volatile OS_U32*)0xfffff090)
#define AIC_SVR6           (*(volatile OS_U32*)0xfffff098)
#define AIC_IVR            (*(volatile OS_U32*)0xfffff100)
#define AIC_IPR            (*(volatile OS_U32*)0xfffff10c)
#define AIC_IDCR           (*(volatile OS_U32*)0xfffff124)
#define AIC_ICCR           (*(volatile OS_U32*)0xfffff128)
#define AIC_IECR           (*(volatile OS_U32*)0xfffff120)
#define AIC_EOICR          (*(volatile OS_U32*)0xfffff130)

#define SF_PMR             (*(volatile OS_U32*)0xfff00018)
//
//      PLL
//
#define APMC_CGMR          (*(volatile unsigned int*)0xffff4020)
#define APMC_SR            (*(volatile unsigned int*)0xffff4030)
//
//      Advanced interrupt controller interrupt sources
//
#define US0IRQ             (2)  // USART Channel 0 interrupt
#define US1IRQ             (3)  // USART Channel 1 interrupt
#define US2IRQ             (4)  // USART Channel 2 interrupt
#define TC0IRQ             (6)  // Timer Channel 0 interrupt

#define OS_TIMER_ID        (TC0IRQ)
#define OS_TIMER_PRIO      (0x00)

#define NUM_INT_SOURCES    (32)

#define INT_PRIORITY_MASK  (0x07)

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _InitPLL()
*
*  Function description
*    Initialize PLL to speed up system
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*/
static void _InitPLL(void) {
  int i;

  AIC_IDCR = 0xFFFFFFFF;       // Disable all interrupts
  AIC_ICCR = 0xFFFFFFFF;       // Clear all interrupts
  //
  // setup PLL
  //
  #define APMC_MOSCS      (1)  // Main oscillator status bit
  #define APMC_PLL_LOCK   (2)  // PLL lock status
  //
  // Speed up the System Frequency.
  //
  APMC_CGMR = 0x002F0002;      // MOSCEN = 1, OSCOUNT = 47  (1.4ms/30us)
  //
  // Read the APMC Status register to detect when the oscillator is stabilized
  //
  while ((APMC_SR & APMC_MOSCS) != APMC_MOSCS) {
    ;
  }
  //
  // Commuting from Slow Clock to Main Oscillator (16Mhz)
  //
  APMC_CGMR = 0x002F4002;      // MOSCEN = 1, OSCOUNT = 47  (1.4ms/30us)
  //
  // Setup the PLL to 32 MHz clock
  //
  APMC_CGMR = 0x032F4102;      // MUL = 1, PLLCOUNT = 3, CSS = 1
  //
  // Read the APMC Status register to detect when the PLL is stabilized
  // Wait a little extra time to ensure stable PLL.
  //
  for (i = 0; i < 1000; i++) {
    while ((APMC_SR & APMC_PLL_LOCK) != APMC_PLL_LOCK) {
      ;
    }
  }
  //
  // Commuting from 16Mhz to PLL @ 32MHz
  //
  APMC_CGMR = 0x032F8102;      // CSS = 2, MUL = 1
  //
  // Now the Master clock is the output of PLL @ 32MHz
  //
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return TC_CVC0;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return AIC_IPR & (1u << TC0IRQ);
}

/*********************************************************************
*
*       _OS_ISR_Tick()
*
*  Function description
*    This is the hardware timer exception handler.
*/
static void _OS_ISR_Tick(void) {
  volatile OS_U32 Dummy;

  Dummy = TC_SRC0;   // Reset Timer Interrupt to reset/reenable further interrupts
  OS_TICK_Handle();  // Call OS tick handler
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_Process();
#endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_DOWNCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  volatile OS_U32 Dummy;

  OS_INT_IncDI();
#if (defined(DEBUG) || defined(OS_DEBUG))               // For debugging activate AIC protected mode
  SF_PMR |= 0x20;                                       // Enable AIC protected mode in SF Protect Mode register (bit 5 AIC)
#endif
  //
  // Initialize OS timer (timer 0, capture mode channel C)
  //
  PCER     = (1 << OS_TIMER_ID);                        // Enable peripheral clock for timer 0
  TC_IDRC0 = 0xff;                                      // Disable all timer/counter 0 interrupts.
  TC_BMR   = 0x02;                                      // Runs on internal clock
  TC_CMRC0 = 0x00004000;                                // Capture mode, CPCTRG=1, TCCLKS=0 (=> MCLK/2)
  TC_RCC0  = OS_TIMER_RELOAD;                           // Set compare register
  TC_CCRC0 = 1;                                         // Enable the clock.
  TC_CCRC0 = 5;                                         // Software trigger.
  Dummy      = TC_SRC0;                                 // Read timer/counter 0 status register to clear flags.
  TC_IERC0 = (1 << 4);                                  // Interrupt on RC compare.
  //
  // Setup interrupt controller for OS timer interrupt using timer 0
  //
  OS_ARM_InstallISRHandler(OS_TIMER_ID, &_OS_ISR_Tick);  // Timer/counter 0 interrupt vector.
  OS_ARM_ISRSetPrio(OS_TIMER_ID, 0x00);                 // Timer/counter 0 interrupt level sensitive at lowest priority (0).
  AIC_ICCR = (1 << OS_TIMER_ID);                        // Clears timer/counter 0 interrupt.
  OS_ARM_EnableISR(OS_TIMER_ID);                        // Enable timer/counter 0 interrupt.
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       embOS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;

  pISR = (OS_ISR_HANDLER*)AIC_IVR;   // Read interrupt vector to release NIRQ to CPU core
#if (OS_DEBUG != 0)
  AIC_IVR = (OS_U32)pISR;            // Write back any value to IVR register to allow interrupt stacking in protected mode
#endif
  OS_INT_EnterNestable();            // Inform OS that interrupt handler is running
  pISR();                            // Call interrupt service routine
  OS_INT_Disable();                  // Disable interrupts and unlock
  AIC_EOICR = 0;                     // interrupt controller =>  Restore previous priority
  OS_INT_LeaveNestable();            // Leave interrupt, perform task switch if required
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Used to install a specific interrupt vector when ARM
*    CPUs with vectored interrupt controller are used.
*
*  Parameters
*    ISRIndex:    Index of the interrupt source, normally the interrupt vector number.
*    pISRHandler: Address of the interrupt handler function.
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler(int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;
  OS_ISR_HANDLER** papISR;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  papISR               = (OS_ISR_HANDLER**)&AIC_SVR0;
  pOldHandler          = *(papISR + ISRIndex);
  *(papISR + ISRIndex) = pISRHandler;
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Used to enable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be enabled.
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  AIC_IECR = (1u << ISRIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Used to disable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be disabled.
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  AIC_IDCR = (1u << ISRIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Function description
*    Used to set or modify the priority of a specific interrupt
*    source by programming the interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be modified.
*    Prio:     The priority which should be set for the specific interrupt.
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  OS_U32* pPrio;
  int     OldPrio;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  pPrio = (OS_U32*)&AIC_SMR0;
  OldPrio = pPrio[ISRIndex];
  pPrio[ISRIndex] = (OldPrio & ~INT_PRIORITY_MASK) | (Prio & INT_PRIORITY_MASK);
  OS_INT_EnableConditional();
  return OldPrio & INT_PRIORITY_MASK;
}

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);     // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
OS_INTERWORK int __low_level_init(void) {
  _InitPLL();
  return 1;  // Always return 1 to enable segment initialization!
}

/*************************** End of file ****************************/
