/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Clock settings
*/
#define OS_SYSTEM_FREQ     (47923200uL)

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ      (OS_SYSTEM_FREQ)
#define OS_TICK_FREQ       (1000uL)
#define OS_INT_FREQ        (OS_TICK_FREQ)
#define OS_TIMER_PRESCALE  (16uL)
#define OS_TIMER_RELOAD    ((OS_TIMER_FREQ / OS_TIMER_PRESCALE / OS_INT_FREQ) - 1)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0uL)
  #define OS_BAUDRATE  (38400uL)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
//
//      Debug unit
//
#define DBGU_BASE_ADDR    (0xFFFFF200uL)
#define DBGU_CPUID        (*(volatile OS_U32*) (DBGU_BASE_ADDR + 0x40uL))
//
//      Reset controller
//
#define RSTC_BASE_ADDR      (0xFFFFFD00uL)
#define RSTC_CR             (*(volatile OS_U32*)(RSTC_BASE_ADDR + 0x00uL))
#define RSTC_SR             (*(volatile OS_U32*)(RSTC_BASE_ADDR + 0x04uL))
#define RSTC_MR             (*(volatile OS_U32*)(RSTC_BASE_ADDR + 0x08uL))
#define RSTC_BODIEN         (1u << 16)  // Brownout interrupt enable
#define RSTC_URSTIEN        (1u <<  4)  // User reset interrupt enable
#define RSTC_BODSTS         (1u <<  1)  // Brownout status
#define RSTC_URSTS          (1u <<  0)  // User reset status
//
//      Real time timer
//
#define RTT_BASE_ADDR       (0xFFFFFD20uL)
#define RTT_MR              (*(volatile OS_U32*)(RTT_BASE_ADDR + 0x00uL))
#define RTT_SR              (*(volatile OS_U32*)(RTT_BASE_ADDR + 0x0CuL))
#define RTT_RTTINCIEN       (1u << 17)
#define RTT_ALMIEN          (1u << 16)
#define RTT_RTTINC          (1u <<  1)
#define RTT_ALMS            (1u <<  0)
//
//      Periodic interval timer
//
#define PIT_BASE_ADDR       (0xFFFFFD30uL)
#define PIT_MR              (*(volatile OS_U32*)(PIT_BASE_ADDR + 0x00uL))
#define PIT_SR              (*(volatile OS_U32*)(PIT_BASE_ADDR + 0x04uL))
#define PIT_PIVR            (*(volatile OS_U32*)(PIT_BASE_ADDR + 0x08uL))
#define PIT_PIIR            (*(volatile OS_U32*)(PIT_BASE_ADDR + 0x0CuL))
//
//      Watchdog
//
#define WDT_BASE_ADDR       (0xFFFFFD40uL)
#define WDT_CR              (*(volatile OS_U32*)(WDT_BASE_ADDR + 0x00uL))
#define WDT_MR              (*(volatile OS_U32*)(WDT_BASE_ADDR + 0x04uL))
#define WDT_SR              (*(volatile OS_U32*)(WDT_BASE_ADDR + 0x08uL))
#define WDT_WDFIEN          (1uL << 12)  // Watchdog interrupt enable flag in mode register
#define WDT_WDERR           (1uL <<  1)  // Watchdog error status flag
#define WDT_WDUNF           (1uL <<  0)  // Watchdog underflow status flag
//
//      Memory controller
//
#define MC_BASE_ADDR        (0xFFFFFF00uL)
#define MC_FMR0             (*(volatile OS_U32*)(MC_BASE_ADDR + 0x60uL))
#define MC_FMR1             (*(volatile OS_U32*)(MC_BASE_ADDR + 0x70uL))
#define MC_FWS_0FWS         (0uL << 8)  // (MC) 1 cycle for Read,  2 for Write operations
#define MC_FWS_1FWS         (1uL << 8)  // (MC) 2 cycles for Read, 3 for Write operations
#define MC_FWS_2FWS         (2uL << 8)  // (MC) 3 cycles for Read, 4 for Write operations
#define MC_FWS_3FWS         (3uL << 8)  // (MC) 4 cycles for Read, 4 for Write operations
#if   (OS_SYSTEM_FREQ <= 30000000)
  #define MC_FWS (MC_FWS_0FWS)
#elif (OS_SYSTEM_FREQ <= 50000000)
  #define MC_FWS (MC_FWS_1FWS)
#else
  #error "Please examine settings for memory controller, Flash wait states"
#endif
//
//      PIO control register
//
#define PIOA_BASE_ADDR      (0xfffff400uL)
#define PIOA_PDR            (*(volatile OS_U32*)(_PIOA_BASE_ADDR + 0x04uL))  // PIOA disable register
#define PIOA_ASR            (*(volatile OS_U32*)(_PIOA_BASE_ADDR + 0x70uL))  // PIOA "A" peripheral select register
#define PIOA_BSR            (*(volatile OS_U32*)(_PIOA_BASE_ADDR + 0x74uL))  // PIOA "B" peripheral select register
//
//      Power management controller
//
#define PMC_BASE_ADDR       (0xFFFFFC00uL)
#define PMC_SCDR            (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x04uL))  // System Clock Disable Register
#define PMC_PCER            (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x10uL))  // Peripheral clock enable register
#define PMC_MOR             (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x20uL))  // main oscillator register
#define PMC_PLLR            (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x2cuL))  // PLL register
#define PMC_MCKR            (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x30uL))  // Master clock register
#define PMC_SR              (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x68uL))  // status register
#define PMC_IMR             (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x6CuL))  // interrupt mask register
#define PMC_PCKRDY2         (1uL << 10)
#define PMC_PCKRDY1         (1uL <<  9)
#define PMC_PCKRDY0         (1uL <<  8)
#define PMC_MCKRDY          (1uL <<  3)
#define PMC_LOCK            (1uL <<  2)
#define PMC_MOSCS           (1uL <<  0)
#define PMC_MASK_ALL        (PMC_PCKRDY2 | PMC_PCKRDY1 | PMC_PCKRDY0 | \
                             PMC_MCKRDY  | PMC_LOCK    | PMC_MOSCS)
//
//      Advanced interrupt controller (AIC)
//
#define AIC_BASE_ADDR       (0xfffff000uL)
#define AIC_SMR_BASE_ADDR   (AIC_BASE_ADDR + 0x00uL)
#define AIC_SVR_BASE_ADDR   (AIC_BASE_ADDR + 0x80uL)
#define AIC_SVR0            (*(volatile OS_U32*)(AIC_SVR_BASE_ADDR + 0x000uL))
#define AIC_SVR1            (*(volatile OS_U32*)(AIC_SVR_BASE_ADDR + 0x004uL))
#define AIC_IVR             (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x100uL))
#define AIC_ISR             (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x108uL))
#define AIC_IPR             (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x10cuL))
#define AIC_IDCR            (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x124uL))
#define AIC_ICCR            (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x128uL))
#define AIC_IECR            (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x120uL))
#define AIC_EOICR           (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x130uL))
#define AIC_SPU             (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x134uL))
#define AIC_DCR             (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x138uL))
//
//      AIC interrupt sources and peripheral IDs
//
#define SYSTEM_IRQ_ID       ( 1uL)  // System IRQ ID
#define US0IRQ_ID           ( 6uL)  // USART Channel 0 interrupt
#define US1IRQ_ID           ( 7uL)  // USART Channel 1 interrupt
#define US2IRQ_ID           ( 8uL)  // USART Channel 2 interrupt
#define TC0IRQ_ID           (12uL)  // Timer Channel 0 interrupt

#define NUM_INT_SOURCES     (32uL)

#define INT_PRIORITY_MASK   (7uL)
#define NUM_INT_PRIORITIES  (8uL)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _HandlePmcIrq()
*
*  Function description
*    Power management controller interrupt
*
*  Notes
*    (1) If not used for application, this handler may be removed
*/
static void _HandlePmcIrq(void) {
#if (OS_DEBUG != 0)
  OS_U32 IrqSource;

  IrqSource  = PMC_IMR;
  IrqSource &= (PMC_SR & PMC_MASK_ALL);
  if (IrqSource) {  // PMC interrupt pending?
    while (1) {
      ;             // Not implemented
    }
  }
#endif
}

/*********************************************************************
*
*       _HandleRttIrq()
*
*  Function description
*    Real time timer interrupt handler
*
*  Notes
*    (1) If not used for application, this handler may be removed
*/
static void _HandleRttIrq(void) {
#if (OS_DEBUG != 0)
  OS_U32 IrqStatus;
  OS_U32 IrqEnabled;

  IrqEnabled = RTT_MR & (RTT_RTTINCIEN  | RTT_ALMIEN);
  IrqStatus  = RTT_SR & (RTT_RTTINC | RTT_ALMS);
  if ((IrqStatus & RTT_RTTINC) && (IrqEnabled & RTT_RTTINCIEN )) {  // RTT inc. interrupt pending?
    while (1) {
      ;                                                             // Not implemented
    }
  }
  if ((IrqStatus & RTT_ALMS) && (IrqEnabled & RTT_ALMIEN )) {       // Alarm interrupt pending?
    while (1) {
      ;                                                             // Not implemented
    }
  }
#endif
}

/*********************************************************************
*
*       _HandleRstcIrq()
*
*  Function description
*    Reset controller interrupt handler
*
*  Notes
*    (1) If not used for application, this handler may be removed
*/
static void _HandleRstcIrq(void) {
#if (OS_DEBUG != 0)
  OS_U32 IrqStatus;
  OS_U32 IrqEnabled;

  IrqEnabled = RSTC_MR & (RSTC_BODIEN | RSTC_URSTIEN);
  IrqStatus  = RSTC_SR & (RSTC_BODSTS | RSTC_URSTS);
  if ((IrqStatus & RSTC_BODSTS) && (IrqEnabled & RSTC_BODIEN )) {  // Brownout interrupt pending?
    while (1) {
      ;                                                            // Not implemented
    }
  }
  if ((IrqStatus & RSTC_URSTS) && (IrqEnabled & RSTC_URSTIEN )) {  // User reset interrupt pending?
    while (1) {
      ;                                                            // Not implemented
    }
  }
#endif
}

/*********************************************************************
*
*       _HandleWdtIrq()
*
*  Function description
*    watchdog timer interrupt handler
*
*  Notes
*    (1) If not used for application, this handler may be removed
*/
static void _HandleWdtIrq(void) {
#if (OS_DEBUG != 0)
  OS_U32 IrqStatus;

  IrqStatus = WDT_SR & (WDT_WDERR | WDT_WDUNF);
  if (IrqStatus && (WDT_MR & WDT_WDFIEN)) {  // Watchdog error interrupt pending?
    while (1) {
      ;                                      // Not implemented
    }
  }
#endif
}

/*********************************************************************
*
*       _DefaultFiqHandler()
*
*  Function description
*    a dummy FIQ handler
*/
static void _DefaultFiqHandler(void) {
  while (1) {
    ;  // Not implemented
  }
}

/*********************************************************************
*
*       _DefaultIrqHandler()
*
*  Function description
*    a dummy IRQ handler
*
*  Notes
*    (1) This handler is initially written into all AIC interrupt vectors
*        It is called if no interrupt vector was installed for specific
*        interrupt source.
*        May be used during debugging to detect uninstalled interrupts.
*/
static void _DefaultIrqHandler(void) {
  volatile OS_U32 IrqSource;

  IrqSource = AIC_ISR;  // detect source of uninstalled interrupt
  while (1) {
    ;
  }
}

/*********************************************************************
*
*       _SpuriousIrqHandler()
*
*  Function description
*    default spurious IRQ handler
*
*  Notes
*    A spurious interrupt is called under one of the following
*    circumstances:
*    1. An external interrupt source is programmed in level-sensitive mode
*       and an active level occurs for only a short time.
*    2. An internal interrupt source is programmed in level sensitive
*       and the output signal of the corresponding embedded peripheral
*       is activated for a short time. (As in the case for the Watchdog.)
*    3. An interrupt occurs just a few cycles before the software begins
*       to mask it, thus resulting in a pulse on the interrupt source.
*
*    The spurious handler is called to detect these errors.
*/
static void _SpuriousIrqHandler(void) {
#if (OS_DEBUG != 0)
  static OS_U32 _SpuriousIrqCnt;

  _SpuriousIrqCnt++;
#endif
}

/*********************************************************************
*
*       _InitAIC()
*
*  Function description
*    Initialize interrupt controller by setting default vectors
*    and clearing all interrupts
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*/
static void _InitAIC(void) {
  unsigned int     i;
  OS_ISR_HANDLER** papISR;

  AIC_IDCR = 0xFFFFFFFFuL;                    // Disable all interrupts
  AIC_ICCR = 0xFFFFFFFFuL;                    // Clear all interrupts
  AIC_SVR0 = (int)_DefaultFiqHandler;         // Dummy FIQ handler
  AIC_SPU  = (int) _SpuriousIrqHandler;       // Dummy spurious handler
  papISR   = (OS_ISR_HANDLER**)AIC_SVR_BASE_ADDR;
  for (i = 1uL; i < NUM_INT_SOURCES; i++)  {  // Initially set all sources
    *(papISR + i) = &_DefaultIrqHandler;      // to dummy irq handler
  }
  for (i = 0; i < NUM_INT_PRIORITIES; i++) {
    AIC_EOICR = 0;                            // Reset interrupt controller
  }
#if OS_DEBUG                                  // For debugging activate AIC protected mode
  AIC_DCR |= 0x01u;                           // Enable AIC protected mode
#endif
}

/*********************************************************************
*
*       _InitRTT()
*
*  Function description
*    Initialize Real time timer.
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*    (2) Not used by embOS, we disable interrupts here to avoid unhandled interrupts
*        May be modified by user if RTT is required for application.
*/
static void _InitRTT(void) {
  //
  // Disable Real-Time Timer interrupts
  //
  RTT_MR &= ~(RTT_RTTINCIEN | RTT_ALMIEN);
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return (PIT_PIIR & 0x000FFFFFuL);
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return (PIT_PIIR & 0xFFF00000uL);
}

/*********************************************************************
*
*       _OS_ISR_Tick()
*
*  Function description
*    This is the hardware timer exception handler.
*/
static void _OS_ISR_Tick(void) {
  volatile OS_U32 Dummy;

  if (PIT_SR & (1uL << 0)) {  // Timer interrupt pending?
    Dummy = PIT_PIVR;         // Reset interrupt pending condition
    OS_USEPARA(Dummy);
    OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
    JLINKDCC_Process();
#endif
  }
  //
  // Call to following handlers may be removed if not used by application
  //
  _HandlePmcIrq();
  _HandleRttIrq();
  _HandleRstcIrq();
  _HandleWdtIrq();
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  OS_INT_IncDI();
  //
  // Initialize timer for embOS, assuming PLL is already initialized
  //
  PIT_MR = ((OS_TIMER_RELOAD & 0x000FFFFFuL) | (1uL << 25) | (1uL << 24));
  //
  // Install PIT ISR Handler
  //
  OS_ARM_InstallISRHandler(SYSTEM_IRQ_ID, _OS_ISR_Tick);
  OS_ARM_EnableISR(SYSTEM_IRQ_ID);
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
    PMC_SCDR = 1uL;     // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       embOS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;

  pISR = (OS_ISR_HANDLER*)AIC_IVR;   // Read interrupt vector to release NIRQ to CPU core
#if (OS_DEBUG != 0)
  AIC_IVR = (OS_U32)pISR;            // Write back any value to IVR register to allow interrupt stacking in protected mode
#endif
  OS_INT_EnterNestable();            // Inform OS that interrupt handler is running
  pISR();                            // Call interrupt service routine
  OS_INT_Disable();                  // Disable interrupts and unlock
  AIC_EOICR = 0;                     // interrupt controller =>  Restore previous priority
  OS_INT_LeaveNestable();            // Leave interrupt, perform task switch if required
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Used to install a specific interrupt vector when ARM
*    CPUs with vectored interrupt controller are used.
*
*  Parameters
*    ISRIndex:    Index of the interrupt source, normally the interrupt vector number.
*    pISRHandler: Address of the interrupt handler function.
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler(int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;
  OS_ISR_HANDLER** papISR;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  papISR = (OS_ISR_HANDLER**)AIC_SVR_BASE_ADDR;
  pOldHandler          = *(papISR + ISRIndex);
  *(papISR + ISRIndex) = pISRHandler;
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Used to enable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be enabled.
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  AIC_IECR = (1uL << ISRIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Used to disable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be disabled.
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  AIC_IDCR = (1uL << ISRIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Function description
*    Used to set or modify the priority of a specific interrupt
*    source by programming the interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be modified.
*    Prio:     The priority which should be set for the specific interrupt.
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  OS_U32* pPrio;
  int     OldPrio;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  pPrio = (OS_U32*)AIC_SMR_BASE_ADDR;
  OldPrio = pPrio[ISRIndex];
  pPrio[ISRIndex] = (OldPrio & ~INT_PRIORITY_MASK) | (Prio & INT_PRIORITY_MASK);
  OS_INT_EnableConditional();
  return OldPrio & INT_PRIORITY_MASK;
}

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
//
// Calculation of Flash microsecond cycle numbers for flash programming
// MCLK frequency (in MHz) * 1.5. The value has to be rounded up
//
#define FMCN_VAL ((((OS_SYSTEM_FREQ + 500000) / 1000000) * 3 ) >> 1)

OS_INTERWORK int __low_level_init(void) {
  WDT_MR = (1uL << 15);                     // Initially disable watchdog
  //
  // Setup Flash microsecond cycle numbers and wait states
  //
  MC_FMR0 = ((0xFFuL << 16) & (FMCN_VAL << 16)) | MC_FWS;
  if (((DBGU_CPUID >> 8) & 0xFFuL) >= 0x09u) {
    //
    // The Flash size is above 256KB, initialize the second Flash controller
    //
    MC_FMR1 = ((0xFFuL << 16) & (FMCN_VAL << 16)) | MC_FWS;
  }
  //
  // Start-up main oscillator
  //
  PMC_MOR  = (   1uL << 0)     // Enable main oscillator
           | (0x40uL << 8)     // Set startup delay time, 512 slow clock cycles
           ;
  //
  // Wait until end of startup time => MOSCS set to 1
  //
  while ((PMC_SR & (1uL << 0)) == 0);
  //
  // Initialize PLL to run at 96 MHz
  //
  PMC_PLLR = 0x05uL            // Divider = 5
            | (28uL  <<  8)    //
            | (25uL  << 16)    //
            | ( 1uL  << 28)    // USBDIV is value + 1 = 2
            ;
  //
  // Wait until pll is locked, until LOCK bit is set to 1
  //
  while((PMC_SR & (1uL << 2)) == 0);
  //
  // Wait until master clock is ready, until MCKRDY is set to 1
  //
  while((PMC_SR & (1uL << 3)) == 0);
  //
  // Select main clock as master clock, set divider
  //
  PMC_MCKR = 0                // Select main clock oscillator as source
           | (1uL << 2)       // Divide clock by 2
           ;
  //
  // Wait until master clock is ready, until MCKRDY is set to 1
  //
  while((PMC_SR & (1uL << 3)) == 0);
  //
  // Select PLL clock as master clock, set divider
  //
  PMC_MCKR = 3                // Select PLL clock as source
           | (1uL <<  2)      // Divide clock by 2
           ;
  //
  // Wait until master clock is ready, until MCKRDY is set to 1
  //
  while((PMC_SR & (1uL << 3)) == 0);
  //
  // Initialize interrupt controller
  //
  _InitAIC();                 // Initialize interrupt controller, clear and reset all interrupts
  //
  // Initialize real time timer
  //
  _InitRTT();
  //
  // Initialize reset control
  //
  RSTC_MR = ((0xA5uL << 24) | (1uL << 0));  // write KEY and URSTEN to allow USER RESET
  return 1;                                 // Always return 1 to enable segment initialization!
}

/*********************************************************************
*
*       Optional speed opimization for IAR
*
**********************************************************************
*/
#ifdef __ICCARM__

#ifndef   OPTIMIZE_SPEED
  #define OPTIMIZE_SPEED (0)
#endif

#if OPTIMIZE_SPEED
/*********************************************************************
*
*       OS_EnableInt()
*/
OS_INTERWORK __arm __ramfunc void OS_EnableInt(void) {
  __set_CPSR(__get_CPSR() & ~0x80uL);  // Clear IRQ only, keep other bits unchanged
}

/*********************************************************************
*
*       OS_DisableInt()
*/
OS_INTERWORK __arm __ramfunc void OS_DisableInt(void) {
  __set_CPSR(__get_CPSR() | 0x80uL);  // Set IRQ only, keep other bits unchanged
}
#endif  // OPTIMIZE_SPEED
#endif  // __ICCARM__

/*************************** End of file ****************************/
