/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Configuration
*/

#ifndef   DDRAM_BASE_ADDR
  #define DDRAM_BASE_ADDR  (0x70000000)  // DDRAM on eval board
#endif

/*********************************************************************
*
*       Clock settings
*/
#define OS_SYSTEM_FREQ  (400000000u)

#define DIV_PLLA        (   3u)     // PLL input clock divider
#define MUL_PLLA        ( 200u)     // Multiplier
#define OUT_PLLA        (0x00u)     // High speed setting
#define COUNT_PLLA      (0x3Fu)     // Startup counter

#define PLLAR_VALUE     ((1u << 29)           \
                        |((MUL_PLLA-1) << 16) \
                        |(OUT_PLLA     << 14) \
                        |(COUNT_PLLA   << 8)  \
                        |(DIV_PLLA     << 0))

#define MCKR_MDIV       (0x03)       // Main clock is processor clock / 3
#define MCKR_PRES       (0)          // Processor clock is selected clock
#define MCKR_CSS        (0x02)       // PLLA is selected clock
#define MCKR_PLLADIV2   (1u << 12)   // PLLA clock is divided by 2

#define MCKR_VALUE      ((MCKR_PLLADIV2)    \
                        |(MCKR_MDIV << 8)   \
                        |(MCKR_PRES << 2))

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ      (OS_SYSTEM_FREQ / 3 / 16)
#define OS_TICK_FREQ       (1000u)
#define OS_INT_FREQ        (OS_TICK_FREQ)
#define OS_TIMER_RELOAD    ((OS_TIMER_FREQ / OS_INT_FREQ) - 1)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (1u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
//
//      Reset controller
//
#define RSTC_BASE_ADDR           (0xFFFFFD00u)
#define RSTC_CR                  (*(volatile OS_U32*)(RSTC_BASE_ADDR + 0x00u))
#define RSTC_SR                  (*(volatile OS_U32*)(RSTC_BASE_ADDR + 0x04u))
#define RSTC_MR                  (*(volatile OS_U32*)(RSTC_BASE_ADDR + 0x08u))
#define RSTC_URSTEN              (1u <<  0)                                         // User reset enable
#define RSTC_BODIEN              (1u << 16)                                         // Brownout interrupt enable
#define RSTC_URSTIEN             (1u <<  4)                                         // User reset interrupt enable
#define RSTC_BODSTS              (1u <<  1)                                         // Brownout status
#define RSTC_URSTS               (1u <<  0)                                         // User reset status
//
//      Real time timer
//
#define RTT_BASE_ADDR            (0xFFFFFD20u)
#define RTT_MR                   (*(volatile OS_U32*)(RTT_BASE_ADDR + 0x00u))
#define RTT_SR                   (*(volatile OS_U32*)(RTT_BASE_ADDR + 0x0Cu))
#define RTT_RTTINCIEN            (1u << 17)
#define RTT_ALMIEN               (1u << 16)
#define RTT_RTTINC               (1u << 1)
#define RTT_ALMS                 (1u << 0)
//
//      Periodic interval timer
//
#define PIT_BASE_ADDR            (0xFFFFFD30u)
#define PIT_MR                   (*(volatile OS_U32*)(PIT_BASE_ADDR + 0x00u))
#define PIT_SR                   (*(volatile OS_U32*)(PIT_BASE_ADDR + 0x04u))
#define PIT_PIVR                 (*(volatile OS_U32*)(PIT_BASE_ADDR + 0x08u))
#define PIT_PIIR                 (*(volatile OS_U32*)(PIT_BASE_ADDR + 0x0Cu))
//
//      Watchdog
//
#define WDT_BASE_ADDR            (0xFFFFFD40u)
#define WDT_CR                   (*(volatile OS_U32*)(WDT_BASE_ADDR + 0x00u))
#define WDT_MR                   (*(volatile OS_U32*)(WDT_BASE_ADDR + 0x04u))
#define WDT_SR                   (*(volatile OS_U32*)(WDT_BASE_ADDR + 0x08u))
#define WDT_WDFIEN               (1u << 12)                                         // Watchdog interrupt enable flag in mode register
#define WDT_WDERR                (1u <<  1)                                         // Watchdog error status flag
#define WDT_WDUNF                (1u <<  0)                                         // Watchdog underflow status flag
//
//      PIO control register
//
#define PIOA_BASE_ADDR           (0xFFFFF200u)
#define PIOB_BASE_ADDR           (0xFFFFF400u)
#define PIOC_BASE_ADDR           (0xFFFFF600u)
#define PIOD_BASE_ADDR           (0xFFFFF800u)
#define PIOE_BASE_ADDR           (0xFFFFFA00u)
//
//      PIOC, used as data BUS
//
#define PIOA_PDR                 (*(volatile OS_U32*) (PIOA_BASE_ADDR + 0x04))    // PIOA disable register
#define PIOA_MDDR                (*(volatile OS_U32*) (PIOA_BASE_ADDR + 0x54))    // PIOA multi driver disable register
#define PIOA_ASR                 (*(volatile OS_U32*) (PIOA_BASE_ADDR + 0x70))    // PIOA peripheral A select register
//
//      Power management controller
//
#define PMC_BASE_ADDR            (0xFFFFFC00u)
#define PMC_CKGR_PLLAR           (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x28u))       // PLLA register
#define PMC_CKGR_PLLBR           (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x2cu))       // PLLB register
#define PMC_SCER                 (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x000))       // System Clock enable register
#define PMC_SCDR                 (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x04u))       // System clock disable register
#define PMC_PCER                 (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x10u))       // Peripheral clock enable register
#define PMC_MOR                  (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x20u))       // Main oscillator register
#define PMC_PLLR                 (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x2cu))       // PLL register
#define PMC_MCKR                 (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x30u))       // Master clock register
#define PMC_SR                   (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x68u))       // Status register
#define PMC_IMR                  (*(volatile OS_U32*)(PMC_BASE_ADDR + 0x6Cu))       // Interrupt mask register
#define PMC_LOCKA                (1u <<  1)
#define PMC_LOCKB                (1u <<  2)
#define PMC_LOCKU                (1u <<  6)
#define PMC_PCKRDY1              (1u <<  9)
#define PMC_PCKRDY0              (1u <<  8)
#define PMC_MCKRDY               (1u <<  3)
#define PMC_MOSCS                (1u <<  0)
#define PMC_MASK_ALL             (PMC_PCKRDY1 | PMC_PCKRDY0 | PMC_LOCKU | PMC_MCKRDY | PMC_LOCKA | PMC_MOSCS)
//
//      Advanced interrupt controller (AIC)
//
#define AIC_BASE_ADDR            (0xFFFFF000u)
#define AIC_SMR_BASE_ADDR        (AIC_BASE_ADDR + 0x00u)
#define AIC_SVR_BASE_ADDR        (AIC_BASE_ADDR + 0x80u)
#define AIC_SVR0                 (*(volatile OS_U32*)(AIC_SVR_BASE_ADDR + 0x000u))
#define AIC_SVR1                 (*(volatile OS_U32*)(AIC_SVR_BASE_ADDR + 0x004u))
#define AIC_IVR                  (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x100u))
#define AIC_ISR                  (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x108u))
#define AIC_IPR                  (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x10cu))
#define AIC_IDCR                 (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x124u))
#define AIC_ICCR                 (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x128u))
#define AIC_IECR                 (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x120u))
#define AIC_EOICR                (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x130u))
#define AIC_SPU                  (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x134u))
#define AIC_DCR                  (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x138u))
#define AIC_FFDR                 (*(volatile OS_U32*)(AIC_BASE_ADDR     + 0x144u))
//
//      AIC interrupt sources and peripheral IDs
//
#define SYSTEM_IRQ_ID            (1u)   // System IRQ ID
#define PIOA_ID                  (2u)
#define PIOB_ID                  (3u)
#define PIOC_ID                  (4u)
#define PIODE_ID                 (5u)

#ifndef   NUM_INT_SOURCES
  #define NUM_INT_SOURCES        (32u)
#endif

#define INT_PRIORITY_MASK        (7u)
#define NUM_INT_PRIORITIES       (8u)
//
//      MATRIX + EBI interface
//
#define MATRIX_BASE_ADDR         (0xFFFFEA00)                                // MATRIX Base Address

#define MATRIX_MCFG              (*(volatile OS_U32*) (MATRIX_BASE_ADDR + 0x000)) // MATRIX Master configuration register
#define MATRIX_EBICSA            (*(volatile OS_U32*) (MATRIX_BASE_ADDR + 0x128)) // MATRIX EBI Chip Select Assignment register
//
//      SDRAM controller
//
#define SDRAMC_BASE_ADDR         (0xFFFFE200)   // SDRAMC Base Address
#define SDRAMC_MR                (*(volatile OS_U32*) (SDRAMC_BASE_ADDR + 0x00)) // (SDRAMC) SDRAM Controller Mode Register
#define SDRAMC_TR                (*(volatile OS_U32*) (SDRAMC_BASE_ADDR + 0x04)) // (SDRAMC) SDRAM Controller Refresh timer Register
#define SDRAMC_CR                (*(volatile OS_U32*) (SDRAMC_BASE_ADDR + 0x08)) // (SDRAMC) SDRAM Controller Configuration Register
#define SDRAMC_LPR               (*(volatile OS_U32*) (SDRAMC_BASE_ADDR + 0x10)) // (SDRAMC) SDRAM Controller Low Power Register
#define SDRAMC_MDR               (*(volatile OS_U32*) (SDRAMC_BASE_ADDR + 0x24)) // (SDRAMC) SDRAM Controller Memory Device Register

#define SDRAMC_MODE_NORMAL_CMD   (0x0) // (SDRAMC) Normal Mode
#define SDRAMC_MODE_NOP_CMD      (0x1) // (SDRAMC) Issue a All Banks Precharge Command at every access
#define SDRAMC_MODE_PRCGALL_CMD  (0x2) // (SDRAMC) Issue a All Banks Precharge Command at every access
#define SDRAMC_MODE_LMR_CMD      (0x3) // (SDRAMC) Issue a Load Mode Register at every access
#define SDRAMC_MODE_RFSH_CMD     (0x4) // (SDRAMC) Issue a Refresh
//
//      SMC, static memory controller
//
#define SMC_BASE_ADDR            (0xFFFFE400)                   // SMC  Base Address
#define SMC2_BASE_ADDR           (SMC_BASE_ADDR + (0x10 * 2))   // SMC2 Base Address
#define SMC_SETUP2               (*(volatile OS_U32*) (SMC2_BASE_ADDR + 0x00))
#define SMC_PULSE2               (*(volatile OS_U32*) (SMC2_BASE_ADDR + 0x04))
#define SMC_CYCLE2               (*(volatile OS_U32*) (SMC2_BASE_ADDR + 0x08))
#define SMC_MODE2                (*(volatile OS_U32*) (SMC2_BASE_ADDR + 0x0C))
//
//      Utility macros
//
#define __readMemory32(Addr, Zone)         (*(volatile OS_U32*) (Addr))
#define __writeMemory32(Data, Addr, Zone)  ((*(volatile OS_U32*) (Addr)) = Data)


/*********************************************************************
*
*       local data
*
**********************************************************************
*/

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

/*********************************************************************
*
*       MMU and cache configuration
*
*  Notes
*    The translation table has to be aligned on 16384 bytes boundary
*    The translation table has to be located in un-initialized memory,
*    because it may be filled from startup, before segments and variables
*    are initialized.
*/
#if   (defined(__ICCARM__))  // IAR
  #pragma data_alignment=16384
  static __no_init OS_U32 _TranslationTable[0x1000];
#elif (defined(__CC_ARM))
  #pragma arm section zidata = "NO_INIT"  // Has to be declared as separate section UNINIT in the scatter file
  __align(16384) static OS_U32 _TranslationTable[0x1000];
  #pragma arm section zidata              // back to default (.bss section)
#elif (defined(__GNUC__))    // GCC
  static OS_U32 _TranslationTable[0x1000] __attribute__ ((aligned (16384), section (".no_init")));
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       OS_GetPC()
*/
#ifdef __ICCARM__
#pragma diag_suppress=Pe940 // IAR specific: No warning for missing return
static OS_U32 OS_GetPC(void) {
  #if (__VER__ < 6300000)
    __asm("mov  r0, lr");
    __asm("subs r0, r0, #4");
  #else
    __asm volatile("mov  r0, lr");
    __asm volatile("subs r0, r0, #4");
  #endif
}
#elif defined __CC_ARM
static __asm OS_U32 OS_GetPC(void) {
  mov  r0, lr
  subs r0, r0, #4
  bx   lr
}
#elif defined __GNUC__
OS_U32 OS_GetPC(void);
#endif

/*********************************************************************
*
*       __delay()
*/
static void __delay(OS_U32 Loops) {
  volatile unsigned int i;
  #define LOOP_CNT (OS_SYSTEM_FREQ / 10000u)
  while (Loops) {  // One loop should last 1ms
    for (i = 0; i < LOOP_CNT; i++) {
    }
    Loops--;
  }
}

/*********************************************************************
*
*       _HandlePmcIrq()
*
*  Function description
*    Power management controller interrupt
*
*  Notes
*    (1) If not used for application, this handler may be removed
*/
static void _HandlePmcIrq(void) {
#if (OS_DEBUG != 0)
  OS_U32 IrqSource;

  IrqSource  = PMC_IMR;
  IrqSource &= (PMC_SR & PMC_MASK_ALL);
  if (IrqSource) {  // PMC interrupt pending?
    while (1) {
      ;             // Not implemented
    }
  }
#endif
}

/*********************************************************************
*
*       _HandleRttIrq()
*
*  Function description
*    Real time timer interrupt handler
*
*  Notes
*    (1) If not used for application, this handler may be removed
*/
static void _HandleRttIrq(void) {
#if (OS_DEBUG != 0)
  OS_U32 IrqStatus;
  OS_U32 IrqEnabled;

  IrqEnabled = RTT_MR & (RTT_RTTINCIEN  | RTT_ALMIEN);
  IrqStatus  = RTT_SR & (RTT_RTTINC | RTT_ALMS);
  if ((IrqStatus & RTT_RTTINC) && (IrqEnabled & RTT_RTTINCIEN )) {  // RTT inc. interrupt pending?
    while (1) {
      ;                                                             // Not implemented
    }
  }
  if ((IrqStatus & RTT_ALMS) && (IrqEnabled & RTT_ALMIEN )) {       // Alarm interrupt pending?
    while (1) {
      ;                                                             // Not implemented
    }
  }
#endif
}

/*********************************************************************
*
*       _HandleRstcIrq()
*
*  Function description
*    Reset controller interrupt handler
*
*  Notes
*    (1) If not used for application, this handler may be removed
*/
static void _HandleRstcIrq(void) {
#if (OS_DEBUG != 0)
  OS_U32 IrqStatus;
  OS_U32 IrqEnabled;

  IrqEnabled = RSTC_MR & (RSTC_BODIEN | RSTC_URSTIEN);
  IrqStatus  = RSTC_SR & (RSTC_BODSTS | RSTC_URSTS);
  if ((IrqStatus & RSTC_BODSTS) && (IrqEnabled & RSTC_BODIEN )) {  // Brownout interrupt pending?
    while (1) {
      ;                                                            // Not implemented
    }
  }
  if ((IrqStatus & RSTC_URSTS) && (IrqEnabled & RSTC_URSTIEN )) {  // User reset interrupt pending?
    while (1) {
      ;                                                            // Not implemented
    }
  }
#endif
}

/*********************************************************************
*
*       _HandleWdtIrq()
*
*  Function description
*    watchdog timer interrupt handler
*
*  Notes
*    (1) If not used for application, this handler may be removed
*/
static void _HandleWdtIrq(void) {
#if (OS_DEBUG != 0)
  OS_U32 IrqStatus;

  IrqStatus = WDT_SR & (WDT_WDERR | WDT_WDUNF);
  if (IrqStatus && (WDT_MR & WDT_WDFIEN)) {  // Watchdog error interrupt pending?
    while (1) {
      ;                                      // Not implemented
    }
  }
#endif
}

/*********************************************************************
*
*       _DefaultFiqHandler()
*
*  Function description
*    a dummy FIQ handler
*/
static void _DefaultFiqHandler(void) {
  while (1) {
    ;  // Not implemented
  }
}

/*********************************************************************
*
*       _DefaultIrqHandler()
*
*  Function description
*    A dummy IRQ handler
*
*  Notes
*    (1) This handler is initially written into all AIC interrupt vectors
*        It is called if no interrupt vector was installed for specific
*        interrupt source.
*        May be used during debugging to detect uninstalled interrupts.
*/
static void _DefaultIrqHandler(void) {
  volatile OS_U32 IrqSource;

  IrqSource = AIC_ISR;  // detect source of uninstalled interrupt
  while (1) {
    ;
  }
}

/*********************************************************************
*
*       _SpuriousIrqHandler()
*
*  Function description
*    default spurious IRQ handler
*
*  Notes
*    A spurious interrupt is called under one of the following
*    circumstances:
*    1. An external interrupt source is programmed in level-sensitive mode
*       and an active level occurs for only a short time.
*    2. An internal interrupt source is programmed in level sensitive
*       and the output signal of the corresponding embedded peripheral
*       is activated for a short time. (As in the case for the Watchdog.)
*    3. An interrupt occurs just a few cycles before the software begins
*       to mask it, thus resulting in a pulse on the interrupt source.
*
*    The spurious handler is called to detect these errors.
*/
static void _SpuriousIrqHandler(void) {
#if (OS_DEBUG != 0)
  static OS_U32 _SpuriousIrqCnt;

  _SpuriousIrqCnt++;
#endif
}

/*********************************************************************
*
*       _InitAIC()
*
*  Function description
*    Initialize interrupt controller by setting default vectors
*    and clearing all interrupts
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*/
static void _InitAIC(void) {
  unsigned int     i;
  OS_ISR_HANDLER** papISR;

  AIC_IDCR = 0xFFFFFFFFu;                    // Disable all interrupts
  AIC_ICCR = 0xFFFFFFFFu;                    // Clear all interrupts
  AIC_FFDR = 0xFFFFFFFF;                     // Reset fast forcingts
  AIC_SVR0 = (int)_DefaultFiqHandler;        // Dummy FIQ handler
  AIC_SPU  = (int) _SpuriousIrqHandler;      // Dummy spurious handler
  papISR   = (OS_ISR_HANDLER**)AIC_SVR_BASE_ADDR;
  for (i = 1u; i < NUM_INT_SOURCES; i++)  {  // Initially set all sources
    *(papISR + i) = &_DefaultIrqHandler;     // to dummy irq handler
  }
  for (i = 0; i < NUM_INT_PRIORITIES; i++) {
    AIC_EOICR = 0;                           // Reset interrupt controller
  }
#if OS_DEBUG                                 // For debugging activate AIC protected mode
  AIC_DCR |= 0x01u;                          // Enable AIC protected mode
#endif
}

/*********************************************************************
*
*       _InitFlash()
*
*  Function description
*    Initialize wait states and BUS interface for external Flash access
*    Needs to be called before main clock is switched to high speed
*    Depends on specific hardware and has to be implemented by user.
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*/
static void _InitFlash(void) {
}

/*********************************************************************
*
*       _InitClocks()
*
* Function description
*   Initialize main clock and PLLs
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*    (2) Settings used here are for AT91SAM99G45-EK eval board
*        and may have to be modified for other target board or frequency
*/
static void _InitClocks(OS_U32 PLLA_Val, OS_U32 MCKR_Val) {
  //
  // Initialize PLLs if target is not running from SDRAM. (Otherwise PLL is already initialized)
  //
  if ((OS_GetPC() & 0xF0000000) != DDRAM_BASE_ADDR) {
    //
    // Enable main clock OSC and wait until clock is stable
    //
    PMC_MOR = ((64/8) << 8)    // Startup time = 64 slow clocks
            | 0x01             // Main oscillator enable
            ;
    while ((PMC_SR & PMC_MOSCS) == 0) {
    }
    //
    // Initialize PLL A for Main clock
    //
    PMC_CKGR_PLLAR = PLLA_Val;
    //
    // Wait until PLLA locks
    //
    while ((PMC_SR & PMC_LOCKA) == 0) {
    }
    //
    // Wait until the master clock is stable
    //
    while ((PMC_SR & PMC_MCKRDY) == 0) {
    }
    //
    // Setup main clock register
    //
    PMC_MCKR = MCKR_Val;
    //
    // Wait until the master clock is stable
    //
    while ((PMC_SR & PMC_MCKRDY) == 0) {
    }
    //
    // Select PLL A clock as clock source
    //
    PMC_MCKR |= 0x02;
    //
    // Wait until the master clock is stable */
    //
    while((PMC_SR & PMC_MCKRDY) == 0) {
    }
  }
}

/*********************************************************************
*
*       _InitRTT()
*
*  Function description
*    Initialize Real time timer.
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*    (2) Not used by embOS, we disable interrupts here to avoid unhandled interrupts
*        May be modified by user if RTT is required for application.
*/
static void _InitRTT(void) {
  //
  // Disable Real-Time Timer interrupts
  //
  RTT_MR &= ~(RTT_RTTINCIEN | RTT_ALMIEN);
}

/*********************************************************************
*
*       _InitSDRAM()
*
*  Function description
*    Initialize SDRAM controller
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*    (2) Settings used here are for AT91SAM9263-EK eval board
*        and may have to be modified for other target board
*/
static void _InitSDRAM(void) {
  OS_U32 TempReg;
  //
  // Setup bus interface for DDRAM if not already initialized
  //
  if ((OS_GetPC() & 0xF0000000) != DDRAM_BASE_ADDR) {  // Target running in DDRAM ?
    //
    // Initialize SDRAM access and SDRAM, tbd
    //



    // 0xFFFFE600 DDR2C Base Address

    // Enable DDR2 clock x2 in PMC
    // AT91C_BASE_PMC, PMC_SCER, AT91C_PMC_DDR

    PMC_SCER |= (1u << 2);             // Enable clock for DDR2 controller

    // Configure the DDR controller

    // Disable anticipated read
    //WRITE(pDdrc, HDDRSDRC2_HS, (READ(pDdrc, HDDRSDRC2_HS) | AT91C_DDRC2_NO_ANT));

    TempReg = __readMemory32(0xFFFFE62C,"Memory");
    TempReg |= (1uL << 2);
    __writeMemory32(TempReg,0xFFFFE62C,"Memory");


    // -----------------------Step 1-------------------
    // Program the memory device type
    // ------------------------------------------------

    // HDDRSDRC2_MDR, AT91C_DDRC2_DBW_16_BITS | 16-bit DDR
    __writeMemory32(0x16,0xFFFFE620,"Memory");

    // -----------------------Step 2-------------------
    // 1. Program the features of DDR2-SDRAM device into
    //    the Configuration Register.
    // 2. Program the features of DDR2-SDRAM device into
    //    the Timing Register HDDRSDRC2_T0PR.
    // 3. Program the features of DDR2-SDRAM device into
    //    the Timing Register HDDRSDRC2_T1PR.
    // 4. Program the features of DDR2-SDRAM device into
    //    the Timing Register HDDRSDRC2_T2PR.
    // ------------------------------------------------

    // HDDRSDRC2_CR, AT91C_DDRC2_NC_DDR10_SDR9  |     // 10 column bits (1K)
    //          AT91C_DDRC2_NR_14          |          // 14 row bits    (8K)
    //          AT91C_DDRC2_CAS_3          |          // CAS Latency 3
    //          AT91C_DDRC2_DLL_RESET_DISABLED        // DLL not reset
    __writeMemory32(0x3D,0xFFFFE608,"Memory");

    // assume timings for 7.5ns min clock period
    // HDDRSDRC2_T0PR, AT91C_DDRC2_TRAS_6       |     //  6 * 7.5 = 45   ns
    //            AT91C_DDRC2_TRCD_2            |     //  2 * 7.5 = 15   ns
    //            AT91C_DDRC2_TWR_2             |     //  2 * 7.5 = 15   ns
    //            AT91C_DDRC2_TRC_8             |     //  8 * 7.5 = 60   ns
    //            AT91C_DDRC2_TRP_2             |     //  2 * 7.5 = 15   ns
    //            AT91C_DDRC2_TRRD_2            |     //  2 * 7.5 = 15  ns
    //            AT91C_DDRC2_TWTR_1            |     //  2 clock cycle
    //            AT91C_DDRC2_TMRD_2                  //  2 clock cycles
    __writeMemory32(0x21228226,0xFFFFE60C,"Memory");

    // pSDDRC->HDDRSDRC2_T1PR = 0x00000008;
    // HDDRSDRC2_T1PR, AT91C_DDRC2_TXP_2  |           //  2 * 7.5 = 15 ns
    //                 200 << 16          |           // 200 clock cycles, TXSRD: Exit self refresh delay to Read command
    //                 16 << 8            |           // 16 * 7.5 = 120 ns TXSNR: Exit self refresh delay to non read command
    //                 AT91C_DDRC2_TRFC_14 << 0       // 14 * 7.5 = 105 ns (must be 105 ns for 512M DDR)
    __writeMemory32(0x02c8100e,0xFFFFE610,"Memory");

    // HDDRSDRC2_T2PR, AT91C_DDRC2_TRTP_2   |         //  2 * 7.5 = 15 ns
    //                 AT91C_DDRC2_TRPA_0   |         //  0 * 7.5 = 0 ns
    //                 AT91C_DDRC2_TXARDS_7 |         //  7 clock cycles
    //                 AT91C_DDRC2_TXARD_2            //  2 clock cycles
    __writeMemory32(0x1072,0xFFFFE614,"Memory");

    // -----------------------Step 3-------------------
    // An NOP command is issued to the DDR2-SDRAM to
    // enable clock.
    // ------------------------------------------------

    __writeMemory32(0x1,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x70000000,"Memory");

    // A minimum pause of 200 us is provided to precede any signal toggle.
    __delay(1);

    // Now clocks which drive DDR2-SDRAM device are enabled

    // -----------------------Step 4-------------------
    // An NOP command is issued to the DDR2-SDRAM
    // ------------------------------------------------
    __writeMemory32(0x1,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x70000000,"Memory");
    // wait 400 ns min
    __delay(1);
    // Now CKE is driven high.

    // -----------------------Step 5-------------------
    // An all banks precharge command is issued to the
    // DDR2-SDRAM.
    // ------------------------------------------------

    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_PRCGALL_CMD
    __writeMemory32(0x2,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x70000000,"Memory");

    // wait 400 ns min
    __delay(1);

    // -----------------------Step 6-------------------
    // An Extended Mode Register set (EMRS2) cycle is
    // issued to chose between commercial or high
    // temperature operations
    // ------------------------------------------------

    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_EXT_LMR_CMD
    __writeMemory32(0x5,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x74000000,"Memory");  // The write address must be chosen so that BA[1] is set to 1 and BA[0] are set to 0.
    // wait 2 cycles min
    __delay(1);

    // -----------------------Step 7-------------------
    // An Extended Mode Register set (EMRS3) cycle is
    // issued to set all registers to 0.
    // ------------------------------------------------
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_EXT_LMR_CMD
    __writeMemory32(0x5,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x76000000,"Memory");  // The write address must be chosen so that BA[1] is set to 1 and BA[0] are set to 1.
    // wait 2 cycles min
    __delay(1);

    // -----------------------Step 8-------------------
    // An Extended Mode Register set (EMRS1) cycle is
    // issued to enable DLL.
    // ------------------------------------------------
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_EXT_LMR_CMD
    __writeMemory32(0x5,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x72000000,"Memory"); // The write address must be chosen so that BA[1] and BA[0] are set to 0.
    // An additional 200 cycles of clock are required for locking DLL
    __delay(1);

    // -----------------------Step 9-------------------
    // Program DLL field into the Configuration Register.
    // -------------------------------------------------

    // HDDRSDRC2_CR, cr | AT91C_DDRC2_DLL_RESET_ENABLED
    TempReg = __readMemory32(0xFFFFE608,"Memory");
    TempReg |= 0xBD;
    __writeMemory32(TempReg,0xFFFFE608,"Memory");

    // -----------------------Step 10-------------------
    // A Mode Register set (MRS) cycle is issued to reset
    // DLL.
    // -------------------------------------------------
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_LMR_CMD
    __writeMemory32(0x3,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x70000000,"Memory"); // The write address must be chosen so that BA[1:0] bits are set to 0.
    // wait 2 cycles min
    __delay(1);

    // -----------------------Step 11-------------------
    // An all banks precharge command is issued to the
    // DDR2-SDRAM.
    // -------------------------------------------------
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_PRCGALL_CMD
    __writeMemory32(0x2,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x70000000,"Memory"); // Perform a write access to any DDR2-SDRAM address to acknowledge this command.
    // wait 2 cycles min
    __delay(1);

    // -----------------------Step 12-------------------
    // Two auto-refresh (CBR) cycles are provided.
    // Program the auto refresh command (CBR) into the
    // Mode Register.
    // -------------------------------------------------
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_RFSH_CMD
    __writeMemory32(0x4,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x70000000,"Memory"); // Performs a write access to any DDR2-SDRAM location twice to acknowledge these commands.
    // wait 2 cycles min
    __delay(1);
    // Set 2nd CBR
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_RFSH_CMD
    __writeMemory32(0x4,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x70000000,"Memory"); // Performs a write access to any DDR2-SDRAM location twice to acknowledge these commands.
    // wait 2 cycles min
    __delay(1);

    // -----------------------Step 13-------------------
    // Program DLL field into the Configuration Register
    // to low(Disable DLL reset).
    // -------------------------------------------------
    // HDDRSDRC2_CR, cr & (~AT91C_DDRC2_DLL_RESET_ENABLED)
    TempReg = __readMemory32(0xFFFFE608,"Memory");
    TempReg &= 0xFFFFFF7F;
    __writeMemory32(TempReg,0xFFFFE608,"Memory");
    // wait 2 cycles min
    __delay(1);

    // -----------------------Step 14-------------------
    // A Mode Register set (MRS) cycle is issued to
    // program the parameters of the DDR2-SDRAM devices
    // -------------------------------------------------
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_LMR_CMD
    __writeMemory32(0x3,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x70000000,"Memory"); // The write address must be chosen so that BA[1:0] are set to 0.
    // wait 2 cycles min
    __delay(1);

    // -----------------------Step 15-------------------
    // Program OCD field into the Configuration Register
    // to high (OCD calibration default)
    // -------------------------------------------------
    TempReg = __readMemory32(0xFFFFE608,"Memory");
    TempReg |= (0x07 << 12);
    __writeMemory32(TempReg,0xFFFFE608,"Memory");
    __delay(1);

    // -----------------------Step 16-------------------
    // An Extended Mode Register set (EMRS1) cycle is
    // issued to OCD default value.
    // -------------------------------------------------
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_EXT_LMR_CMD
    __writeMemory32(0x5,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x72000000,"Memory"); // The write address must be chosen so that BA[1] is set to 0 and BA[0] is set to 1.
    // wait 2 cycles min
    __delay(1);

    // -----------------------Step 17-------------------
    // Program OCD field into the Configuration Register
    // to low (OCD calibration mode exit).
    // -------------------------------------------------
    TempReg = __readMemory32(0xFFFFE608,"Memory");
    TempReg &= 0xFFFF8FFF;
    __writeMemory32(TempReg,0xFFFFE608,"Memory");
    __delay(1);

    // -----------------------Step 18-------------------
    // An Extended Mode Register set (EMRS1) cycle is
    // issued to enable OCD exit.
    // -------------------------------------------------
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_EXT_LMR_CMD
    __writeMemory32(0x5,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x76000000,"Memory"); // The write address must be chosen so that BA[1] is set to 1 and BA[0] is set to 1.
    // wait 2 cycles min
    __delay(1);


    // -----------------------Step 19,20-------------------
    // A mode Normal command is provided. Program the
    // Normal mode into Mode Register.
    // -------------------------------------------------
    // HDDRSDRC2_MR, AT91C_DDRC2_MODE_NORMAL_CMD
    __writeMemory32(0x0,0xFFFFE600,"Memory");
    __writeMemory32(0x0,0x70000000,"Memory"); // Perform a write access to any DDR2-SDRAM address.
    // wait 2 cycles min
    __delay(1);

    // -----------------------Step 21-------------------
    // Write the refresh rate into the count field in the
    // Refresh Timer register. The DDR2-SDRAM device requires a
    // refresh every 15.625 ¦Ìs or 7.81 ¦Ìs. With a 100
    // MHz frequency, the refresh timer count register must to
    // be set with (15.625 /100 MHz) = 1562 i.e. 0x061A or
    // (7.81 /100MHz) = 781 i.e. 0x030d
    // -------------------------------------------------
    // HDDRSDRC2_RTR, 0x00000411
    __writeMemory32(0x00000411,0xFFFFE604,"Memory");

    // Read optimization" shall be un-selected on this revision.
    __writeMemory32(0x04,0xFFFFE62C,"Memory");

    // OK now we are ready to work on the DDRSDR

    // wait for end of calibration
    __delay(1);

  }
}

/*********************************************************************
*
*       _InitSMC()
*
*  Function description
*    Initialize static memory controller
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*    (2) Settings used here are for AT91SAM9G45-EK eval board
*        and may have to be modified for other target boards
*/
static void _InitSMC(void) {
}


/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return (PIT_PIIR & 0x000FFFFFu);
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return (PIT_PIIR & 0xFFF00000u);
}

/*********************************************************************
*
*       _OS_ISR_Tick()
*
*  Function description
*    This is the hardware timer exception handler.
*/
static void _OS_ISR_Tick(void) {
  volatile OS_U32 Dummy;

  if (PIT_SR & (1u << 0)) {  // Timer interrupt pending?
    Dummy = PIT_PIVR;        // Reset interrupt pending condition
    OS_USEPARA(Dummy);
    OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
    JLINKDCC_Process();
#endif
  }
  //
  // Call to following handlers may be removed if not used by application
  //
  _HandlePmcIrq();
  _HandleRttIrq();
  _HandleRstcIrq();
  _HandleWdtIrq();
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  OS_INT_IncDI();
  //
  // Ensure caches are synchronized
  //
  OS_ARM_CACHE_Sync();
  //
  // Initialize timer for embOS, assuming PLL is already initialized
  //
  PIT_MR = ((OS_TIMER_RELOAD & 0x000FFFFFu) | (1u << 25) | (1u << 24));
  //
  // Enable timer interrupt in interrupt controller
  //
  OS_ARM_InstallISRHandler(SYSTEM_IRQ_ID, _OS_ISR_Tick);
  OS_ARM_EnableISR(SYSTEM_IRQ_ID);
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       embOS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;

  pISR = (OS_ISR_HANDLER*)AIC_IVR;   // Read interrupt vector to release NIRQ to CPU core
#if (OS_DEBUG != 0)
  AIC_IVR = (OS_U32)pISR;            // Write back any value to IVR register to allow interrupt stacking in protected mode
#endif
  OS_INT_EnterNestable();            // Inform OS that interrupt handler is running
  pISR();                            // Call interrupt service routine
  OS_INT_Disable();                  // Disable interrupts and unlock
  AIC_EOICR = 0;                     // interrupt controller =>  Restore previous priority
  OS_INT_LeaveNestable();            // Leave interrupt, perform task switch if required
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Used to install a specific interrupt vector when ARM
*    CPUs with vectored interrupt controller are used.
*
*  Parameters
*    ISRIndex:    Index of the interrupt source, normally the interrupt vector number.
*    pISRHandler: Address of the interrupt handler function.
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler(int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;
  OS_ISR_HANDLER** papISR;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  papISR               = (OS_ISR_HANDLER**)AIC_SVR_BASE_ADDR;
  pOldHandler          = *(papISR + ISRIndex);
  *(papISR + ISRIndex) = pISRHandler;
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Used to enable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be enabled.
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  AIC_IECR = (1u << ISRIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Used to disable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be disabled.
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  AIC_IDCR = (1u << ISRIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Function description
*    Used to set or modify the priority of a specific interrupt
*    source by programming the interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be modified.
*    Prio:     The priority which should be set for the specific interrupt.
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  OS_U32* pPrio;
  int     OldPrio;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  pPrio = (OS_U32*)AIC_SMR_BASE_ADDR;
  OldPrio = pPrio[ISRIndex];
  pPrio[ISRIndex] = (OldPrio & ~INT_PRIORITY_MASK) | (Prio & INT_PRIORITY_MASK);
  OS_INT_EnableConditional();
  return OldPrio & INT_PRIORITY_MASK;
}

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
OS_INTERWORK int __low_level_init(void) {
  WDT_MR = (1u << 15);                   // Initially disable watchdog
  //
  // Setup waitstates for external BUS interface before switching to high speed
  //
  _InitFlash();
  //
  // Initialize main clock oscillator and PLLs
  //
  _InitClocks(PLLAR_VALUE, MCKR_VALUE);
  //
  // Initialize real time timer
  //
  _InitRTT();
  //
  // Initialize SDRAM
  //
  _InitSDRAM();
  //
  //  Initialize static memory controller
  //
  _InitSMC();
  //
  // Init MMU and caches. This defines the virtual memory map, which is used during execution.
  // Memory mapping should be complete, meaning 4096 entries.
  // Code below fills in ascending VAddr order
  //
  //                                             Mode                    VAddr  PAddr  Size[MB]
  if ((OS_GetPC() & 0xF0000000) == DDRAM_BASE_ADDR) {  // When target runs in DDRAM, remap DDRAM to 0x00, else remap iRAM
    OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B, 0x000, 0x700, 0x001);  // SDRAM, remapped
  } else {
    OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B, 0x000, 0x003, 0x001);  // Internal SRAM, remapped
  }
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x001, 0x001, 0x001);  // ITCM, uncached
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x002, 0x002, 0x001);  // DTCM, uncached
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x003, 0x003, 0x001);  // Internal SRAM, original address, uncached, unbuffered (allows DMA access)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,   0x004, 0x004, 0x001);  // Internal ROM
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x005, 0x005, 0x001);  // LCD controller
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x006, 0x006, 0x001);  // UDPHS (DMA)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x007, 0x007, 0x001);  // UHP OHCI
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x008, 0x008, 0x001);  // UHP EHCI
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_MMU_UNMAPPED,    0x009, 0x009, 0x001);  // Reserved area (might be used to mirror SRAM)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_MMU_UNMAPPED,    0x00A, 0x00A, 0x006);  // Undefined (Abort)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,   0x100, 0x100, 0x100);  // EBI CS0 area
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,   0x200, 0x200, 0x100);  // EBI CS1 area (SDRAM)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,   0x300, 0x300, 0x100);  // EBI CS2 area
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x400, 0x400, 0x100);  // EBI CS3 area (NAND-FLASH)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x500, 0x500, 0x100);  // EBI CS4 area (Compact Flash Slot0)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x600, 0x600, 0x100);  // EBI CS4 area (Compact Flash Slot1)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,   0x700, 0x700, 0x080);  // DDRAM, original address, 128MB
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x780, 0x700, 0x080);  // DDRAM, mirrored, 128MB
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x800, 0x800, 0x6FF);  // Undefined (Abort)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0xfff, 0xfff, 0x001);  // SFR area
  OS_ARM_MMU_Enable      (&_TranslationTable[0]);
  OS_ARM_ICACHE_Enable();
  OS_ARM_DCACHE_Enable();
  //
  // Initialize interrupt controller
  //
  _InitAIC();
  //
  // Initialize reset control
  //
  RSTC_MR = ((0xA5u << 24) | RSTC_URSTEN);  // write KEY and URSTEN to allow USER RESET
  return 1;                                 // Always return 1 to enable segment initialization!
}

/*************************** End of file ****************************/
