/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : BSP for Atmel SAMA5D27 SOM module
*/

#include "BSP.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/
#define LED0_BIT           (10u)  // Red, PA10, high active
#define LED1_BIT           (1u)   // Green, PB1, high active
#define LED2_BIT           (31u)  // Blue, PA31, high active

#define PIO_BASE_ADDR      (0xFC038000u)

#define PIO_WPMR1          (*(volatile unsigned int*) (PIO_BASE_ADDR + 0x5A0u)) // Output Data Status Register
#define PIO_WPKEY          (0x50494Fu)

#define PIOA_BASE_ADDR     (PIO_BASE_ADDR)
#define PIOA_MSKR1         (*(volatile unsigned int*) (PIOA_BASE_ADDR + 0x00u))  // PIO Mask Register
#define PIOA_CFGR1         (*(volatile unsigned int*) (PIOA_BASE_ADDR + 0x04u))  // PIO Configuration Register
#define PIOA_SODR1         (*(volatile unsigned int*) (PIOA_BASE_ADDR + 0x10u))  // Set Output Data Register
#define PIOA_CODR1         (*(volatile unsigned int*) (PIOA_BASE_ADDR + 0x14u))  // Clear Output Data Register
#define PIOA_ODSR1         (*(volatile unsigned int*) (PIOA_BASE_ADDR + 0x18u))  // Output Data Status Register

#define PIOB_BASE_ADDR     (PIO_BASE_ADDR + 0x40u)
#define PIOB_MSKR1         (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x00u))  // PIO Mask Register
#define PIOB_CFGR1         (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x04u))  // PIO Configuration Register
#define PIOB_SODR1         (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x10u))  // Set Output Data Register
#define PIOB_CODR1         (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x14u))  // Clear Output Data Register
#define PIOB_ODSR1         (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x18u))  // Output Data Status Register

#define PIO_CFGRX_DIR_BIT  (8u)

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_Init()
*/
void BSP_Init(void) {
  unsigned int Protection;

  //
  // Unlock access to PIO configuration
  //
  Protection = (PIO_WPMR1) & (1u << 0);
  PIO_WPMR1  = ((PIO_WPKEY) << 8);
  //
  // Enable LEDs, set pins to output and disable initially
  //
  PIOA_MSKR1 = (1u << LED0_BIT) | (1u << LED2_BIT);
  PIOB_MSKR1 = (1u << LED1_BIT);
  PIOA_CFGR1 = (1u << PIO_CFGRX_DIR_BIT);
  PIOB_CFGR1 = (1u << PIO_CFGRX_DIR_BIT);
  PIOA_CODR1 = (1u << LED0_BIT) | (1u << LED2_BIT);
  PIOB_CODR1 = (1u << LED1_BIT);
  //
  // Finally restore protection state
  //
  if (Protection != 0) {
    PIO_WPMR1  = ((PIO_WPKEY) << 8) | (1u << 0);
  }
}

/*********************************************************************
*
*       BSP_SetLED()
*/
void BSP_SetLED(int Index) {
  unsigned int Protection;

  //
  // Unlock access to PIO configuration
  //
  Protection = (PIO_WPMR1) & (1u << 0);
  PIO_WPMR1  = ((PIO_WPKEY) << 8);
  //
  // Enable led port pin: Set to output & enable open drain
  //
  switch (Index) {
  case 0:
    PIOB_SODR1 = (1u << LED1_BIT);
    break;
  case 1:
    PIOA_SODR1 = (1u << LED2_BIT);
    break;
  case 2:
    PIOA_SODR1 = (1u << LED0_BIT);
    break;
  default:
    break;
  }
  //
  // Finally restore protection state
  //
  if (Protection != 0) {
    PIO_WPMR1  = ((PIO_WPKEY) << 8) | (1u << 0);
  }  
}

/*********************************************************************
*
*       BSP_ClrLED()
*/
void BSP_ClrLED(int Index) {
  unsigned int Protection;

  //
  // Unlock access to PIO configuration
  //
  Protection = (PIO_WPMR1) & (1u << 0);
  PIO_WPMR1  = ((PIO_WPKEY) << 8);

  switch (Index) {
  case 0:
    PIOB_CODR1 = (1u << LED1_BIT);
    break;
  case 1:
    PIOA_CODR1 = (1u << LED2_BIT);
    break;
  case 2:
    PIOA_CODR1 = (1u << LED0_BIT);
    break;
  default:
    break;
  }
  //
  // Finally restore protection state
  //
  if (Protection != 0) {
    PIO_WPMR1  = ((PIO_WPKEY) << 8) | (1u << 0);
  }    
}

/*********************************************************************
*
*       BSP_ToggleLED()
*/
void BSP_ToggleLED(int Index) {
  unsigned int Protection;

  //
  // Unlock access to PIO configuration
  //
  Protection = (PIO_WPMR1) & (1u << 0);
  PIO_WPMR1  = ((PIO_WPKEY) << 8);

  switch (Index) {
  case 0:
    if ((PIOB_ODSR1 & (1u << LED1_BIT)) == 0) {
      PIOB_SODR1 = (1u << LED1_BIT);
    } else {
      PIOB_CODR1 = (1u << LED1_BIT);
    }  
    break;
  case 1:
    if ((PIOA_ODSR1 & (1u << LED2_BIT)) == 0) {
      PIOA_SODR1 = (1u << LED2_BIT);
    } else {
      PIOA_CODR1 = (1u << LED2_BIT);
    }  
    break;
  case 2:
    if ((PIOA_ODSR1 & (1u << LED0_BIT)) == 0) {
      PIOA_SODR1 = (1u << LED0_BIT);
    } else {
      PIOA_CODR1 = (1u << LED0_BIT);
    }  
    break;
  default:
    break;
  }
  //
  // Finally restore protection state
  //
  if (Protection != 0) {
    PIO_WPMR1  = ((PIO_WPKEY) << 8) | (1u << 0);
  }  
}

/*************************** End of file ****************************/
