/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
File    : BSP.c
Purpose : BSP for Atmel SAMA5D2 Xplained Ultra eval board
*/

#include "BSP.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/
#define NUM_LEDS                      (3)

#define LED_RED_PIN                   (6u)
#define LED_GREEN_PIN                 (5u)
#define LED_BLUE_PIN                  (0u)

#define PIOB_BASE_ADDR                (0xFC038040u)
#define PIO_MSKR1                     (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x00u))  // PIO Mask Register
#define PIO_CFGR1                     (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x04u))  // PIO Configuration Register
#define PIO_SODR1                     (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x10u))  // Set Output Data Register
#define PIO_CODR1                     (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x14u))  // Clear Output Data Register
#define PIO_ODSR1                     (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x18u))  // Output Data Status Register
#define PIO_WPMR1                     (*(volatile unsigned int*) (PIOB_BASE_ADDR + 0x5A0u)) // Output Data Status Register
#define PIO_WPKEY                     (0x50494Fu)

#define PIO_CFGRX_DIR_BIT             (8u)
#define PIO_CFGR_OPD_BIT              (14u)

/*********************************************************************
*
*       Typedefs
*
**********************************************************************
*/

typedef void ETH_ISR_HANDLER(void);

typedef struct _PORT_INFO {
  int PortPin;
} _PORT_INFO;

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static int              _LEDState;

static const _PORT_INFO _aLEDInfo[] = {
  {LED_RED_PIN}, 
  {LED_GREEN_PIN},
  {LED_BLUE_PIN}
};

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_Init()
*/
void BSP_Init(void) {
  unsigned int Protection;

  //
  // Unlock access to PIO configuration
  //
  Protection = (PIO_WPMR1) & (1u << 0);
  PIO_WPMR1  = ((PIO_WPKEY) << 8);
  //
  // Switch off LEDs initially by setting led port pins to input / high resistance
  //
  PIO_MSKR1  = (1u << LED_RED_PIN) | (1u << LED_GREEN_PIN) | (1u << LED_BLUE_PIN);
  PIO_CFGR1  = 0;
  //
  // Finally restore protection state
  //
  if (Protection != 0) {
    PIO_WPMR1  = ((PIO_WPKEY) << 8) | (1u << 0);
  }
}

/*********************************************************************
*
*       BSP_SetLED()
*/
void BSP_SetLED(int Index) {
  unsigned int Protection;

  if (Index < NUM_LEDS) {
    const _PORT_INFO* pLedInfo = &_aLEDInfo[Index];
    //
    // Unlock access to PIO configuration
    //
    Protection = (PIO_WPMR1) & (1u << 0);
    PIO_WPMR1  = ((PIO_WPKEY) << 8);
    //
    // Enable led port pin: Set to output & enable open drain
    //
    PIO_MSKR1  = (1u << pLedInfo->PortPin);
    PIO_CFGR1  = (1u << PIO_CFGRX_DIR_BIT) | (1u << PIO_CFGR_OPD_BIT);
    _LEDState |= (1 << Index);
    //
    // Finally restore protection state
    //
    if (Protection != 0) {
      PIO_WPMR1  = ((PIO_WPKEY) << 8) | (1u << 0);
    }
  }
}

/*********************************************************************
*
*       BSP_ClrLED()
*/
void BSP_ClrLED(int Index) {
  unsigned int Protection;

  if (Index < NUM_LEDS) {
    const _PORT_INFO* pLedInfo = &_aLEDInfo[Index];
    //
    // Unlock access to PIO configuration
    //
    Protection = (PIO_WPMR1) & (1u << 0);
    PIO_WPMR1  = ((PIO_WPKEY) << 8);
    //
    // Disable led: Set to input / high resistance
    //
    PIO_MSKR1  =  (1u << pLedInfo->PortPin);
    PIO_CFGR1  =  0;
    _LEDState &= ~(1 << Index);
    //
    // Finally restore protection state
    //
    if (Protection != 0) {
      PIO_WPMR1  = ((PIO_WPKEY) << 8) | (1u << 0);
    }
  }
}

/*********************************************************************
*
*       BSP_ToggleLED()
*/
void BSP_ToggleLED(int Index) {
  if (_LEDState & (1 << Index)) {
    BSP_ClrLED(Index);
  } else {
    BSP_SetLED(Index);
  }
}

/*************************** End of file ****************************/
