/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : UART implementation

Additional information:

  Device : AT91SAMA5D44
  Board  : AT91SAMA5D4x MB

  Unit | UART   | Board connector
  ===============================
  0    | USART3 | DBGU (J24)

*/

#include "BSP_UART.h"
#include "RTOS.h"      // For OS_ARM_* and OS_INT_* functions.

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/

#define SYSTEM_FREQ           (498000000uL)
#define USART_FREQ            (SYSTEM_FREQ / 6uL)

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/
//
// USART3
//
#define USART3_BASE_ADDR            (0xFC00C000uL)
#define US3_CR                      (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x000uL))
#define US3_MR                      (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x004uL))
#define US3_IER                     (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x008uL))
#define US3_IDR                     (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x00CuL))
#define US3_IMR                     (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x010uL))
#define US3_CSR                     (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x014uL))
#define US3_RHR                     (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x018uL))
#define US3_THR                     (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x01CuL))
#define US3_BRGR                    (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x020uL))
#define US3_WPMR                    (*(volatile unsigned long*)(USART3_BASE_ADDR + 0x0E4uL))  // USART Write Protection Mode Register

#define USART_RXRDY                 (1uL << 0)  // Rx status flag
#define USART_TXRDY                 (1uL << 1)  // Tx RDY Status flag
#define USART_RX_ERROR_FLAGS        (0xE0uL)    // Parity, framing, overrun error
#define USART_BAUDDIVIDE(Baudrate)  ((USART_FREQ + ((Baudrate) * 8uL)) / ((Baudrate) * 16uL) - 1)

#define USART_IRQ_PRIO              (0x00uL)    // Lowest priority for UART interrupts

#define USART3_ID                   (30uL)

#define USART3                      (0uL)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static BSP_UART_TX_CB* _pfWriteCB;
static BSP_UART_RX_CB* _pfReadCB;

/*********************************************************************
*
*       Prototypes
*
*  Declare ISR handler here to avoid "no prototype" warning.
*  They are not declared in any CMSIS header.
*
**********************************************************************
*/

#if defined(__cplusplus)
  extern "C" {                // Make sure we have C-declarations in C++ programs.
#endif

static void USART3_RX_TX_IRQHandler(void);

#if defined(__cplusplus)
}                             // Make sure we have C-declarations in C++ programs.
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _SetBaudrate()
*
*  Function description
*    Configures the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
static void _SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  BSP_UART_USE_PARA(Unit);
  BSP_UART_USE_PARA(Baudrate);
}

/*********************************************************************
*
*       Global functions, IRQ handler
*
**********************************************************************
*/

/*********************************************************************
*
*       USART3_RX_TX_IRQHandler()
*
*  Function description
*    UART Rx & Tx interrupt handler.
*/
static void USART3_RX_TX_IRQHandler(void) {
  unsigned int Status;
  unsigned int Unit;
  unsigned int Data;

  OS_INT_EnterNestable();
  Unit   = USART3;
  Status = US3_CSR;                         // Examine status register
  do {
    if (Status & USART_RXRDY) {             // Data received?
      Data = US3_RHR;
      if (Status & USART_RX_ERROR_FLAGS) {  // Any error ?
        US3_CR = (1uL << 8);                // RSTSTA: Reset Status Bits PARE, FRAME, OVRE and RXBRK
      } else {
        if (_pfReadCB != 0) {
          _pfReadCB(Unit, Data);            // Forward the received byte to the callback.
        }
      }
    }
    if (Status & US3_IMR & USART_TXRDY) {   // Check Tx status => Send next character
      if ((_pfWriteCB != 0) && (_pfWriteCB(Unit) != 0)) {
        US3_IDR = USART_TXRDY;              // Disable further tx interrupts
      }
    }
    Status = US3_CSR;                       // Examine status register
  } while (Status & US3_IMR & (USART_TXRDY | USART_RXRDY));
  OS_INT_LeaveNestable();
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_UART_Init()
*
*  Function description
*    Initializes the UART peripheral.
*
*  Parameters
*    Unit       : Unit number (typically zero-based).
*    Baudrate   : Baud rate to configure [Hz].
*    NumDataBits: Number of data bits to use.
*    Parity     : One of the following values:
*                   * BSP_UART_PARITY_NONE
*                   * BSP_UART_PARITY_ODD
*                   * BSP_UART_PARITY_EVEN
*    NumStopBits: Number of stop bits to use.
*/
void BSP_UART_Init(unsigned int Unit, unsigned long Baudrate, unsigned char NumDataBits, unsigned char Parity, unsigned char NumStopBits) {
  BSP_UART_USE_PARA(NumDataBits);
  BSP_UART_USE_PARA(Parity);
  BSP_UART_USE_PARA(NumStopBits);
  OS_INT_IncDI();
  switch (Unit) {
  case USART3:
    US3_IDR  = 0xFFFFFFFFuL;            // Initially disable all USART interrupts
    US3_CR   = (1uL <<  2)              // RSTRX:  Reset Receiver: 1 = The receiver logic is reset.
             | (1uL <<  3);             // RSTTX:  Reset Transmitter: 1 = The transmitter logic is reset.
    US3_CR   = (0uL <<  2)              // RSTRX:  Reset Receiver: 1 = The receiver logic is reset.
             | (0uL <<  3)              // RSTTX:  Reset Transmitter: 1 = The transmitter logic is reset.
             | (1uL <<  4)              // RXEN:   Receiver Enable: 1 = The receiver is enabled if RXDIS is 0.
             | (0uL <<  5)              // RXDIS:  Receiver Disable: 0 = Noeffect.
             | (1uL <<  6)              // TXEN:   Transmitter Enable: 1 = The transmitter is enabled if TXDIS is 0.
             | (0uL <<  7)              // TXDIS:  Transmitter Disable: 0 = Noeffect.
             | (1uL <<  8)              // RSTSTA: Reset Status Bits: 1 = Resets the status bits PARE, FRAME, OVRE and RXBRK in the US_CSR.
             | (0uL <<  9)              // STTBRK: Start Break: 0 = Noeffect.
             | (0uL << 10)              // STPBRK: Stop Break: 0 = Noeffect.
             | (0uL << 11)              // STTTO:  Start Time-out: 0 = Noeffect.
             | (0uL << 12);             // SENDA:  Send Address: 0 = Noeffect.
    US3_WPMR = (0x555341uL << 8);       // Disable Write Protection
    US3_BRGR = USART_BAUDDIVIDE(Baudrate);
    US3_MR   = (0uL <<  4)              // USCLKS: Clock Selection: 0 = MCK
             | (3uL <<  6)              // CHRL:   Character Length: 3 = Eight bits
             | (0uL <<  8)              // SYNC:   Synchronous Mode Select: 0 = USART operates in Asynchronous Mode.
             | (4uL <<  9)              // PAR:    Parity Type: 0x4 = No parity
             | (0uL << 12)              // NBSTOP: Number of Stop Bits: 0 = 1 stop bit
             | (0uL << 14)              // CHMODE: Channel Mode: 0 = Normal mode
             | (0uL << 17)              // MODE9:  9-bit Character Length: 0 = CHRL defines character length.
             | (0uL << 18);             // CKLO:   Clock Output Select: 0 = The USART does not drive the SCK pin.
    US3_WPMR = (0x555341uL << 8)        // Enable Write Protection
             | (1uL        << 0);
    US3_IER  = (1uL <<  0);             // Enable Rx Interrupt
    //
    // Setup interrupt controller for USART
    //
    OS_ARM_InstallISRHandler(USART3_ID, &USART3_RX_TX_IRQHandler);  // OS UART interrupt handler vector
    OS_ARM_EnableISR(USART3_ID);                                    // Enable OS usart interrupts
    break;
  }
  OS_INT_DecRI();
}

/*********************************************************************
*
*       BSP_UART_DeInit()
*
*  Function description
*    De-initializes the UART peripheral.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*/
void BSP_UART_DeInit(unsigned int Unit) {
  BSP_UART_USE_PARA(Unit);
}

/*********************************************************************
*
*       BSP_UART_SetBaudrate()
*
*  Function description
*    Configures/changes the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
void BSP_UART_SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  _SetBaudrate(Unit, Baudrate);
}

/*********************************************************************
*
*       BSP_UART_SetReadCallback()
*
*  Function description
*    Sets the callback to execute upon an Rx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetReadCallback(unsigned Unit, BSP_UART_RX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfReadCB = pf;
}

/*********************************************************************
*
*       BSP_UART_SetWriteCallback()
*
*  Function description
*    Sets the callback to execute upon a Tx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetWriteCallback(unsigned int Unit, BSP_UART_TX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfWriteCB = pf;
}

/*********************************************************************
*
*       BSP_UART_Write1()
*
*  Function description
*    Sends one byte via UART.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    Data: (First) data byte to send.
*
*  Additional information
*    The first byte of a transfer is typically sent from application
*    context. Further bytes of the transfer are then sent from the
*    Tx interrupt handler by also calling this function from interrupt
*    context.
*/
void BSP_UART_Write1(unsigned int Unit, unsigned char Data) {
  switch (Unit) {
  case USART3:
    while ((US3_CSR & USART_TXRDY) == 0) {
      ;  // Wait until THR becomes available
    }
    US3_THR = Data;
    US3_IER = USART_TXRDY;  // Enable Tx interrupt
    break;
  }
}

/*************************** End of file ****************************/
