/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/********************************************************************
*
*       Configuration
*/
#ifndef   OS_TIMER_IS_PIT
  #define OS_TIMER_IS_PIT  (0)
#endif

#ifndef   OS_ALLOW_NESTED_FIQ
  #define OS_ALLOW_NESTED_FIQ  (1)
#endif

#ifndef   OS_HANDLE_UNSECURE_INTS_ONLY
  #define OS_HANDLE_UNSECURE_INTS_ONLY  (0)  // If set, FIQ (SAIC) is not handled here, this is default
#endif

/*********************************************************************
*
*       Clock settings
*/
#define OS_SYSTEM_FREQ     (498000000u)

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ      (OS_SYSTEM_FREQ / 6 / 2)
#define OS_TICK_FREQ       (1000u)
#define OS_INT_FREQ        (OS_TICK_FREQ)
#define OS_TIMER_RELOAD    ((OS_TIMER_FREQ / OS_INT_FREQ) - 1)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/********************************************************************
*
*       External memory settings
*/
#define   DDR_RAM_TYPE (1)
#define   DDR_CS_ADDR  (0x20000000u)

#ifdef __ICCARM__
  #define OS_ARM_FUNC  __arm
  #define ATTRIBUTE_NAKED_NO_INLINE
#endif

#ifdef __GNUC__
  #define OS_ARM_FUNC
  #define ATTRIBUTE_NAKED_NO_INLINE __attribute__ ((naked, noinline))
#endif

#ifdef __CC_ARM
  #define OS_ARM_FUNC
  #define OS_INTERWORK
  #define ATTRIBUTE_NAKED_NO_INLINE __attribute__ ((noinline))
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
//
// Clock configuration
//
#define PMC_BASE_ADDR                 (0xF0018000u)
#define PMC_PCER0                     (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x10u))   // PMC Peripheral Clock Enable Register 0
#define PMC_PCDR0                     (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x14u))   // PMC Peripheral Clock Enable Register 0
#define PMC_CKGR_MOR                  (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x20u))   // PMC Main oscillator register
#define PMC_CKGR_MCFR                 (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x24u))   // PMC Main oscillator register
#define PMC_CKGR_PLLAR                (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x28u))   // PMC Clock generator PLLA Register
#define PMC_MCKR                      (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x30u))   // PMC Master clock register
#define PMC_SR                        (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x68u))   // PMC Status register
#define PMC_PLLICPR                   (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x80u))   // PMC PLL Charge Pump Current register
#define PMC_PCER1                     (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x100u))  // PMC Peripheral Clock Enable Register 1
#define PMC_PCDR1                     (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x104u))  // PMC Peripheral Clock Enable Register 1
#define PMC_WPMR                      (*(volatile unsigned int*)(PMC_BASE_ADDR + 0x0E4u))  // PMC Write Protect Mode Register

#define PMC_WPKEY                     (0x504D43u)
#define PCLOCK_ENABLE                 (1u)
#define PCLOCK_DISABLE                (0u)
#define PLL_MULA ((OS_SYSTEM_FREQ * 2 / 12000000u) - 1)

//
// Watchdog register
//
#define WDT_BASE_ADDR                 (0xFC068640u)
#define WDT_MR                        (*(volatile unsigned int*)(WDT_BASE_ADDR + 0x04u))   // WDT Mode Register
#define WDDIS                         (15u)
//
// Advanced interrupt controller register (non secure, IRQ)
//
#define AIC_BASE_ADDR                 (0xFC06E000u)
#define AIC_SSR                       (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x00u))  // Source Select Register
#define AIC_SMR                       (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x04u))  // Source Mode Register
#define AIC_SVR                       (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x08u))  // Source Vector Register
#define AIC_IVR                       (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x10u))  // Interrupt Vector Register
#define AIC_IPR0                      (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x20u))  // Interrupt Pending Register 0
#define AIC_IPR1                      (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x24u))  // Interrupt Pending Register 1
#define AIC_IPR2                      (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x28u))  // Interrupt Pending Register 2
#define AIC_IPR3                      (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x2Cu))  // Interrupt Pending Register 3
#define AIC_EOICR                     (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x38u))  // End of Interrupt Command Register
#define AIC_SPU                       (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x3Cu))  // Spurious Interrupt Vector Register
#define AIC_IECR                      (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x40u))  // Interrupt Enable Command Register
#define AIC_IDCR                      (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x44u))  // Interrupt Disable Command Register
#define AIC_ICCR                      (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x48u))  // Interrupt Clear Command Register
#define AIC_DCR                       (*(volatile unsigned int*) (AIC_BASE_ADDR + 0x6Cu))  // Debug Control Register
#define AIC_WPMR                      (*(volatile unsigned int*) (AIC_BASE_ADDR + 0xE4u))  // AIC Write Protect Mode Register

#define NUM_INT_SOURCES               (68u)
#define INT_PRIORITY_MASK             (0x07u)
#define AIC_WPKEY                     (0x414943u)

//
// Secure Advanced interrupt controller register, FIQ
//
#define SAIC_BASE_ADDR                (0xFC068400u)
#define SAIC_SSR                      (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x00u))  // Source Select Register
#define SAIC_SMR                      (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x04u))  // Source Mode Register
#define SAIC_SVR                      (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x08u))  // Source Vector Register
#define SAIC_IVR                      (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x10u))  // Interrupt Vector Register
#define SAIC_EOICR                    (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x38u))  // End of Interrupt Command Register
#define SAIC_SPU                      (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x3Cu))  // Spurious Interrupt Vector Register
#define SAIC_IECR                     (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x40u))  // Interrupt Enable Command Register
#define SAIC_IDCR                     (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x44u))  // Interrupt Disable Command Register
#define SAIC_ICCR                     (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x48u))  // Interrupt Clear Command Register
#define SAIC_DCR                      (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0x6Cu))  // Debug Control Register
#define SAIC_WPMR                     (*(volatile unsigned int*) (SAIC_BASE_ADDR + 0xE4u))  // AIC Write Protect Mode Register

//
// Secure interrupt redirection
//
#define SFR_AICREDIR                  (*(volatile unsigned int *)0xF8028054u)
#define SFR_UID                       (*(volatile unsigned int *)0xF8028050u)
#define AICREDIR_KEY                  0x5F67B102u

//
// TC Timer counter, TC0 channel 0 used as embOS timer
//
#define TC0_BASE_ADDR                 (0xF801C000u)
#define TC1_BASE_ADDR                 (0xfffa0040u)
#define TC2_BASE_ADDR                 (0xfffa0080u)

/****** Assign timer used for OS *********************************/

#define OS_TC_BASE_ADDR               (TC0_BASE_ADDR)

#define OS_TC_CCR                     (*(volatile OS_U32*) (OS_TC_BASE_ADDR + 0x00u))
#define OS_TC_CMR                     (*(volatile OS_U32*) (OS_TC_BASE_ADDR + 0x04u))
#define OS_TC_CV                      (*(volatile OS_U32*) (OS_TC_BASE_ADDR + 0x10u))
#define OS_TC_RC                      (*(volatile OS_U32*) (OS_TC_BASE_ADDR + 0x1cu))
#define OS_TC_SR                      (*(volatile OS_U32*) (OS_TC_BASE_ADDR + 0x20u))
#define OS_TC_IER                     (*(volatile OS_U32*) (OS_TC_BASE_ADDR + 0x24u))
#define OS_TC_IDR                     (*(volatile OS_U32*) (OS_TC_BASE_ADDR + 0x28u))
#define OS_TC_BCR                     (*(volatile OS_U32*) (OS_TC_BASE_ADDR + 0xc0u))
#define OS_TC_BMR                     (*(volatile OS_U32*) (OS_TC_BASE_ADDR + 0xc4u))

#define TC_WPKEY                      (0x54494Du)

//
// PIT timer
//
#define PIT_BASE_ADDR                 (0xFC068630u)
#define PIT_MR                        (*(volatile unsigned int*)(PIT_BASE_ADDR + 0x00u))   // PIT Mode Register
#define PIT_SR                        (*(volatile unsigned int*)(PIT_BASE_ADDR + 0x04u))   // PIT Status Register
#define PIT_PIVR                      (*(volatile unsigned int*)(PIT_BASE_ADDR + 0x08u))   // PIT Periodic Interval Value Register
#define PIT_PIIR                      (*(volatile unsigned int*)(PIT_BASE_ADDR + 0x0Cu))   // PIT Periodic Interval Image Register
#define PIT_INT_ID                    (3u)
#define PITEN                         (24u)
#define PITIEN                        (25u)
#define PIT_ID                        (3u)
#define PIT_FIXED_PRESCALER           (16u)

#if OS_TIMER_IS_PIT
  #define OS_TIMER_ID (PIT_ID)

  #define AIC_TC_PENDIG_REG  (AIC_IPR0)
  #define AIC_TC_PENDING_BIT (3)
  #define OS_TIMER_INT_RESET (PIT_PIVR)

#else
  #if   (OS_TC_BASE_ADDR == TC0_BASE_ADDR)
    #define OS_TIMER_ID (40u)
    #define AIC_TC_PENDIG_REG  (AIC_IPR1)
    #define AIC_TC_PENDING_BIT (8)
  #elif (OS_TC_BASE_ADDR == TC1_BASE_ADDR)
    #define OS_TIMER_ID (41u)
    #define AIC_TC_PENDIG_REG  (AIC_IPR1)
    #define AIC_TC_PENDING_BIT (9)
  #elif (OS_TC_BASE_ADDR == TC2_BASE_ADDR)
    #define OS_TIMER_ID (42u)
    #define AIC_TC_PENDIG_REG  (AIC_IPR1)
    #define AIC_TC_PENDING_BIT (10)
  #endif
  #define OS_TIMER_INT_RESET (OS_TC_SR)

#endif

#define OS_TIMER_INT_PRIO             (0u)   // In Atmel ATSAMA5D4x AIC, 3 priority bits are implemented, 0 is the lowest priority

//
// MATRIX 32
//
#define MATRIX0_BASE_ADDR             (0xFC054000u)
#define MATRIX0_WPMR                  (*(volatile unsigned int*)(MATRIX0_BASE_ADDR + 0x1E4u))   // Security peripheral select register 1
#define MATRIX0_SPSELR1               (*(volatile unsigned int*)(MATRIX0_BASE_ADDR + 0x2C0u))   // Security peripheral select register 1
#define MATRIX0_SPSELR2               (*(volatile unsigned int*)(MATRIX0_BASE_ADDR + 0x2C4u))   // Security peripheral select register 2
#define MATRIX0_SPSELR3               (*(volatile unsigned int*)(MATRIX0_BASE_ADDR + 0x2C8u))   // Security peripheral select register 3
//
// MATRIX 64
//
#define MATRIX1_BASE_ADDR             (0xF001C000u)
#define MATRIX1_WPMR                  (*(volatile unsigned int*)(MATRIX1_BASE_ADDR + 0x1E4u))   // Security peripheral select register 1
#define MATRIX1_SPSELR1               (*(volatile unsigned int*)(MATRIX1_BASE_ADDR + 0x2C0u))   // Security peripheral select register 1
#define MATRIX1_SPSELR2               (*(volatile unsigned int*)(MATRIX1_BASE_ADDR + 0x2C4u))   // Security peripheral select register 2
#define MATRIX1_SPSELR3               (*(volatile unsigned int*)(MATRIX1_BASE_ADDR + 0x2C8u))   // Security peripheral select register 3

#define NOT_SECURED                   (1u)
#define SECURED                       (0u)
#define MAT_WPKEY                     (0x4D4154u)

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/

extern unsigned int __vector;

static void _OS_ISR_Undefined(void);

#if (OS_HANDLE_UNSECURE_INTS_ONLY == 0)
typedef OS_U32 TEST_ARRAY[3];
#endif

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

#if (OS_HANDLE_UNSECURE_INTS_ONLY == 0)
static const TEST_ARRAY _aIsInMatrix64 = {0x000D2504u, 0x001C0000u, 0x00000008u};
#endif

/*********************************************************************
*
*       MMU and cache configuration
*
*  Notes
*    The translation table has to be aligned on 16384 bytes boundary
*    The translation table has to be located in un-initialized memory,
*    because it may be filled from startup, before segments and variables
*    are initialized.
*/
#if  (defined(__ICCARM__))  // IAR
  #pragma data_alignment=16384
  static __no_init OS_U32 _TranslationTable[0x1000];
#elif (defined(__CC_ARM))
  #pragma arm section zidata = "NO_INIT"  // Has to be declared as separate section UNINIT in the scatter file
  __align(16384) static OS_U32 _TranslationTable[0x1000];
  #pragma arm section zidata              // back to default (.bss section)
#elif (defined(__GNUC__))   // GCC
  static OS_U32 _TranslationTable[0x1000] __attribute__ ((aligned (16384), section (".no_init")));
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _SpuriousIrqHandler()
*
*  Function description
*    default spurious IRQ handler
*
*  Notes
*    A spurious interrupt is called under one of the following
*    circumstances:
*    1. An external interrupt source is programmed in level-sensitive mode
*       and an active level occurs for only a short time.
*    2. An internal interrupt source is programmed in level sensitive
*       and the output signal of the corresponding embedded peripheral
*       is activated for a short time. (As in the case for the Watchdog.)
*    3. An interrupt occurs just a few cycles before the software begins
*       to mask it, thus resulting in a pulse on the interrupt source.
*
*    The spurious handler is called to detect these errors.
*/
static void _SpuriousIrqHandler(void) {
#if (OS_DEBUG != 0)
  static OS_U32 _SpuriousIrqCnt;

  _SpuriousIrqCnt++;
#endif
}

/*********************************************************************
*
*       _DefaultFiqHandler()
*
*  Function description
*    Dummy FIQ handler
*/
static void _DefaultFiqHandler(void) {
  while (1) {
    ;  // Not implemented
  }
}

#if (OS_HANDLE_UNSECURE_INTS_ONLY == 0)
/*********************************************************************
*
*       _ISRGetSecurity()
*
*  Notes:
*    (1) The peripheral is "secure" when it is marked "secure" (bit == 0)
*        in in the corresponding MATRIX SPSELR register
*        We use a table to find out which MATRIX has to be accessed.
*    (2) This function is not needed when RTOSInit is configured to
*        handle unsecured peripheral interrupts only.
*/
static OS_U32 _ISRGetSecurity(int ISRIndex) {
  const  TEST_ARRAY *pTestArray;
  int    IsInM64;
  OS_U32 *pMatrix;
  OS_U32 RegIndex;
  OS_U32 BitPos;
  OS_U32 Security;

#if OS_DEBUG
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    OS_INT_Disable();
    return 1;
  }
#endif
  //
  // Calculate register index from ISRIndex
  //
  RegIndex = ((ISRIndex) >> 5);
  BitPos   = ((ISRIndex) & 0x1fu);
  //
  // Get the security flag from the corresponding MATRIX registers.
  //
  pTestArray = &_aIsInMatrix64;
  IsInM64    = *pTestArray[RegIndex] & (1u << BitPos);
  if (IsInM64) {
    pMatrix = (OS_U32*) &(MATRIX1_SPSELR1);
  } else {
    pMatrix = (OS_U32*) &(MATRIX0_SPSELR1);
  }
  pMatrix   += (RegIndex);
  Security   = *pMatrix;
  Security >>= BitPos;
  Security  &=  (1u << 0);;
  return Security;
}

/*********************************************************************
*
*       _DisableIsrEx()
*
*  Notes:
*    (1) Access to source mode register may be locked by write protection
*    (2) ISRIndex range is not checked here, this is done in _ISRGetSecurity()
*/
static void _DisableIsrEx(int ISRIndex) {
  OS_U32 Security;
  OS_U32 IntState;

  OS_INT_PreserveAndDisableAll(&IntState);
  //
  // Examine which interrupt controller has to be addressed
  //
  Security = _ISRGetSecurity(ISRIndex);
  if (Security == NOT_SECURED) {
    AIC_SSR   = ISRIndex;
    AIC_IDCR  = 1u;
  } else {
    SAIC_SSR  = ISRIndex;
    SAIC_IDCR = 1u;
  }
  OS_INT_RestoreAll(&IntState);
}

/*********************************************************************
*
*       _ISRSetSecurity()
*
*  Function description
*    Sets the security option for a specified peripheral
*
*  Parameter
*    ISRIndex: Peripheral ID
*    Security: SECURED or NOT_SECURED
*
*  Returns
*    0: Success
*  !=0: Error
*
*  Notes:
*    (1) If the Scurity of the addressed peripheral changes,
*        The ISR is automatically disabled in the previously
*        connected interrupt controller.
*/
static int _ISRSetSecurity(int ISRIndex, OS_U32 Security) {
  int    Result;
  const  TEST_ARRAY *pTestArray;
  OS_U32 *pMatrix;
  int    IsInM64;
  OS_U32 SecurityVal;

  OS_U32 MatProtection;
  OS_U32 RegIndex;
  OS_U32 BitMask;
  OS_U32 IntState;

#if OS_DEBUG
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    OS_INT_Disable();
    return 1;
  }
#endif

  Result = 0;
  //
  // Calculate register offsets from ISRIndex
  //
  RegIndex = ((ISRIndex) >> 5);
  BitMask  = (1uL << ((ISRIndex) & 0x1Fu));
  //
  // Find out which MATRIX registers to use
  //
  pTestArray = &_aIsInMatrix64;
  IsInM64    = *pTestArray[RegIndex] & BitMask;

  OS_INT_PreserveAndDisableAll(&IntState);
  //
  // Save write protection state of MATRIX and enable write access
  //
  if (IsInM64) {
    pMatrix = (OS_U32*) &(MATRIX1_SPSELR1);
    MatProtection = MATRIX1_WPMR;
    MATRIX1_WPMR  = ((MAT_WPKEY) << 8);  // Disable write protection
  } else {
    pMatrix = (OS_U32*) &(MATRIX0_SPSELR1);
    MatProtection = MATRIX0_WPMR;
    MATRIX0_WPMR  = ((MAT_WPKEY) << 8);  // Disable write protection
  }
  //
  // Check cuurent security setting
  //
  pMatrix     += RegIndex;
  SecurityVal  = *pMatrix;
  if ((((SecurityVal & BitMask) == 0) && (Security == NOT_SECURED)) || (((SecurityVal & BitMask) != 0) && (Security == SECURED))) {
    //
    // The security option of the peripheral shall be changed. Disable ISR in current controller
    //
    _DisableIsrEx(ISRIndex);
  }
  //
  // Set the security option
  //
  if (Security == SECURED) {
    SecurityVal &= ~(BitMask);
  } else {
    SecurityVal |= BitMask;
  }
  *pMatrix = SecurityVal;
  //
  // Finally restore write protection
  //
  if (MatProtection != 0) {
    if (IsInM64) {
      MATRIX1_WPMR = ((MAT_WPKEY) << 8) | (1u << 0);
    } else {
      MATRIX0_WPMR = ((MAT_WPKEY) << 8) | (1u << 0);
    }
  }
  OS_INT_RestoreAll(&IntState);
  //
  // xxx AW: We may add a check here. By now, the result is always OK.
  //
  return Result;
}

#endif // OS_HANDLE_UNSECURE_INTS_ONLY

/*********************************************************************
*
*       _PMC_PclockEnable()
*
*  Function description
*    Enables/Disables peripheral clock for specific peripheral
*/
static void _PMC_PclockEnable(OS_U32 Id, OS_U8 Enable) {
  OS_U32 Protect;

  Protect   = (PMC_WPMR) & (1u << 0);
  PMC_WPMR  = (PMC_WPKEY << 8);     // Disable write protect
  if (Enable == PCLOCK_DISABLE) {
    //
    // Disable peripheral clock
    //
    if ((Id) >= 32u) {
      PMC_PCDR1 = (1uL << ((Id) - 32u));
    } else {
      PMC_PCDR0 = (1uL << (Id));
    }
  } else {
    //
    // Enable peripheral clock
    //
    if ((Id) >= 32u) {
      PMC_PCER1 = (1uL << ((Id) - 32u));
    } else {
      PMC_PCER0 = (1uL << (Id));
    }
  }
  if (Protect != 0) {
    //
    // PMC was protected, restore state
    //
    PMC_WPMR = ((PMC_WPKEY) << 8) | (1u << 0);
  }
}

/*********************************************************************
*
*       _SetVectorBaseAddr()
*
*  Function description
*    Sets the vector table base address
*/
#ifdef __GNUC__
static OS_INTERWORK OS_ARM_FUNC void _SetVectorBaseAddr(unsigned int Addr) ATTRIBUTE_NAKED_NO_INLINE;
#endif

static OS_INTERWORK OS_ARM_FUNC void _SetVectorBaseAddr(unsigned int Addr) {
#ifdef __ICCARM__
  __MCR(15, 0, Addr, 12, 0, 0);
#else
  __asm(".balign 4\n\
         .arm\n\
         mcr p15, 0, r0, c12, c0, 0\n\
         bx  lr\n");
#endif
  OS_USEPARA(Addr);
}

/*********************************************************************
*
*       _InitAIC()
*
*  Function description
*    Initialize interrupt controller by setting default vectors
*    and clearing all interrupts
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*/
#if OS_DEBUG
  // Enable AIC protected mode, GMSK: Enable nIRQ and nFIQ controlled by AIC
  #define AIC_DCR_VAL   ((1u << 0) | (0u << 1))
#else
  // Disable AIC protected mode, GMSK: Enable nIRQ and nFIQ controlled by AIC
  #define AIC_DCR_VAL   ((0u << 0) | (0u << 1))
#endif

static void _InitAIC(void) {
  OS_U32 i;
  OS_U32 Key;
  OS_U32 SerNo;

  _SetVectorBaseAddr((OS_U32)&__vector);  // Set interrupt vector table base addr
  //
  // Set all Interrupt sources to non-secure, regardless, some may be Secure Always...)
  //
  MATRIX0_WPMR    = ((MAT_WPKEY) << 8);   // Disable write protection
  MATRIX1_WPMR    = ((MAT_WPKEY) << 8);   // Disable write protection
  i = 0xFFFFFFFFuL;
  MATRIX0_SPSELR1 = i;
  MATRIX0_SPSELR2 = i;
  MATRIX0_SPSELR3 = i;
  MATRIX1_SPSELR1 = i;
  MATRIX1_SPSELR2 = i;
  MATRIX1_SPSELR3 = i;
  //
  // Redirect secure interrupts to non-secure interrupt controller.
  //
  SerNo = SFR_UID;
  Key   = SerNo ^ AICREDIR_KEY;
  //
  // bits[31:1] = key
  // bit[0] = 1 => all interrupts redirected to AIC
  // bit[0] = 0 => secure interrupts directed to SAIC, others to AIC (default)
  SFR_AICREDIR = Key & 0xFFFFFFFE | 0x1uL;
  //
  // Reset and initialize non secure AIC (IRQ) and SAIC (FIQ)
  //
  AIC_WPMR  = ((AIC_WPKEY) << 8);                // Disable write protection
  SAIC_WPMR = ((AIC_WPKEY) << 8);                // Disable write protection

  AIC_SSR   = 0;                                 // Select FIQ
  AIC_SVR   = (unsigned int)_DefaultFiqHandler;  // Set default FIQ handler

  SAIC_SSR  = 0;                                 // Select FIQ
  SAIC_SVR  = (unsigned int)_DefaultFiqHandler;  // Set default FIQ handler

  for (i = 1u; i < NUM_INT_SOURCES; i++) {
    //
    // Write a default vector for all peripherals for Non Secure interrupts (IRQ)
    //
    AIC_SSR   = i;
    AIC_SVR   = (unsigned int)_OS_ISR_Undefined;  // Set default interrupt handler
    AIC_IDCR  = 1u;                               // Disable all interrupts
    AIC_ICCR  = 1u;                               // Clear pending interrupt flag
    //
    // Write a default vector for all peripherals for Secure interrupts (FIQ)
    //
    SAIC_SSR  = i;
    SAIC_SVR  = (unsigned int)_OS_ISR_Undefined;  // Set default interrupt handler
    SAIC_IDCR = 1u;                               // Disable all interrupts
    SAIC_ICCR = 1u;                               // Clear pending interrupt flag
  }
  //
  // Write a vector for spurious interrupts
  //
  AIC_SPU  = (unsigned int)_SpuriousIrqHandler;
  SAIC_SPU = (unsigned int)_SpuriousIrqHandler;
  //
  // Perform 8 IT acknowledge (write any value in EOICR)
  //
  for (i = 0u; i < 8u ; i++) {
    AIC_EOICR  = 0;
    SAIC_EOICR = 0;
  }
  //
  // Setup AIC debug control, enable IRQ and FIQ
  //
  AIC_DCR  = AIC_DCR_VAL;
  SAIC_DCR = AIC_DCR_VAL;
}

/*********************************************************************
*
*       _PLL_Init()
*
*  Function description
*    Initialize the PLL
*/
static void _PLL_Init(void) {
  OS_U32 Protection;

  //
  // Initialize PLL if code is not executed from extenel DDR RAM
  //
  if ((OS_U32) _PLL_Init < DDR_CS_ADDR) {
    //
    // Unlock PMC protection
    //
    Protection = (PMC_WPMR) & (1u << 0);
    PMC_WPMR   = ((PMC_WPKEY) << 8);
    //
    // Switch to external 12MHz oscillator if not already done
    //
    if ((PMC_CKGR_MOR & (1uL << 24 )) == 0) {
      //
      // Enable external 12 MHz oscillator
      //
      PMC_CKGR_MOR |= ((1ul << 0) | (0x37u << 16));  // Requires key at bit position 16
      //
      // Wait until main oscillator is ready
      //
      while((PMC_CKGR_MCFR & (1uL << 24)) == 0) {
      }
      //
      // Disable bypass
      //
      PMC_CKGR_MOR = (PMC_CKGR_MOR & ~(1uL << 1)) | (0x37u << 16);
      //
      // Now switch main clock to external clock
      //
      PMC_CKGR_MOR |= ((1uL << 24) | (0x37u << 16));
      //
      // Wait until clock switching is done
      //
      while((PMC_SR & (1uL << 16)) == 0) {
      }
      //
      // Wait until main clock is ready
      //
      while((PMC_SR & (1uL << 3)) == 0) {
      }
      //
      // To save power, we might disable the internal RC oscillator here
      //
    }
    //
    // Select Main oscillator as input for MCK and PCK
    //
    PMC_MCKR = (PMC_MCKR & ~(0x03uL << 0)) | (0x01uL << 0);
    //
    // Wait until main clock is ready
    //
    while((PMC_SR & (1uL << 3)) == 0) {
    }
    //
    // Setup the PLLA control register, Multiplier
    //
    PMC_CKGR_PLLAR = (1uL   << 0)     // b0:       Magic value? Atmel sets this undocumented bit in the software samples
                   | (0x3Fu << 8)     // b13..b8:  PLLA count, number of counts until lock bit is set
                   | (0x0u  << 14)    // b17..b14: OUTA, has to be 0x00
                   | (PLL_MULA << 18) // b24..b18: Multiplier = PLLA_MUL + 1
                   | (1uL   << 29)    // b29:      Has to be written as 1
                   ;                  // All other bits unspecifed, write 0

    PMC_PLLICPR    = (0x0u << 0)      // b1..0:    ICP_PLLA, according electrical characteristis
                   ;                  // All other bits 0
    //
    // Wait until PLL is locked
    //
    while((PMC_SR & (1uL << 1)) == 0) {
    }
    //
    // Activate PLLA divider (DIV 2) and wait until main clock is ready
    //
    PMC_MCKR = PMC_MCKR | (1uL << 12);
    while((PMC_SR & (1uL << 3)) == 0) {
    }
    //
    // Reset the clock prescaler and wait auntil main clock is ready
    //
    PMC_MCKR &= ~(0x7u << 4);
    while((PMC_SR & (1uL << 3)) == 0) {
    }
    //
    // Set the Main clock divider and wait until main clock is ready
    // b9..8 of MCKR: MDIV, 0x03 => Master clock is Output / 3; SysClk DDR is 2 x MCK, DDRCK is MCK
    //
    PMC_MCKR = (PMC_MCKR & ~(0x3u << 8)) | (0x03u << 8);
    while((PMC_SR & (1uL << 3)) == 0) {
    }
    //
    // Switch main clock to PLL and wait until main clock is ready
    //
    PMC_MCKR = (PMC_MCKR & ~(0x03u << 0)) | (0x02u << 0);
    while((PMC_SR & (1uL << 3)) == 0) {
    }
    //
    // Finally restore the protection state of PMC
    //
    if (Protection != 0) {
      /* Was protected before, restore state */
      PMC_WPMR  = ((PMC_WPKEY) << 8) | (1u << 0);
    }
  }
}

/*********************************************************************
*
*       __readMemory32()
*/
#define __readMemory32(Addr, Zone)         (*(volatile OS_U32*) (Addr))

/*********************************************************************
*
*       __writeMemory32()
*/
#define __writeMemory32(Data, Addr, Zone)  ((*(volatile OS_U32*) (Addr)) = Data)

/*********************************************************************
*
*       __delay()
*/
static void __delay(OS_U32 Loops) {
  volatile unsigned int i;
  #define LOOP_CNT (OS_SYSTEM_FREQ / 10000u)
  while (Loops) {  // One loop should last 1ms
    for (i = 0; i < LOOP_CNT; i++) {
    }
    Loops--;
  }
}

/*********************************************************************
*
*       _DDR2_Init()
*
*  Function description
*    Initialize the DDR2 RAM connected to the external DDR RAM 0 area
*    This function does nothing if the code is running from this area
*    The function is a copy of the CSpy macro file function which
*    initializes the DDRAM.
*    To work, it reuires implementation of the macros or functions
*    __readMemory32()
*    __writedMemory32()
*    __delay()
*
*  Notes:
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*/
static void _DDR2_Init(void) {
  unsigned int type;

  //
  // Initialize DDRAM if code is not executed from extenal DDRAM
  //
  if ((OS_U32) _PLL_Init < DDR_CS_ADDR) {

    type = DDR_RAM_TYPE;

    // ------------------ DDR Controller Registers --------------


    //  0xFFFFEA00    Mode Register                       MPDDRC_MR
    //  0xFFFFEA04    Refresh Timer Register              MPDDRC_RTR
    //  0xFFFFEA08    Configuration Register              MPDDRC_CR
    //  0xFFFFEA0C    Timing Parameter 0 Register         MPDDRC_TPR0
    //  0xFFFFEA10    Timing Parameter 1 Register         MPDDRC_TPR1
    //  0xFFFFEA14    Timing Parameter 2 Register         MPDDRC_TPR2
    //  0xFFFFEA1C    Low-power Register                  MPDDRC_LPR
    //  0xFFFFEA20    Memory Device Register              MPDDRC_MD
    //  0xFFFFEA28    LPDDR2 Low-power Register           MPDDRC_LPDDR2_LPR
    //  0xFFFFEA2C    LPDDR2 Calibration and MR4 Register MPDDRC_LPDDR2_CAL_MR4
    //  0xFFFFEA30    LPDDR2 Timing Calibration Register  MPDDRC_LPDDR2_TIM_CAL
    //  0xFFFFEA34    IO Calibration                      MPDDRC_IO_CALIBR
    //  0xFFFFEA38    OCMS Register                       MPDDRC_OCMS
    //  0xFFFFEA3C    OCMS KEY1 Register                  MPDDRC_OCMS_KEY1         Write-only
    //  0xFFFFEA40    OCMS KEY2 Register                  MPDDRC_OCMS_KEY2         Write-only
    //  0xFFFFEA74    DLL Master Offset Register          MPDDRC_DLL_MOR
    //  0xFFFFEA78    DLL Slave Offset Register           MPDDRC_DLL_SOR
    //  0xFFFFEA7C    DLL Master Status Register          MPDDRC_DLL_MSR           Read-only
    //  0xFFFFEA80    DLL Slave 0 Status Register         MPDDRC_DLL_S0SR          Read-only
    //  0xFFFFEA84    DLL Slave 1 Status Register         MPDDRC_DLL_S1SR          Read-only
    //  0xFFFFEAE4    Write Protect Control Register      MPDDRC_WPCR
    //  0xFFFFEAE8    Write Protect Status Register       MPDDRC_WPSR              Read-only

    // -----------------------------------------------

    // ---------------- DDR2 Timings (133MHz) -----------------

                             // -------- Configuration Register
    //  n_row      =  14
    //  n_col      =  10
    //  n_bank     =  8
                             // -------- Refresh Timer Register
    //  t_refresh  =  520
                             // -------- Timing 0 Register
    //  t_tras     =  TRAS   6
    //  t_trcd     =  TRCD   2
    //  t_twr      =  TWR    2
    //  t_trc      =  TRC    8
    //  t_trp      =  TRP    2
    //  t_trrd     =  TRRD   2
    //  t_twtr     =  TWTR   2
    //  t_tmrd     =  TMRD   2
                             // -------- Timing 1 Register
    //  t_trfc     =  TRFC   26/14
    //  t_txsnr    =  TXSNR  28/16
    //  t_txsrd    =  TXSRD  208
    //  t_txp      =  TXP    2
                             // -------- Timing 2 Register
    //  t_txard    =  TXARD  7
    //  t_tards    =  TXARDS 7
    //  t_trpa     =  TRPA   3
    //  t_trtp     =  TRTP   2
    //  t_tfaw     =  TFAW  10
    // -----------------------------------------------
    __delay(2);
    __writeMemory32(0x00008000,0xFFFFFE44,"Memory");   // Disable Watchdog
    __writeMemory32(0x0,0xFFFFEA1C,"Memory");
    __writeMemory32(0x00020000,0xFFFFFD00,"Memory");   // Enable MPDDR controller clock
    __writeMemory32(0x00000004,0xFFFFFC00,"Memory");   // System Clock Enable Register : Enable DDR clock

    __writeMemory32(0x00000020,0xFFFFEA24,"Memory");   // DDRSDRC High Speed Register (MPDDRC_HS)  : hidden option -> calibration during autorefresh
    __writeMemory32(0x00030000,0xF0038004,"Memory");   // SFR_DDRCFG  DDR Configuration  Force DDR_DQ and DDR_DQS input buffer always on

    __writeMemory32(0x01010001,0xFFFFEA78,"Memory");   // MPDDRC DLL Slave Offset Register : set DLL Slave x Delay Line Offset
    __writeMemory32(0xC5011F07,0xFFFFEA74,"Memory");   // MPDDRC DLL Master Offset Register : DLL Master Delay Line Offset + DLL CLK90 Delay Line Offset + DLL Offset Selection
    //__writeMemory32(0x011F07,0xFFFFEA74,"Memory");   // MPDDRC DLL Master Offset Register : DLL Master Delay Line Offset + DLL CLK90 Delay Line Offset + DLL Offset Selection

    __writeMemory32(0x00870a03,0xFFFFEA34,"Memory");   // MPDDRC I/O Calibration Register : set Resistor Divider + IO Calibration (delay between an IO Calibration Command and any Valid commands)
    //__writeMemory32(0x00870303,0xFFFFEA34,"Memory");   // MPDDRC I/O Calibration Register : set Resistor Divider + IO Calibration (delay between an IO Calibration Command and any Valid commands)

    __writeMemory32(0x00000028,0xFFFFEA24,"Memory");   // DDRSDRC High Speed Register : enable calibration during autorefresh
    __writeMemory32(0x00030001,0xF0038004,"Memory");   // SFR_DDRCFG  DDR Configuration : Force DDR_DQ and DDR_DQS input buffer always on
    __writeMemory32(0x01010001,0xFFFFEA78,"Memory");   // MPDDRC DLL Slave Offset Register : set DLL Slave x Delay Line Offset
    __writeMemory32(0xC5011f07,0xFFFFEA74,"Memory");   // MPDDRC DLL Master Offset Register : DLL Master Delay Line Offset + DLL CLK90 Delay Line Offset + DLL Offset Selection
    __writeMemory32(0x00870a03,0xFFFFEA34,"Memory");   // MPDDRC I/O Calibration Register : set Resistor Divider + IO Calibration (delay between an IO Calibration Command and any Valid commands)
    __writeMemory32(0x00000028,0xFFFFEA24,"Memory");   // DDRSDRC High Speed Register : enable calibration during autorefresh
    __writeMemory32(0x00030001,0xF0038004,"Memory");   // SFR_DDRCFG  DDR Configuration : Force DDR_DQ and DDR_DQS input buffer always on


    __writeMemory32(0x00000006,0xFFFFEA20,"Memory");   // Memory Device Register  :  32bit mode - DDR2 mode
    if (type == 0) {
        __writeMemory32(0x00B0003D,0xFFFFEA08,"Memory");   // Configuration Register  :  row = 13, column(DDR) = 10, CAS 3, DLL reset disable, phase error correction is enabled / normal driver strength
    } else {
        __writeMemory32(0x00B0004D,0xFFFFEA08,"Memory");   // Configuration Register  :  row = 14, column(DDR) = 10, CAS 3, DLL reset disable, phase error correction is enabled / normal driver strength
    }
    __writeMemory32(0x22228226,0xFFFFEA0C,"Memory");   // Timing 0 Register : tras  | trcd  | twr   | trc  | trp | trrd | twtr | tmrd
    //__writeMemory32(0x22228226,0xFFFFEA0C,"Memory");   // Timing 0 Register : tras  | trcd  | twr   | trc  | trp | trrd | twtr | tmrd
    __writeMemory32(0x02D0100E,0xFFFFEA10,"Memory");   // Timing 1 Register : trfc  | txsnr | txsrd | txp
    __writeMemory32(0x000A2277,0xFFFFEA14,"Memory");   // Timing 2 Register : txard | tards | trpa  | trtp | tfaw
    __writeMemory32(0x00000001,0xFFFFEA00,"Memory");   // Mode register : command  NOP --> ENABLE CLOCK output
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);   // wait 1 ms


    __writeMemory32(0x00000001,0xFFFFEA00,"Memory");   // Mode register : command  NOP --> ENABLE CLOCK output
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);
    __writeMemory32(0x00000001,0xFFFFEA00,"Memory");   // Mode register : command  NOP --> ENABLE CLOCK output
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);

    __writeMemory32(0x00000002,0xFFFFEA00,"Memory");   // Mode register : command  All Banks Precharge
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);

    __writeMemory32(0x00000005,0xFFFFEA00,"Memory");   // Mode register : command  Extended Load Mode Register : Set EMR Ext Mode Reg EMSR2 BA0=0 BA1=1
    if (type == 0) {
        __writeMemory32(0x00000000,0x24000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    } else {
        __writeMemory32(0x00000000,0x28000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    }
    __delay(1);


    __writeMemory32(0x00000005,0xFFFFEA00,"Memory");   // Mode register : command  Extended Load Mode Register : Set EMR Ext Mode Reg EMSR3 BA0=1 BA1=1
    if (type == 0) {
        __writeMemory32(0x00000000,0x26000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    } else {
        __writeMemory32(0x00000000,0x2C000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    }

    __delay(1);
    __writeMemory32(0x00000005,0xFFFFEA00,"Memory");   // Mode register : command  Extended Load Mode Register : Set EMR Ext Mode Reg EMSR1 BA0=1 BA1=0 ENABLE DLL

    if (type == 0) {
        __writeMemory32(0x00000000,0x22000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    } else {
        __writeMemory32(0x00000000,0x24000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    }
    __delay(1);

    __writeMemory32(0x00B000BD,0xFFFFEA08,"Memory");   // Configuration Register : Enable DLL reset

    __writeMemory32(0x00000003,0xFFFFEA00,"Memory");   // Mode register : command  RESET DLL
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);
    __writeMemory32(0x00000002,0xFFFFEA00,"Memory");   // Mode register : command  All Banks Precharge
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);
    __writeMemory32(0x00000004,0xFFFFEA00,"Memory");   // Mode register : 2 * command  Auto-Refresh
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);
    __writeMemory32(0x00000004,0xFFFFEA00,"Memory");   // Mode register :
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);

    __writeMemory32(0x00B0003D,0xFFFFEA08,"Memory");   // Configuration Register : disable DLL reset

    __writeMemory32(0x00000003,0xFFFFEA00,"Memory");   // Mode register :  MRS  initialize device operation (CAS latency, burst length and disable DLL reset)
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);

    __writeMemory32(0x00B0703D,0xFFFFEA08,"Memory");   // Configuration Register : OCD default value

    __writeMemory32(0x00000005,0xFFFFEA00,"Memory");   // Mode register : EMRS1   OCD Default values
    if (type == 0) {
        __writeMemory32(0x00000000,0x22000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    } else {
        __writeMemory32(0x00000000,0x24000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    }
    __delay(1);

    __writeMemory32(0x00B0003D,0xFFFFEA08,"Memory");   // Configuration Register : OCD exit

    __writeMemory32(0x00000005,0xFFFFEA00,"Memory");   // Mode register : EMRS1   OCD exit
    if (type == 0) {
        __writeMemory32(0x00000000,0x22000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    } else {
        __writeMemory32(0x00000000,0x24000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    }
    __delay(1);
    __writeMemory32(0x00000000,0xFFFFEA00,"Memory");   // Mode register : command  Normal mode
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);
    __writeMemory32(0x00000000,0x20000000,"Memory");   // DDR2 memory : access memory to validate preeceeding command
    __delay(1);

    __writeMemory32(0x0000061A,0xFFFFEA04,"Memory");   // Refresh Timer register : 520 for 133 MHz
  }
}

/*********************************************************************
*
*       _OS_ISR_Undefined()
*
*  Function description
*    Is called when an uninstalled interrupt was detected
*    As interrupt pending condition of peripherals has to be reset,
*    program will not continue when interrupt is ignored.
*/
static void _OS_ISR_Undefined(void) {
  volatile int Dummy;

  Dummy = 1;
  while (Dummy > 0) {
    ;  // You may set a breakpoint here to detect non-installed ISR-handler
  }
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return OS_TC_CV;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return (AIC_TC_PENDIG_REG & (1uL << AIC_TC_PENDING_BIT));
}

/*********************************************************************
*
*       _OS_ISR_Tick()
*
*  Function description
*    This is the hardware timer exception handler.
*/
static void _OS_ISR_Tick(void) {
  volatile unsigned int Dummy;

  Dummy = OS_TIMER_INT_RESET;   // Reset interrupt pending condition
  OS_USEPARA(Dummy);
  OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_Process();
#endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  volatile OS_U32 Dummy;

  OS_INT_IncDI();
  //
  // Ensure, caches are synchronized
  //
  OS_ARM_CACHE_Sync();
  //
  // Install Timer ISR Handler and set interrupt priority
  //
  OS_ARM_InstallISRHandler(OS_TIMER_ID, _OS_ISR_Tick);
  OS_ARM_ISRSetPrio(OS_TIMER_ID, OS_TIMER_INT_PRIO);
  OS_ARM_EnableISR(OS_TIMER_ID);
  //
  // Initialize and start timer
  //
  _PMC_PclockEnable(OS_TIMER_ID, PCLOCK_ENABLE);

#if OS_TIMER_IS_PIT
  PIT_MR = ((OS_TIMER_RELOAD & 0x000FFFFFu) | (1u << PITEN) | (1u << PITIEN));
#else
  OS_TC_IDR   = 0xffu;                // Disable all timer/counter 0 interrupts.
  OS_TC_CMR   = 0x00004000u;          // Capture mode, CPCTRG=1, TCCLKS=0 (=> MCLK/2)
  OS_TC_RC    = OS_TIMER_RELOAD;      // Set compare register
  OS_TC_CCR   = (1u << 0)             // Enable clock
              | (1u << 2)             // Software trigger.
              ;
  Dummy       = OS_TC_SR;             // Read timer/counter 0 status register to clear flags.
  OS_TC_IER   = (1u << 4);            // Interrupt on RC compare.
#endif
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       embOS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;

  pISR = (OS_ISR_HANDLER*)AIC_IVR;   // Read interrupt vector to release NIRQ to CPU core
#if (OS_DEBUG != 0)
  AIC_IVR = (OS_U32)pISR;            // Write back any value to IVR register to allow interrupt stacking in protected mode
#endif
  OS_INT_EnterNestable();            // Inform OS that interrupt handler is running
  pISR();                            // Call interrupt service routine
  AIC_EOICR = 0;                     // interrupt controller =>  Restore previous priority
  OS_INT_LeaveNestable();
}

/*********************************************************************
*
*       OS_fiq_handler()
*
*  Function description
*    Detect reason for FRQ and call correspondig service routine.
*    OS_fiq_handler is called from the low level irq handler
*    function FIQ_Handler()
*    which is part of the startup file for the project
*/
#ifdef __cplusplus
extern "C" {     /* Make sure we have C-declarations in C++ programs */
#endif
OS_INTERWORK void OS_fiq_handler(void);
#ifdef __cplusplus
}
#endif
OS_INTERWORK void OS_fiq_handler(void) {
  OS_ISR_HANDLER* pISR;

  pISR = (OS_ISR_HANDLER*) SAIC_IVR;   // Read interrupt vector to release nIRQ to CPU core
#if OS_DEBUG
  SAIC_IVR = (OS_U32) pISR;            // Write back any value to IVR register to allow interrupt stacking in protected mode
#endif
#if OS_ALLOW_NESTED_FIQ
  __enable_fiq();                      // Now FIQ interrupts may be reenabled
#endif
  pISR();                              // Call interrupt service routine
  __disable_fiq();                     // Disable fiq and unlock
  AIC_EOICR = 0u;                      // interrupt controller =>  Restore previous priority
}

#if (OS_HANDLE_UNSECURE_INTS_ONLY)

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Used to install a specific interrupt vector when ARM
*    CPUs with vectored interrupt controller are used.
*
*  Parameters
*    ISRIndex:    Index of the interrupt source, normally the interrupt vector number.
*    pISRHandler: Address of the interrupt handler function.
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler(int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;
  OS_U32           Protection;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  Protection  = (AIC_WPMR) & (1u << 0);     /* Get current protection state */
  AIC_WPMR    = ((AIC_WPKEY) << 8);         /* Allow write access to SVR    */
  AIC_SSR     = ISRIndex;
  pOldHandler = (OS_ISR_HANDLER*) AIC_SVR;
  AIC_SVR     = (unsigned int)pISRHandler;  /* Set new interrupt handler    */
  if (Protection != 0) {
    //
    // AIC was write protected, restore state
    //
    AIC_WPMR   = ((AIC_WPKEY) << 8) | (1ul << 0);  /* Lock write access to AIC */
  }
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Used to enable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be enabled.
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  AIC_SSR  = ISRIndex;
  AIC_IECR = 1u;
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Used to disable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be disabled.
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  AIC_SSR  = ISRIndex;
  AIC_IDCR = 1u;
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Function description
*    Used to set or modify the priority of a specific interrupt
*    source by programming the interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be modified.
*    Prio:     The priority which should be set for the specific interrupt.
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  int    OldPrio;
  OS_U32 Protection;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  Protection = (AIC_WPMR) & (1u << 0);  /* Get current protection state */
  AIC_WPMR   = (AIC_WPKEY) << 8;        /* Allow write access to AIC    */
  AIC_SSR    = ISRIndex;
  OldPrio    = AIC_SMR & INT_PRIORITY_MASK;
  AIC_SMR    = (OldPrio & ~INT_PRIORITY_MASK) | (Prio & INT_PRIORITY_MASK);
  if (Protection != 0) {
    //
    // Was protected before, restore state
    //
    AIC_WPMR  = ((AIC_WPKEY) << 8) | (1u << 0);
  }
  OS_INT_EnableConditional();
  return OldPrio & INT_PRIORITY_MASK;
}

#else  // The following functions support secured and unsecured peripherals

/*********************************************************************
*
*       OS_ARM_ISRSetSecure()
*
*  Notes:
*    (1) This Function is available only when secured and unsecured
*        peripherals should be handled.
*/
int OS_ARM_ISRSetSecure(int ISRIndex) {
  return _ISRSetSecurity(ISRIndex, SECURED);
}

/*********************************************************************
*
*       OS_ARM_ISRSetUnSecure()
*
*  Notes:
*    (1) This Function is available only when secured and unsecured
*        peripherals should be handled.
*/
int OS_ARM_ISRSetUnSecure(int ISRIndex) {
  return _ISRSetSecurity(ISRIndex, NOT_SECURED);
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Notes:
*    (1) This version handles secured and unsecured peripherals
*        The security option of the peripheral has to be set before
*        calling this function
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler (int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;
  OS_U32 Security;
  OS_U32 Protection;
  OS_U32 IntState;

  OS_INT_PreserveAndDisableAll(&IntState);
  //
  // Examine which interrupt controller has to be addressed
  //
  Security = _ISRGetSecurity(ISRIndex);
  if (Security == NOT_SECURED) {
    Protection  = (AIC_WPMR) & (1u << 0);  /* Get current protection state */
    AIC_WPMR    = ((AIC_WPKEY) << 8);      /* Allow write access to SVR    */
    AIC_SSR     = ISRIndex;
    pOldHandler = (OS_ISR_HANDLER*) AIC_SVR;
    AIC_SVR     = (unsigned int)pISRHandler;  // Set new interrupt handler
  } else {
    Protection  = (SAIC_WPMR) & (1u << 0); /* Get current protection state */
    SAIC_WPMR   = ((AIC_WPKEY) << 8);      /* Allow write access to SVR    */
    SAIC_SSR    = ISRIndex;
    pOldHandler = (OS_ISR_HANDLER*) SAIC_SVR;
    SAIC_SVR    = (unsigned int)pISRHandler;  // Set new interrupt handler
  }
  if (Protection != 0) {
    //
    // AIC was write protected, restore state
    //
    if (Security == NOT_SECURED) {
      AIC_WPMR   = ((AIC_WPKEY) << 8) | (1ul << 0);  /* Lock write access to AIC */
    } else {
      SAIC_WPMR  = ((AIC_WPKEY) << 8) | (1ul << 0);  /* Lock write access to AIC */
    }
  }
  OS_INT_RestoreAll(&IntState);
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Notes:
*    (1) This version handles secured and unsecured peripherals
*        The security option of the peripheral has to be set before
*        calling this function
*/
void OS_ARM_EnableISR(int ISRIndex) {
  OS_U32 Security;
  OS_U32 IntState;

  OS_INT_PreserveAndDisableAll(&IntState);
  //
  // Examine which interrupt controller has to be addressed
  //
  Security = _ISRGetSecurity(ISRIndex);
  if (Security == NOT_SECURED) {
    AIC_SSR   = ISRIndex;
    AIC_IECR  = 1u;
  } else {
    SAIC_SSR  = ISRIndex;
    SAIC_IECR = 1u;
  }
  OS_INT_RestoreAll(&IntState);
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Notes:
*    (1) This version handles secured and unsecured peripherals
*        The security option of the peripheral has to be set before
*        calling this function
*/
void OS_ARM_DisableISR(int ISRIndex) {
  _DisableIsrEx(ISRIndex);
}
/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Notes:
*    (1) This version handles secured and unsecured peripherals
*        The security option of the peripheral has to be set before
*        calling this function
*    (2) Access to source mode register may be locked by write protection
*    (3) ISRIndex range is not checked here, this is done in _ISRGetSecurity()
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  int    OldPrio;
  OS_U32 Security;
  OS_U32 Protection;
  OS_U32 IntState;

  OS_INT_PreserveAndDisableAll(&IntState);
  //
  // Examine which interrupt controller has to be addressed
  //
  Security = _ISRGetSecurity(ISRIndex);
  if (Security == NOT_SECURED) {
    Protection = (AIC_WPMR) & (1u << 0);  /* Get current protection state */
    AIC_WPMR   = (AIC_WPKEY) << 8;        /* Allow write access to AIC    */
    AIC_SSR    = ISRIndex;
    OldPrio    = AIC_SMR & INT_PRIORITY_MASK;
    AIC_SMR    = (OldPrio & ~INT_PRIORITY_MASK) | (Prio & INT_PRIORITY_MASK);
    if (Protection != 0) {
      /* Was protected before, restore state */
      AIC_WPMR  = ((AIC_WPKEY) << 8) | (1u << 0);
    }
  } else {
    Protection = (SAIC_WPMR) & (1u << 0); /* Get current protection state */
    SAIC_WPMR  = (AIC_WPKEY) << 8;        /* Allow write access to SAIC   */
    SAIC_SSR   = ISRIndex;
    OldPrio    = SAIC_SMR & INT_PRIORITY_MASK;
    SAIC_SMR   = (OldPrio & ~INT_PRIORITY_MASK) | (Prio & INT_PRIORITY_MASK);
    if (Protection != 0) {
      /* Was protected before, restore state */
      SAIC_WPMR  = ((AIC_WPKEY) << 8) | (1u << 0);
    }
  }
  OS_INT_RestoreAll(&IntState);
  return OldPrio & INT_PRIORITY_MASK;
}

#endif  // OS_HANDLE_UNSECURE_INTS_ONLY

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
int __low_level_init(void) {
  //
  // Disable watchdog
  //
  WDT_MR = (1u << WDDIS);
  //
  // Initialize PLL
  //
  _PLL_Init();
  //
  // Initialize DDR2 RAM
  //
  _DDR2_Init();
  //
  // Init interrupt controller
  //
  _InitAIC();
  //
  // Init MMU and caches. This defines the virtual memory map, which is used during execution.
  // Memory mapping should be complete, meaning 4096 entries. If so, a call of OS_ARM_MMU_InitTT() is not required.
  // Code below fills in ascending VAddr order
  //
  OS_ARM_MMU_InitTT(_TranslationTable);
  //                                         Mode                                Access permissions        VAddr  PAddr  Size[MB]
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_MMU_UNMAPPED,                                          0x000, 0x000, 0x001);  // Internal ROM 128KB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x001, 0x001, 0x001);  // NFC SRAM, 4KB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_WRITE_BACK_ALLOC,                            0x002, 0x002, 0x001);  // Internal SRAM, 128 KB, original address
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x003, 0x003, 0x001);  // VDEC, 1MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x004, 0x004, 0x001);  // UDPHS RAM, 1MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x005, 0x005, 0x001);  // OHCI controller
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x006, 0x006, 0x001);  // EHCI controller
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x007, 0x007, 0x001);  // AXI Matrix
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x008, 0x008, 0x001);  // DAP
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x009, 0x009, 0x001);  // SMD
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x00A, 0x00A, 0x001);  // L2CC
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x00B, 0x002, 0x001);  // Internal SRAM, 128 KB, Remapped, uncached access
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_MMU_UNMAPPED,                                          0x00C, 0x00C, 0x0F4);  // Unmapped area
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_WRITE_BACK_ALLOC,                            0x100, 0x100, 0x100);  // EBI, CS0 area, 256MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_WRITE_BACK_ALLOC,                            0x200, 0x200, 0x100);  // External DDR RAM 0 area, 256 MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_MMU_UNMAPPED,                                          0x300, 0x300, 0x100);  // External DDR RAM 1 area, Not assembled
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_WRITE_BACK_ALLOC,                            0x400, 0x200, 0x100);  // External DDR RAM 0 area, AES
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_WRITE_BACK_ALLOC,                            0x500, 0x500, 0x100);  // External DDR RAM 1 area, AES
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x600, 0x600, 0x100);  // EBI, CS1 area, 256MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x700, 0x700, 0x100);  // EBI, CS2 area, 256MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x800, 0x800, 0x080);  // EBI, CS3 area, 128MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_MMU_UNMAPPED,                                          0x880, 0x880, 0x080);  // Undefined (Abort)
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0x900, 0x900, 0x100);  // NFC, 256 MB
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED,                            0xA00, 0x200, 0x100);  // External DDR RAM 0 area, 256 MB, Remapped, uncached access
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_MMU_UNMAPPED,                                          0xB00, 0xA00, 0x400);  // Undefined area, may be used to remapp memory with different cache mode
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED | OS_ARM_MMU_EXECUTE_NEVER, 0xF00, 0xF00, 0x100);  // SFRs, non cacheable, non bufferable
  OS_ARM_MMU_Enable      (_TranslationTable);
  OS_ARM_ICACHE_Enable();
  OS_ARM_DCACHE_Enable();
  return 1;  // Always return 1 to enable segment initialization!
}

/*************************** End of file ****************************/
