/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : UART implementation

Additional information:

  Device : LPC2106
  Board  : KS2106 LPC210X KickStart

  Unit | UART   | Board connector
  ===============================
  0    | UART0  | P0
  1    | UART1  | P1

*/

#include "BSP_UART.h"
#include "RTOS.h"      // For OS_INT_Enter()/OS_INT_Leave(). Remove this line and OS_INT_* functions if not using OS.

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/

#define PLL_MULTIPLIER                  (4uL)         // PLL multiplies by 4
#define EXTERNAL_FREQ                   (14745600uL)  // external crystal frequency
#define SYSTEM_FREQ                     (EXTERNAL_FREQ * PLL_MULTIPLIER)
#define UART_FREQ                       (SYSTEM_FREQ / 4uL)

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/
//
// UART0
//
#define UART0_BASE_ADDR                 (0xE000C000uL)
#define UART0_RBR                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x00uL))
#define UART0_THR                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x00uL))
#define UART0_IER                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x04uL))
#define UART0_IIR                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x08uL))
#define UART0_FCR                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x08uL))
#define UART0_LCR                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x0CuL))
#define UART0_LSR                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x14uL))
#define UART0_SCR                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x1CuL))
#define UART0_DLL                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x00uL))
#define UART0_DLM                       (*(volatile unsigned char*)(UART0_BASE_ADDR + 0x04uL))
//
// UART1
//
#define UART1_BASE_ADDR                 (0xE0010000uL)
#define UART1_RBR                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x00uL))
#define UART1_THR                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x00uL))
#define UART1_IER                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x04uL))
#define UART1_IIR                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x08uL))
#define UART1_FCR                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x08uL))
#define UART1_LCR                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x0CuL))
#define UART1_LSR                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x14uL))
#define UART1_SCR                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x1CuL))
#define UART1_DLL                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x00uL))
#define UART1_DLM                       (*(volatile unsigned char*)(UART1_BASE_ADDR + 0x04uL))
//
// UART configuration
//
#define UART0_INT_SOURCE                ( 6uL)
#define UART1_INT_SOURCE                ( 7uL)
#define UART_INT_INDEX                  (14uL)

#define UART0_PCONP_SHIFT               (3uL)
#define UART1_PCONP_SHIFT               (4uL)

#define UART0_PINSEL                    ((1uL <<  0) | (1uL <<  2))
#define UART1_PINSEL                    ((1uL << 16) | (1uL << 18))
//
// UART shifts, masks, IDs ...
//
#define UART_RX_FULL_INT_ENABLE_SHIFT   (0uL)
#define UART_TX_EMPTY_INT_ENABLE_SHIFT  (1uL)

#define UART_INT_MASK                   (0x0EuL)
#define UART_ERROR_INT_STATUS           (0x06uL)
#define UART_RX_INT_STATUS              (0x04uL)
#define UART_TX_INT_STATUS              (0x02uL)

#define UART0                           (0uL)
#define UART1                           (1uL)

#define UART_BAUDDIVIDE(Baudrate)       ((UART_FREQ + ((Baudrate) * 8uL)) / ((Baudrate) * 16uL))

//
// Port mode settings
//
#define PINSEL0                         (*(volatile unsigned long*)0xE002C000uL)
//
// Power, reset clock control register
//
#define PCONP                           (*(volatile unsigned long*)0xE01FC0C4uL)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static BSP_UART_TX_CB* _pfWriteCB;
static BSP_UART_RX_CB* _pfReadCB;

/*********************************************************************
*
*       Prototypes
*
*  Declare ISR handler here to avoid "no prototype" warning.
*  They are not declared in any CMSIS header.
*
**********************************************************************
*/

#if defined(__cplusplus)
  extern "C" {                // Make sure we have C-declarations in C++ programs.
#endif

void UART0_RX_TX_IRQHandler(void);
void UART1_RX_TX_IRQHandler(void);

#if defined(__cplusplus)
}                             // Make sure we have C-declarations in C++ programs.
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _SetBaudrate()
*
*  Function description
*    Configures the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
static void _SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  BSP_UART_USE_PARA(Unit);
  BSP_UART_USE_PARA(Baudrate);
}

/*********************************************************************
*
*       Global functions, IRQ handler
*
**********************************************************************
*/

/*********************************************************************
*
*       UART0_RX_TX_IRQHandler()
*
*  Function description
*    UART Rx & Tx interrupt handler.
*
*  Additional information
*    Needs to inform the OS that we are in interrupt context.
*/
void UART0_RX_TX_IRQHandler(void) {
  volatile unsigned char Data;
           unsigned int  Unit;
           unsigned char Status;

  Unit   = UART0;
  Status = UART0_IIR;                                           // Examine interrupt identification register
  if ((Status & UART_INT_MASK) == UART_ERROR_INT_STATUS) {      // Error pending ?
    Data = UART0_RBR;                                           // Discard data
    Data = UART0_LSR;                                           // reset error
  } else if ((Status & UART_INT_MASK) == UART_RX_INT_STATUS) {  // Data received?
    if (_pfReadCB != 0) {
      _pfReadCB(Unit, UART0_RBR);                               // Forward the received byte to the callback.
    }
  }
  if ((Status & UART_INT_MASK) == UART_TX_INT_STATUS) {  // Tx interrupt pending?
    if ((_pfWriteCB != 0) && (_pfWriteCB(Unit) != 0)) {
      UART0_IER &= ~(1uL << UART_TX_EMPTY_INT_ENABLE_SHIFT);           // Disable Tx interrupts
    }
  }
}

/*********************************************************************
*
*       UART1_RX_TX_IRQHandler()
*
*  Function description
*    UART Rx & Tx interrupt handler.
*
*  Additional information
*    Needs to inform the OS that we are in interrupt context.
*/
void UART1_RX_TX_IRQHandler(void) {
  volatile unsigned char Data;
           unsigned int  Unit;
           unsigned char Status;

  Unit   = UART1;
  Status = UART1_IIR;                                           // Examine interrupt identification register
  if ((Status & UART_INT_MASK) == UART_ERROR_INT_STATUS) {      // Error pending ?
    Data = UART1_RBR;                                           // Discard data
    Data = UART1_LSR;                                           // reset error
  } else if ((Status & UART_INT_MASK) == UART_RX_INT_STATUS) {  // Data received?
    if (_pfReadCB != 0) {
      _pfReadCB(Unit, UART1_RBR);                               // Forward the received byte to the callback.
    }
  } else if ((Status & UART_INT_MASK) == UART_TX_INT_STATUS) {  // Tx interrupt pending?
    if ((_pfWriteCB != 0) && (_pfWriteCB(Unit) != 0)) {
      UART1_IER &= ~(1uL << UART_TX_EMPTY_INT_ENABLE_SHIFT);           // Disable Tx interrupts
    }
  }
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_UART_Init()
*
*  Function description
*    Initializes the UART peripheral.
*
*  Parameters
*    Unit       : Unit number (typically zero-based).
*    Baudrate   : Baud rate to configure [Hz].
*    NumDataBits: Number of data bits to use.
*    Parity     : One of the following values:
*                   * BSP_UART_PARITY_NONE
*                   * BSP_UART_PARITY_ODD
*                   * BSP_UART_PARITY_EVEN
*    NumStopBits: Number of stop bits to use.
*/
void BSP_UART_Init(unsigned int Unit, unsigned long Baudrate, unsigned char NumDataBits, unsigned char Parity, unsigned char NumStopBits) {
  BSP_UART_USE_PARA(NumDataBits);
  BSP_UART_USE_PARA(Parity);
  BSP_UART_USE_PARA(NumStopBits);
  OS_INT_IncDI();
  switch (Unit) {
  case UART0:
    PINSEL0   |= UART0_PINSEL;
    PCONP     |= (1uL << UART0_PCONP_SHIFT);
    UART0_IER  = 0x00uL;      // Initially disable all interrupts
    UART0_LCR  = 0x80uL;      // Set DLAB to initialize Baudrate generator
    UART0_DLL  =  (UART_BAUDDIVIDE(Baudrate)       & 0xFFuL);
    UART0_DLM  = ((UART_BAUDDIVIDE(Baudrate) >> 8) & 0xFFuL);
    UART0_LCR &= ~0x80uL;     // reset DLAB to lock baudrate generator access
    UART0_LCR  = (3uL << 0)   // 8 data bits
               | (0uL << 2)   // 1 stop bit
               | (0uL << 3)   // NO parity
               | (0uL << 4)   // Parity setting (bit 5:4) does not care
               | (0uL << 6)   // Disable Break transmission
               | (0uL << 7);  // Clear DLAB
    UART0_FCR  = (1uL << 0)   // Enable FIFO (according to documentation)
               | (1uL << 1)   // Reset RxFIFO
               | (1uL << 2)   // Reset TxFIFO
               | (0uL << 6)   // Set trigger-level for RxFIFO
               ;
    //
    //  Install OS UART interrupt handler
    //
    OS_ARM_InstallISRHandler(UART_INT_INDEX, &UART0_RX_TX_IRQHandler);  // OS UART interrupt vector.
    OS_ARM_AssignISRSource(UART_INT_INDEX, UART0_INT_SOURCE);
    OS_ARM_EnableISR(UART_INT_INDEX);                                   // Enable UART interrupt vector.
    OS_ARM_EnableISRSource(UART0_INT_SOURCE);                           // Enable UART interrupt source.
    UART0_IER |= (1uL << UART_RX_FULL_INT_ENABLE_SHIFT);                // Enable Rx interrupts
    break;
  case UART1:
    PINSEL0   |= UART1_PINSEL;
    PCONP     |= (1uL << UART1_PCONP_SHIFT);
    UART1_IER  = 0x00uL;      // Initially disable all interrupts
    UART1_LCR  = 0x80uL;      // Set DLAB to initialize Baudrate generator
    UART1_DLL  =  (UART_BAUDDIVIDE(Baudrate)       & 0xFFuL);
    UART1_DLM  = ((UART_BAUDDIVIDE(Baudrate) >> 8) & 0xFFuL);
    UART1_LCR &= ~0x80uL;     // reset DLAB to lock baudrate generator access
    UART1_LCR  = (3uL << 0)   // 8 data bits
               | (0uL << 2)   // 1 stop bit
               | (0uL << 3)   // NO parity
               | (0uL << 4)   // Parity setting (bit 5:4) does not care
               | (0uL << 6)   // Disable Break transmission
               | (0uL << 7);  // Clear DLAB
    UART1_FCR  = (1uL << 0)   // Enable FIFO (according to documentation)
               | (1uL << 1)   // Reset RxFIFO
               | (1uL << 2)   // Reset TxFIFO
               | (0uL << 6)   // Set trigger-level for RxFIFO
               ;
    //
    //  Install OS UART interrupt handler
    //
    OS_ARM_InstallISRHandler(UART_INT_INDEX, &UART1_RX_TX_IRQHandler);  // OS UART interrupt vector.
    OS_ARM_AssignISRSource(UART_INT_INDEX, UART1_INT_SOURCE);
    OS_ARM_EnableISR(UART_INT_INDEX);                                   // Enable UART interrupt vector.
    OS_ARM_EnableISRSource(UART1_INT_SOURCE);                           // Enable UART interrupt source.
    UART1_IER |= (1uL << UART_RX_FULL_INT_ENABLE_SHIFT);                // Enable Rx interrupts
    break;
  }
  OS_INT_DecRI();
}

/*********************************************************************
*
*       BSP_UART_DeInit()
*
*  Function description
*    De-initializes the UART peripheral.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*/
void BSP_UART_DeInit(unsigned int Unit) {
  BSP_UART_USE_PARA(Unit);
}

/*********************************************************************
*
*       BSP_UART_SetBaudrate()
*
*  Function description
*    Configures/changes the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
void BSP_UART_SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  _SetBaudrate(Unit, Baudrate);
}

/*********************************************************************
*
*       BSP_UART_SetReadCallback()
*
*  Function description
*    Sets the callback to execute upon an Rx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetReadCallback(unsigned Unit, BSP_UART_RX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfReadCB = pf;
}

/*********************************************************************
*
*       BSP_UART_SetWriteCallback()
*
*  Function description
*    Sets the callback to execute upon a Tx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetWriteCallback(unsigned int Unit, BSP_UART_TX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfWriteCB = pf;
}

/*********************************************************************
*
*       BSP_UART_Write1()
*
*  Function description
*    Sends one byte via UART.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    Data: (First) data byte to send.
*
*  Additional information
*    The first byte of a transfer is typically sent from application
*    context. Further bytes of the transfer are then sent from the
*    Tx interrupt handler by also calling this function from interrupt
*    context.
*/
void BSP_UART_Write1(unsigned int Unit, unsigned char Data) {
  switch (Unit) {
  case UART0:
    UART0_THR  = Data;                                     // Send character
    UART0_IER |= (1uL << UART_TX_EMPTY_INT_ENABLE_SHIFT);  // Enable Tx interrupts
    break;
  case UART1:
    UART1_THR  = Data;                                     // Send character
    UART1_IER |= (1uL << UART_TX_EMPTY_INT_ENABLE_SHIFT);  // Enable Tx interrupts
    break;
  }
}

/*************************** End of file ****************************/
