/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose: Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Configuration
*/

#ifndef   OS_INIT_PLL
  #define OS_INIT_PLL  (1)  // Init PLL during startup
#endif

//
// Under some circumstances the MAM does not work correctly even in partially enabled mode,
// so we rather disable the MAM. For details see LPC2148 errata sheet (MAM.1 section)
//
#ifndef   OS_INIT_MAM
  #define OS_INIT_MAM  (0)  // Init memory accelerator
#endif

#ifndef   OS_IGNORE_UNDEFINED_INTERRUPT
  #define OS_IGNORE_UNDEFINED_INTERRUPT  (0)
#endif

/*********************************************************************
*
*       Clock settings
*/
#if (OS_INIT_PLL == 1)
  #define OS_PLL_MULTIPLIER  (4)  // PLL multiplies by 4
#else
  #define OS_PLL_MULTIPLIER  (1)  // PLL not used
#endif

#define OS_SYSTEM_FREQ    (48000000u)

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ     (OS_SYSTEM_FREQ / 2)
#define OS_TICK_FREQ      (1000u)
#define OS_INT_FREQ       (OS_TICK_FREQ)
#define OS_TIMER_RELOAD   (OS_TIMER_FREQ / OS_INT_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Memory accelerator settings
*/

#if (OS_INIT_MAM == 1)  // Define timer value for MAM
  #if   (OS_SYSTEM_FREQ < 2000000)
    #define OS_MAMTIM_VALUE    1
  #elif (OS_SYSTEM_FREQ < 4000000)
    #define OS_MAMTIM_VALUE    2
  #else
    #define OS_MAMTIM_VALUE    3
  #endif
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
//
// Port, port mode settings
//
#define PINSEL0                 (*(volatile OS_U32*)0xE002C000u)
//
// Power, reset clock control unit register
//
#define PLLCON                  (*(volatile OS_U32*)0xE01FC080u)
#define PLLCFG                  (*(volatile OS_U32*)0xE01FC084u)
#define PLLSTAT                 (*(volatile OS_U32*)0xE01FC088u)
#define PLLFEED                 (*(volatile OS_U32*)0xE01FC08Cu)
#define PCON                    (*(volatile OS_U32*)0xE01FC0C0u)
#define PCONP                   (*(volatile OS_U32*)0xE01FC0C4u)
#define APBDIV                  (*(volatile OS_U32*)0xE01FC100u)
//
// PLL and memeory timing
//
#define MAMCR                   (*(volatile OS_U32*)(0xE01FC000u))
#define MAMTIM                  (*(volatile OS_U32*)(0xE01FC004u))
//
// Timer sfr definition
//
#define TIM0_BASE_ADDR          (0xE0004000u)
#define TIM1_BASE_ADDR          (0xE0008000u)
#define TIM_IR_OFFS             (0x00)
#define TIM_TCR_OFFS            (0x04)
#define TIM_TC_OFFS             (0x08)
#define TIM_PR_OFFS             (0x0C)
#define TIM_MCR_OFFS            (0x14)
#define TIM_MR0_OFFS            (0x18)
//
// Assign timer sfrs used for OS timer
// initially, we use timer 0
//
#define OS_TIM_BASE_ADDR        TIM0_BASE_ADDR
#define OS_TIM_IR_BIT           (0u)  // Used to reset timer interrupt (Timer IR register), Match channel 0
#define OS_TIMER_INT_SOURCE     (4u)  // Channel number of timer input t0 VIC
#define TIMER_PCONP_BIT         (1u)

#define OS_TIM_IR               (*(volatile OS_U32*)(OS_TIM_BASE_ADDR + TIM_IR_OFFS))
#define OS_TIM_TCR              (*(volatile OS_U32*)(OS_TIM_BASE_ADDR + TIM_TCR_OFFS))
#define OS_TIM_TC               (*(volatile OS_U32*)(OS_TIM_BASE_ADDR + TIM_TC_OFFS))
#define OS_TIM_PR               (*(volatile OS_U32*)(OS_TIM_BASE_ADDR + TIM_PR_OFFS))
#define OS_TIM_MCR              (*(volatile OS_U32*)(OS_TIM_BASE_ADDR + TIM_MCR_OFFS))
#define OS_TIM_MR0              (*(volatile OS_U32*)(OS_TIM_BASE_ADDR + TIM_MR0_OFFS))
//
// Vectored interrupt controller (VIC)
//
#define VIC_BASE_ADDR           (0xFFFFF000u)
#define VIC_IRQSTATUS           (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0000u))
#define VIC_FIQSTATUS           (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0004u))
#define VIC_RAWINTERRUPT        (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0008u))
#define VIC_INTSELECT           (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x000Cu))
#define VIC_INTENABLE           (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0010u))
#define VIC_INTENABLECLEAR      (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0014u))
#define VIC_SOFTINT             (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0018u))
#define VIC_SOFTINTCLEAR        (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x001Cu))
#define VIC_VECTORADDR          (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0030u))
#define VIC_DEFVECTADDR         (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0034u))
#define VIC_VECT_BASE_ADDR      (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0100u))
#define VIC_VECTCNTL_BASE_ADDR  (*(volatile OS_U32*)(VIC_BASE_ADDR + 0x0200u))
#define NUM_INT_SOURCES         (32u)  // Maximum number of sources. Specific devices may support less
#define INT_SOURCE_MASK         (0x1Fu)

#define NUM_INT_VECTORS         (16u)

#define OS_TIMER_INT_INDEX      (15u)      // Assign to lowest priority interrupt channel

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _InitPLL()
*
*  Function description
*    Initialize PLL
*    WARNING: Improper settings may lock CPU
*    Please examine whether init sequence fits your harware configuration
*    We assume a CPU running with external oscillator which is multilied by four
*    which is allowed for a maximum external frequency of 15MHz
*/
static void _InitPLL(void) {
  #if (OS_INIT_PLL == 1)
  PLLCFG  = (OS_PLL_MULTIPLIER - 1)     // => Multiplication value, defaul = 4
          | (1u << 5);                  // => P Value = 2
  PLLCON  = (1u << 0);                  // Enable PLL
  PLLFEED = 0xAA;                       // Update PLL configuration
  PLLFEED = 0x55;
  while ((PLLSTAT & (1u << 10)) == 0);  // wait until PLL is locked (bit 10 is set)
  PLLCON  = (1u << 0) | (1u << 1);      // Enable PLL and select as clock source
  PLLFEED =  0xAAu;                     // Update Pll register
  PLLFEED =  0x55u;
  APBDIV  =  0x02u;                     // Set APB Bus clock to half of CPU clock (CCLK)
  #endif
}

/*********************************************************************
*
*       _SpuriousIrqHandler()
*
*  Function description
*    default spurious IRQ handler
*
*  Notes
*    A spurious interrupt is called under one of the following
*    circumstances:
*    1. An external interrupt source is programmed in level-sensitive mode
*       and an active level occurs for only a short time.
*    2. An internal interrupt source is programmed in level sensitive
*       and the output signal of the corresponding embedded peripheral
*       is activated for a short time. (As in the case for the Watchdog.)
*    3. An interrupt occurs just a few cycles before the software begins
*       to mask it, thus resulting in a pulse on the interrupt source.
*
*    The spurious handler is called to detect these errors.
*/
static void _SpuriousIrqHandler(void) {
#if (OS_DEBUG != 0)
  static OS_U32 _SpuriousIrqCnt;

  _SpuriousIrqCnt++;
#endif
}

/*********************************************************************
*
*       _Init_VIC()
*
*  Function description
*    Clears and disables all interrupt vectors in the VIC.
*    Sets the default vector to handle spurious interrupts.
*/
static void _Init_VIC(void) {
  unsigned int Index;

  VIC_INTENABLECLEAR = 0xFFFFFFFFu;                   // Disable all source interrupts
  VIC_VECTORADDR     = 0;                             // Write to vector, reset VIC
  for (Index = 0; Index < NUM_INT_VECTORS; Index++) {
    OS_ARM_DisableISR(Index);                         // Disable channel
    OS_ARM_AssignISRSource(Index, 0);                 // Assign to hardware channel 0
    OS_ARM_InstallISRHandler(Index, NULL);            // Set vector to 0;
  }
  VIC_SOFTINTCLEAR   = 0xFFFFFFFFu;                   // Disable all soft interrupts
  VIC_INTSELECT      = 0;                             // Set all channels to generate IRQ
  VIC_DEFVECTADDR    = (OS_U32)&_SpuriousIrqHandler;  // In case of spurious interrupts, the default vector is delivered by the VIC
}

/*********************************************************************
*
*       _OS_ISR_Undefined()
*
*  Function description
*    Is called when an not installed interrupt was detected.
*    As interrupt pending condition of peripherals has to be reset,
*    program will not continue when interrupt is ignored.
*/
static void _OS_ISR_Undefined(void) {
#if (OS_IGNORE_UNDEFINED_INTERRUPT == 0)
  volatile int Dummy;

  Dummy = 1;
  while (Dummy > 0) {
    ;  // You may set a breakpoint here to detect non-installed ISR-handler
  }
#endif
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
* Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return OS_TIM_TC;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
* Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return OS_TIM_IR & (1u << OS_TIM_IR_BIT);
}

/*********************************************************************
*
*       _OS_ISR_Tick()
*
*  Function description
*    This is the hardware timer exception handler.
*/
static void _OS_ISR_Tick(void) {
  OS_TIM_IR = (1u << OS_TIM_IR_BIT);  // Clear OS timer interrupt flag
  OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_Process();
#endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  OS_INT_IncDI();
  _Init_VIC();                              // Initialize VIC, clear and disable all interrupts
  //
  // Initialize timer for OS
  //
  PCONP      |=  (1u << TIMER_PCONP_BIT);  // Enable Clock for OS timer
  OS_TIM_TCR &= ~(1u << 0);                // Disable timer
  OS_TIM_IR   =  (1u << OS_TIM_IR_BIT);    // Clear OS timer interrupt flag
  OS_TIM_PR   = 0;                         // Count on every pclk
  OS_TIM_MR0  = OS_TIMER_RELOAD;           // Initialize match register 1
  OS_TIM_MCR  =  (1u << 0)                 // Interrupt on match channel 0
              |  (1u << 1);                // Counter reset on match channel 0
  OS_TIM_TC   = 0;                         // Clear counter
  OS_TIM_TCR |=  (1u << 0);                // Enable timer
  //
  // Enable timer interrupt in interrupt controller
  //
  OS_ARM_InstallISRHandler(OS_TIMER_INT_INDEX, &_OS_ISR_Tick); // Timer/counter interrupt vector.
  OS_ARM_AssignISRSource(OS_TIMER_INT_INDEX, OS_TIMER_INT_SOURCE);
  OS_ARM_EnableISR(OS_TIMER_INT_INDEX);                        // Enable timer/counter 0 interrupt.
  OS_ARM_EnableISRSource(OS_TIMER_INT_SOURCE);                 // Enable timer/counter 0 interrupt.
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       embOS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;

  pISR = (OS_ISR_HANDLER*)VIC_VECTORADDR;    // Get current interrupt handler
  OS_INT_EnterNestable();
  if (pISR != NULL) {
    pISR();                                   // Call installed interrupt service routine
  } else {
    _OS_ISR_Undefined();
  }
  VIC_VECTORADDR = 0;                         // Clear current interrupt pending condition, reset VIC
  OS_INT_LeaveNestable();
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Used to install a specific interrupt vector when ARM
*    CPUs with vectored interrupt controller are used.
*
*  Parameters
*    ISRIndex:    Index of the interrupt source, normally the interrupt vector number.
*    pISRHandler: Address of the interrupt handler function.
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler(int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;
  OS_ISR_HANDLER** papISR;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_VECTORS) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  papISR = (OS_ISR_HANDLER**)&VIC_VECT_BASE_ADDR;
  pOldHandler          = *(papISR + ISRIndex);
  *(papISR + ISRIndex) = pISRHandler;
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_AssignISRSource()
*/
void OS_ARM_AssignISRSource(int ISRIndex, int Source) {
  OS_U32* pCntlRegister;
  OS_U32  ControlReg;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_VECTORS) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  pCntlRegister = (OS_U32*)&VIC_VECTCNTL_BASE_ADDR;
  OS_INT_Disable();
  ControlReg  = *(pCntlRegister + ISRIndex) & ~INT_SOURCE_MASK;
  *(pCntlRegister + ISRIndex) = (ControlReg | (Source & INT_SOURCE_MASK));
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Used to enable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be enabled.
*/
void OS_ARM_EnableISR(int ISRIndex) {
  OS_U32* pCntlRegister;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_VECTORS) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  pCntlRegister = (OS_U32*)&VIC_VECTCNTL_BASE_ADDR;
  OS_INT_Disable();
  *(pCntlRegister + ISRIndex) |= (1u << 5);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Used to disable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be disabled.
*/
void OS_ARM_DisableISR(int ISRIndex) {
  OS_U32* pCntlRegister;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_VECTORS) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  pCntlRegister = (OS_U32*)&VIC_VECTCNTL_BASE_ADDR;
  OS_INT_Disable();
  *(pCntlRegister + ISRIndex) &= ~(1u << 5);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_EnableISRSource()
*/
void OS_ARM_EnableISRSource(int SourceIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)SourceIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  VIC_INTENABLE |= (1u << SourceIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_DisableISRSource()
*/
void OS_ARM_DisableISRSource(int SourceIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)SourceIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  VIC_INTENABLECLEAR = (1u << SourceIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Function description
*    Used to set or modify the priority of a specific interrupt
*    source by programming the interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be modified.
*    Prio:     The priority which should be set for the specific interrupt.
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  //
  // Not supported for LPC 21xx CPU with VIC
  //
  (void)ISRIndex;  // Suppress warning
  (void)Prio;      // Suppress warning
  return 0;
}

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
OS_INTERWORK int __low_level_init(void) {
  _InitPLL();
  #if (OS_INIT_MAM == 1)       // Speedup CPU
    MAMCR  = 0x00u;            // Disable Memory accelerator module
  #if 0
    // Under some circumstances the MAM does not work correctly even in partially enabled mode
    // So we rather disable the MAM.
    // For details S. LPC2148 errata sheet (MAM.1 section)
    MAMTIM = OS_MAMTIM_VALUE;  // Depends on OS_FSYS
    MAMCR  = 0x02;             // Fully enable Memory accelerator module
  #endif
  #endif
  return 1;                    // Always return 1 to enable segment initialization!
}

/*************************** End of file ****************************/
