/***********************************************************************
 * $Id:: lpc313x_cgu_default.c 1515 2008-12-17 00:14:54Z pdurgesh      $
 *
 * Project: LPC313x CGU driver
 *
 * Description:
 *     This file contains driver support for the LPC313x CGU.
 *
 ***********************************************************************
 * Software that is described herein is for illustrative purposes only
 * which provides customers with programming information regarding the
 * products. This software is supplied "AS IS" without any warranties.
 * NXP Semiconductors assumes no responsibility or liability for the
 * use of the software, conveys no license or title under any patent,
 * copyright, or mask work right to the product. NXP Semiconductors
 * reserves the right to make changes in the software without
 * notification. NXP Semiconductors also make no representation or
 * warranty that such application will be suitable for the specified
 * use without further testing or modification.
 *********************************************************************/

#include "lpc313x_cgu_driver.h"


/**
* @file
* @brief The vhClk functions offer a general API for setting domains,
* subdomains, initialisingdefault clock settings, ... . The definition of
* domains, subdomains and their default setting
* can be configured by modifying the domain configuration tables found in this
* file.
*  For a full explanation of capabilities and functionality of the CGU HDLi IP
*  component, please consult  the technical documentation. This file contains
*  a number of structures which configure the pnx0101
* system to a default clock configuration. This configuration contains 3
* components: The definition of the various clock domains and subdomains, the
* input clock for a domain and the default clock speed setting for each of
* these domains and subdomains. While the latter can be adopted at runtime by
* using the appropriate vhClk functions,the former tow are statically
* determined at compile time. It is currently not possible to change the
* domain definitions or input clock definitions at runtime.
* The domain defintion consist of a 3-level table structure:
* <UL>
* <LI>The first level table contains CGU_SB_NR_BASE entries, where
* CGU_SB_NR_BASE is dependent on the CGU HDLi configuration. A first
* level table entry either conatins the starting adress of a second level table
* or NULL when the configuration does not define subdomains for
* that domain. Only domains which have at least one fractional divider assigned
* to it can have a non null entry in the top level domain configuration table.
* <LI>A second level table contains a number of entries that have to equal the
* number of fractional dividers that are assigned to that domain by the CGU
* HDLi configuration. Each entry contains the start adress
* of a 3rd level table.
* <LI>A 3rd level table contains a variable number of entries, where each entry
* is the ID of a clock assigned to the subdomain. The last entry should
* always be the CGU_SB_INVALID_CLK_ID entry, which acts as a delimter character
* for the table. If the subdomain does not contain any clocks, the table
* should have one entry, namely the CGU_SB_INVALID_CLK_ID entry.
* </UL>
* The default clock speed setting consists of a 2-level table structure.
* <UL>
* <LI>The first level table contains VH_CLK_PREDIFINED_CLOCK_MODES entries,
* where VH_CLK_PREDIFINED_CLOCK_MODES is the number of distinct default
* configurations that are defiend to exist in the system. Each entry
* contains the start adress of a 2nd level table.
* <LI>The second level table contains CGU_SB_NR_FRACDIV entries, where
* CGU_SB_NR_FRACDIV is is dependent on the CGU HDLi configuration. Each
* entry is of type fracdivConfiguration_st_t, and contains the default
* fractional divider settings for the subdomain. An entry of 0,0,x denotes that
*  the subdomain does not have a fractional divider assigend to it in default
* mode, and instead the clocks of th subdomain are connected straight to the
* base clock for that domain.
* </UL>
* The domain input clock table contains CGU_SB_NR_BASE entries, where
* CGU_SB_NR_BASE is dependent on the CGU HDLi configuration. Each entry
* is one of the possible base input clocks as defined by the
* CGU HDLi configuration.
*/

/*Below are the default configurations
 * FFAST : 12MHz
 * MASTER PLL Freq: 170MHz;
 * AUDIOPLL Freq:   1024Fs, Fs = 44.1kHz
 *
 * Domain                                 Input           Subdomain          Divider Ratio
 * 0 - CLK_SWITCHBOX_SYS_BASE0            MASTER PLL      Subdomain0         1/2
 *                                                        Subdomain1         1
 *                                                        Subdomain2         1/2
 *                                                        Subdomain3         1/4
 *                                                        Subdomain4         1/4
 *                                                        Subdomain5         1/4
 *                                                        Subdomain6         1/2
 *
 * 1 - CLK_SWITCHBOX_AHB0_APB0_BASE1      FFAST           Subdomain0         1/3
 *                                                        Subdomain1         1/2
 *
 * 2 - CLK_SWITCHBOX_AHB0_APB1_BASE2      FFAST           Subdomain0         1/2
 *                                                        Subdomain1         1/2
 *
 * 3 - CLK_SWITCHBOX_AHB0_APB2_BASE3      MASTER PLL      Subdomain0         1/2
 *                                                        Subdomain1         1/4
 *                                                        Subdomain2         not used
 *
 * 4 - CLK_SWITCHBOX_AHB0_APB2_BASE3      MASTER PLL      Subdomain0         1/2
 *
 * 5 - CLK_SWITCHBOX_IPINT0_BASE5         MASTER PLL      Subdomain0         1/2
 *
 * 6 - CLK_SWITCHBOX_UARTCLK_BASE6        MASTER PLL      Subdomain0         1/2
 *
 * 7 - CLK_SWITCHBOX_CLK1024FS_BASE7      AUDIO PLL       Subdomain0         1/1024
 *                                                        Subdomain1         1/16
 *                                                        Subdomain2         1/4
 *                                                        Subdomain3         1/16
 *                                                        Subdomain4         1/128
 *                                                        Subdomain5         1/8
 *
 * 8 - CLK_SWITCHBOX_DAI_BCK0_BASE8        DAI_BCK0          -                -
 *
 * 9 - CLK_SWITCHBOX_DAI_BCK1_BASE9        DAI_BCK1          -                -
 *
 * 10 - CLK_SWITCHBOX_SPI_BASE10           MASTER PLL       Subdomain0        1/2
 *
 * 11 - CLK_SWITCHBOX_SYSCLK0_BASE11       FFAST              -                 -
 *
 * End of configuration table*/

#define CGU_DEF_DOMAIN0_DIV0	( D0_BIT(CGU_SB_APB0_CLK_ID) | D0_BIT(CGU_SB_APB1_CLK_ID) | \
                               D0_BIT(CGU_SB_APB2_CLK_ID) | D0_BIT(CGU_SB_APB3_CLK_ID) | \
                               D0_BIT(CGU_SB_APB4_CLK_ID) | D0_BIT(CGU_SB_AHB2INTC_CLK_ID) | \
                               D0_BIT(CGU_SB_AHB0_CLK_ID) | D0_BIT(CGU_SB_DMA_PCLK_ID) | \
                               D0_BIT(CGU_SB_DMA_CLK_GATED_ID) | D0_BIT(CGU_SB_NANDFLASH_CTRL_S0_CLK_ID) | \
                               D0_BIT(CGU_SB_NANDFLASH_CTRL_PCLK_ID) | D0_BIT(CGU_SB_ARM926_BUSIF_CLK_ID) | \
                               D0_BIT(CGU_SB_SD_MMC_HCLK_ID) | D0_BIT(CGU_SB_USB_OTG_AHB_CLK_ID) | \
                               D0_BIT(CGU_SB_ISRAM0_CLK_ID) | D0_BIT(CGU_SB_ISRAM1_CLK_ID) | \
                               D0_BIT(CGU_SB_ISROM_CLK_ID) | D0_BIT(CGU_SB_AHB_MPMC_CFG_CLK_ID) | \
                               D0_BIT(CGU_SB_AHB_MPMC_CFG_CLK2_ID) | D0_BIT(CGU_SB_INTC_CLK_ID) )

#define CGU_DEF_DOMAIN0_DIV1	(D0_BIT(CGU_SB_ARM926_CORE_CLK_ID))

#define CGU_DEF_DOMAIN0_DIV2	( D0_BIT(CGU_SB_NANDFLASH_CTRL_AES_CLK_ID) | \
                               D0_BIT(CGU_SB_NANDFLASH_CTRL_NAND_CLK_ID))

#define CGU_DEF_DOMAIN0_DIV3	(D0_BIT(CGU_SB_NANDFLASH_CTRL_ECC_CLK_ID))

#define CGU_DEF_DOMAIN0_DIV4	(D0_BIT(CGU_SB_SD_MMC_CCLK_IN_ID))
#define CGU_DEF_DOMAIN0_DIV5	(D0_BIT(CGU_SB_CLOCK_OUT_ID))
#define CGU_DEF_DOMAIN0_DIV6	(D0_BIT(CGU_SB_EBI_CLK_ID))

#define CGU_DEF_DOMAIN1_DIV7	(D1_BIT(CGU_SB_ADC_CLK_ID))
#define CGU_DEF_DOMAIN1_DIV8	( D1_BIT(CGU_SB_AHB2APB0_ASYNC_PCLK_ID) | \
                               D1_BIT(CGU_SB_EVENT_ROUTER_PCLK_ID) | D1_BIT(CGU_SB_ADC_PCLK_ID) | \
                               D1_BIT(CGU_SB_WDOG_PCLK_ID) | D1_BIT(CGU_SB_IOCONF_PCLK_ID) | \
                               D1_BIT(CGU_SB_CGU_PCLK_ID) | D1_BIT(CGU_SB_SYSCREG_PCLK_ID) | \
                               D1_BIT(CGU_SB_OTP_PCLK_ID) | D1_BIT(CGU_SB_CIC_RNG_PCLK_ID) )

#define CGU_DEF_DOMAIN2_DIV9	( D2_BIT(CGU_SB_AHB2APB1_ASYNC_PCLK_ID) | \
                               D2_BIT(CGU_SB_TIMER0_PCLK_ID) | D2_BIT(CGU_SB_TIMER1_PCLK_ID) | \
                               D2_BIT(CGU_SB_TIMER2_PCLK_ID) | D2_BIT(CGU_SB_TIMER3_PCLK_ID) | \
                               D2_BIT(CGU_SB_PWM_PCLK_ID) | D2_BIT(CGU_SB_PWM_PCLK_REGS_ID) | \
                               D2_BIT(CGU_SB_I2C0_PCLK_ID) | D2_BIT(CGU_SB_I2C1_PCLK_ID) )

#define CGU_DEF_DOMAIN2_DIV10	(D2_BIT(CGU_SB_PWM_CLK_ID))

#define CGU_DEF_DOMAIN3_DIV11	( D3_BIT(CGU_SB_AHB2APB2_ASYNC_PCLK_ID) | \
                                D3_BIT(CGU_SB_CCP_IPINT_PCLK_ID) | D3_BIT(CGU_SB_CCP_IPINT_APB_PCLK_ID) | \
                                D3_BIT(CGU_SB_UART_APB_CLK_ID) | D3_BIT(CGU_SB_LCD_INTERFACE_PCLK_ID) | \
                                D3_BIT(CGU_SB_SPI_PCLK_ID) | D3_BIT(CGU_SB_SPI_PCLK_GATED_ID) )

#define CGU_DEF_DOMAIN3_DIV12	(D3_BIT(CGU_SB_LCD_INTERFACE_LCDCLK_ID))
/* Currently no clocks are connected to this subdomain */
#define CGU_DEF_DOMAIN3_DIV13	(0)

#define CGU_DEF_DOMAIN4_DIV14	( D4_BIT(CGU_SB_AHB2APB3_ASYNC_PCLK_ID) | \
                                D4_BIT(CGU_SB_ADSS_CFG_PCLK_ID) | D4_BIT(CGU_SB_EDGE_DET_PCLK_ID) | \
                                D4_BIT(CGU_SB_SAO1_PCLK_ID) | D4_BIT(CGU_SB_DAO0_PCLK_ID) | \
                                D4_BIT(CGU_SB_SAO2_PCLK_ID) | D4_BIT(CGU_SB_DAO1_PCLK_ID) | \
                                D4_BIT(CGU_SB_SAI1_PCLK_ID) | D4_BIT(CGU_SB_DAI0_PCLK_ID) | \
                                D4_BIT(CGU_SB_SAI2_PCLK_ID) | D4_BIT(CGU_SB_DAI1_PCLK_ID) | \
                                D4_BIT(CGU_SB_SAO3_PCLK_ID) | D4_BIT(CGU_SB_CLASSD_PCLK_ID) )

#define CGU_DEF_DOMAIN5_DIV15	(D5_BIT(CGU_SB_CCP_IPINT_CLK_IP_ID))

#define CGU_DEF_DOMAIN6_DIV16	(0)

#define CGU_DEF_DOMAIN7_DIV17	( D7_BIT(CGU_SB_CR_DAO_WS_ID) | \
                                D7_BIT(CGU_SB_CR_DAO_WS0_ID) | D7_BIT(CGU_SB_CR_DAO_WS1_ID) | \
                                D7_BIT(CGU_SB_CR_DAI_WS0_ID) | D7_BIT(CGU_SB_CR_DAI_WS1_ID) )

#define CGU_DEF_DOMAIN7_DIV18	( D7_BIT(CGU_SB_CR_DAO_BCK0_N_N_ID) | \
                                D7_BIT(CGU_SB_CR_DAO_BCK1_N_N_ID) )

#define CGU_DEF_DOMAIN7_DIV19	( D7_BIT(CGU_SB_CR_DAO_CLK0_ID) | \
                                D7_BIT(CGU_SB_CR_CLK_256FS_ID) )

#define CGU_DEF_DOMAIN7_DIV20	( D7_BIT(CGU_SB_CR_DAI_BCK0_N_N_ID) | \
                                D7_BIT(CGU_SB_CR_DAI_BCK1_N_N_ID) )

#define CGU_DEF_DOMAIN7_DIV21	(D7_BIT(CGU_SB_CLASSD_NS_CLK_ID))
#define CGU_DEF_DOMAIN7_DIV22	(D7_BIT(CGU_SB_CLASSD_DSP_CLK_ID))

#define CGU_DEF_DOMAIN10_DIV23	( D10_BIT(CGU_SB_SPI_IP_CLK_ID) | \
                                 D10_BIT(CGU_SB_SPI_IP_CLK_GATED_ID) )



/* Domain8Subdomain0 and Domain9Subdomain0 are directly connected. So, no entry exists here */
/* Domain11Subdomain0 is directly connected. So, no entry exists here */

/*------------------------------------------------------------------------------------------
  Global variables:
 ------------------------------------------------------------------------------------------*/

const CGU_CLKS_INIT_T g_cgu_default_clks =
{
  /* domain0 [7] - AHB0ClockDomain*/
  {
    CGU_FIN_SELECT_HPPLL1,
    {
      {{1, 1, 3}, CGU_DEF_DOMAIN0_DIV0}, /*FracDiv0     AHB0ClockDomain_SYSBase */
      {{0, 0, 0}, CGU_DEF_DOMAIN0_DIV1}, /*FracDiv1     AHB0ClockDomain_ARM926EJS*/
      {{1, 1, 3}, CGU_DEF_DOMAIN0_DIV2}, /*FracDiv2     AHB0ClockDomain_NANDFlash_CLK */
      {{1, 1, 6}, CGU_DEF_DOMAIN0_DIV3}, /*FracDiv3     AHB0ClockDomain_NANDFlash_ECC */
      {{1, 1, 6}, CGU_DEF_DOMAIN0_DIV4}, /*FracDiv4     AHB0ClockDomain_MMC_CCLK_IN */
      {{1, 1, 6}, CGU_DEF_DOMAIN0_DIV5}, /*FracDiv5     AHB0ClockDomain_clock_out */
      {{1, 1, 3}, CGU_DEF_DOMAIN0_DIV6}  /*FracDiv6     AHB0ClockDomain_EBI_IP2024 */
    }
  },
  /* domain1 [2] - AHB2APB0ClockDomain*/
  {
    CGU_FIN_SELECT_FFAST,
    {
      {{1, 1, 38}, CGU_DEF_DOMAIN1_DIV7},/*FracDiv7     AHB2APB0ClockDomain_ADC_CLK */
      {{1, 1, 2}, CGU_DEF_DOMAIN1_DIV8}  /*FracDiv8     AHB2APB0ClockDomain_Wdog_PCLK */
    }
  },
  /* domain2 [2] - AHB2APB1ClockDomain*/
  {
    CGU_FIN_SELECT_FFAST,
    {
      {{1, 1, 2}, CGU_DEF_DOMAIN2_DIV9}, /*FracDiv9     AHB2APB1ClockDomain_Timer_PCLK */
      {{1, 1, 2}, CGU_DEF_DOMAIN2_DIV10} /*FracDiv10    AHB2APB1ClockDomain_PWM_CLK */
    }
  },
  /* domain3 [3] - AHB2APB2ClockDomain*/
  {
    CGU_FIN_SELECT_HPPLL1,
    {
      {{1, 1, 3}, CGU_DEF_DOMAIN3_DIV11},/*FracDiv11    AHB2APB2ClockDomain_IPINT_PCLK */
      {{1, 1, 40}, CGU_DEF_DOMAIN3_DIV12},/*FracDiv12    AHB2APB2ClockDomain_LCD_CLK */
      {{0, 0, 0}, CGU_DEF_DOMAIN3_DIV13} /*FracDiv13    Not used */
    }
  },
  /* domain4 [1] - AHB2APB3ClockDomain*/
  {
    CGU_FIN_SELECT_FFAST,
    {
      {{1, 1, 2}, CGU_DEF_DOMAIN4_DIV14} /*FracDiv14    AHB2APB3ClockDomain_ADSS */
    }
  },
  /* domain5 [1] - IPINTClockDomain*/
  {
    CGU_FIN_SELECT_HPPLL1,
    {
      {{1, 1, 3}, CGU_DEF_DOMAIN5_DIV15} /*FracDiv15    IPINTClockDomain */
    }
  },
  /* domain6 [1] - UART0ClockDomain*/
  {
    CGU_FIN_SELECT_FFAST,
    {
      {{0, 0, 0}, CGU_DEF_DOMAIN6_DIV16} /*FracDiv16    UARTClockDomain */
    }
  },
  /* domain7 [6] - DAIDAOClockDomain*/
  {
    CGU_FIN_SELECT_HPPLL0,
    {
      {{1, 1, 256}, CGU_DEF_DOMAIN7_DIV17},/*FracDiv17    AudioClockDomain_DAO_WS SubDomain */
      {{1, 1, 4}, CGU_DEF_DOMAIN7_DIV18},  /*FracDiv18    AudioClockDomain_DAO_BCK SubDomain*/
      {{0, 0, 0}, CGU_DEF_DOMAIN7_DIV19}, /*FracDiv19    AudioClockDomain_CR_DAO_CLK SubDomain */
      {{1, 1, 4}, CGU_DEF_DOMAIN7_DIV20},  /*FracDiv20    AudioClockDomain_DAI_BCK SubDomain */
      {{1, 1, 32}, CGU_DEF_DOMAIN7_DIV21}, /*FracDiv21    AudioClockDomain_CLASSD_NS Subdomain */
      {{1, 1, 2}, CGU_DEF_DOMAIN7_DIV22}  /*FracDiv22    AudioClockDomain_CLASSD_DSP_CLK Subdomain */
    }
  },
  /* domain8  - DAIBCK0ClockDomain*/
  {
    CGU_FIN_SELECT_XT_DAI_BCK0
  },

  /* domain9  - DAIBCK1ClockDomain*/
  {
    CGU_FIN_SELECT_XT_DAI_BCK1
  },
  /* domain10 [1] - SPIClockDomain*/
  {
    CGU_FIN_SELECT_HPPLL1,
    {
      {{1, 1, 3}, CGU_DEF_DOMAIN10_DIV23} /*FracDiv23    SPIClockDomain*/
    }
  },
  /* domain11  - SYSClockDomain*/
  {
    CGU_FIN_SELECT_FFAST
  },

  /* dyn_fdiv_cfg[7]  */
  {
    /*For DYN_SEL, the bits below are all configured as 1.*/
    /*
       bit 	  Name							    Description:

    	0	simple_dma_trans[1]					simple_dma transfers can enable high-speed
    	1	simple_dma_ready					simple_dma last transfers can enable high-speed
    	2	arm926_901616_lp_i_trans[1]			arm926ejs_i transfers can enable high-speed
    	3	arm926_901616_lp_i_ready			arm926ejs_i_ready last transfers can enable high-speed
    	4	arm926_901616_lp_d_trans[1]			arm926ejs_d transfers can enable high-speed
    	5	arm926_901616_lp_d_ready			arm926ejs_d transfers can enable high-speed
    	6	usb_otg_mst_trans[1]				USB OTG transfers can enable high spee
    	7	usb_otg_mst_ready					USB OTG last transfers can enable high-speed
    	8	ahb_mpmc_pl172_refresh_hispeed_req	External SDRAM refresh generator  transfers can enable high speed
    */
    {
      0x1FF,
      {1, 1, 64} /*DynFD0*/
    },
    {
      0x1FF,
      {0, 0, 0}	/*DynFD1*/
    },
    {
      0x1FF,
      {1, 1, 3}	/*DynFD2*/
    },
    {
      0x1FF,
      {1, 1, 6}	/*DynFD3*/
    },
    {
      0x1FF,
      {1, 1, 6}	/*DynFD4*/
    },
    {
      0x1FF,
      {1, 1, 6}	/*DynFD5*/
    },
    {
      0x1FF,
      {1, 1, 3}	/*DynFD6*/
    }
  }
};


