/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose: Initializes and handles the hardware for embOS
*/

#include "RTOS.h"
#include "lpc313x_intc_driver.h"
#include "lpc313x_timer_driver.h"
#include "lpc313x_cgu_driver.h"
#include "ea3131_board.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Configuration
*/

#ifndef   OS_IGNORE_UNDEFINED_INTERRUPT
  #define OS_IGNORE_UNDEFINED_INTERRUPT  (0)
#endif

/*********************************************************************
*
*       System tick settings
*/
#define OS_TICK_FREQ  (1000u)
#define OS_INT_FREQ   (OS_TICK_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/

#define INTC_BASE_ADDR   (0x60000000u)
#define OFF_IRQ_STATUS   (0x0100u)
#define INTC_IRQ_STATUS  (*(volatile OS_U32*)(INTC_BASE_ADDR + OFF_IRQ_STATUS))

/*********************************************************************
*
*       Interrupt handler table
*/

extern PFV g_irq_func_ptrs[IRQ_END_OF_INTERRUPTS + 1];

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static int _OSFreq;
static int _Timer0Handle;  // Timer device handle

/*********************************************************************
*
*       MMU and cache configuration
*
*  Notes
*    The translation table has to be aligned on 16384 bytes boundary
*    The translation table has to be located in un-initialized memory,
*    because it may be filled from startup, before segments and variables
*    are initialized.
*/
#if   (defined(__ICCARM__))  // IAR
  #pragma data_alignment=16384
  static __no_init OS_U32 _TranslationTable[0x1000];
#elif (defined(__CC_ARM))
  #pragma arm section zidata = "NO_INIT"  // Has to be declared as separate section UNINIT in the scatter file
  __align(16384) static OS_U32 _TranslationTable[0x1000];
  #pragma arm section zidata              // back to default (.bss section)
#elif (defined(__GNUC__))   // GCC
  static OS_U32 _TranslationTable[0x1000] __attribute__ ((aligned (16384), section (".no_init")));
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _OS_ISR_Undefined()
*
*  Function description
*    Is called when an not installed interrupt was detected.
*    As interrupt pending condition of peripherals has to be reset,
*    program will not continue when interrupt is ignored.
*/
static void _OS_ISR_Undefined(int Index) {
#if (OS_IGNORE_UNDEFINED_INTERRUPT == 0)
  volatile int Dummy;

  Dummy = Index;
  while (Dummy > 0) {
    ;  // You may set a breakpoint here to detect non-installed ISR-handler
  }
#endif
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
* Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return TIMER_CNTR0->value;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
* Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return INTC->pending & (1 << IRQ_TIMER0);
}

/*********************************************************************
*
*       _OS_ISR_Tick()
*
*  Function description
*    This is the hardware timer exception handler.
*/
static void _OS_ISR_Tick(void) {
  timer_ioctl(_Timer0Handle, TMR_CLEAR_INTS, 0);  // Clear interrupt flag
  OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_Process();
#endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_SYSTIMER_CONFIG SysTimerConfig = {_OSFreq, OS_INT_FREQ, OS_TIMER_DOWNCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

  OS_INT_IncDI();
  OS_ARM_CACHE_Sync();  // Ensure caches are synchronized
  //
  // Setup clocks and memory
  //
  {
    UNS_32 clkin_freq[CGU_FIN_SELECT_MAX];

    //
    // Init clock frequencies
    //
    clkin_freq[0] = XTAL_IN;  // CGU_FIN_SELECT_FFAST
    clkin_freq[1] = 0;        // CGU_FIN_SELECT_XT_DAI_BCK0
    clkin_freq[2] = 0;        // CGU_FIN_SELECT_XT_DAI_WS0
    clkin_freq[3] = 0;        // CGU_FIN_SELECT_XT_DAI_BCK1
    clkin_freq[4] = 0;        // CGU_FIN_SELECT_XT_DAI_WS1
    clkin_freq[5] = 0;        // CGU_FIN_SELECT_HPPLL0
    clkin_freq[6] = 0;        // CGU_FIN_SELECT_HPPLL1
    //
    // Init CGU driver
    //
    cgu_init(clkin_freq);
  }
  //
  // Initialize timer for OS
  //
  _Timer0Handle = timer_open((void *)TIMER_CNTR0, 0);  // Open timers - this will enable the clocks for the timers
  //
  // Setup timer 0
  //
  timer_ioctl(_Timer0Handle, TMR_SET_PERIODIC_MODE, 1);  // Set timer periodic mode
  timer_ioctl(_Timer0Handle, TMR_SET_MSECS, 1);          // Set timer latch value
  timer_ioctl(_Timer0Handle, TMR_CLEAR_INTS, 0);         // Clear pending interrupts
  timer_ioctl(_Timer0Handle, TMR_ENABLE, 1);             // Start timer
  //
  // Enable timer interrupt in interrupt controller
  //
  int_initialize(0xFFFFFFFFu);                        // Initialize interrupt system
  int_install_irq_handler(IRQ_TIMER0, _OS_ISR_Tick);  // Install systick interrupt
  int_enable(IRQ_TIMER0);                             // Enable systick interrupt
  //
  // Save PCLK freq. for later reference
  //
  _OSFreq = cgu_get_clk_freq(CGU_SB_TIMER0_PCLK_ID);
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       embOS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_U32 IntSrc;

  IntSrc = INTC_IRQ_STATUS;
  IntSrc >>= 3;  // Ignore lsb 3 bits
  OS_INT_Enter();
  if (g_irq_func_ptrs[IntSrc]) {
    g_irq_func_ptrs[IntSrc]();         // Call interrupt service routine
  } else {
    _OS_ISR_Undefined(IntSrc);
  }
  OS_INT_Leave();
}

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
void ea3131_init(void);

#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
OS_INTERWORK int __low_level_init(void) {
  //
  // Disable Interrupts; should normally not be necessary, but may be helpful for some debugger
  //
  __disable_interrupt();
  ea3131_init();
  //
  // Init MMU and caches. This defines the virtual memory map, which is used during execution.
  // Memory mapping should be complete, meaning 4096 entries.
  // Code below fills in ascending VAddr order
  //
  //                                             Mode                    VAddr  PAddr  Size[MB]
#if (!defined(SDRAM_TARGET) || defined(SDRAM_SPI_NOR))
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,   0x000, 0x000, 0x001);    // 4k Byte shadowed from internal SRAM0 for vectors, cachable, bufferable
#else
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,   0x000, 0x300, 0x001);    // 4k Byte shadowed from external SDRAM0 for vectors, cachable, bufferable
#endif
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_MMU_UNMAPPED,    0x001, 0x001, 0x10F);    // Unmapped area
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,   0x110, 0x110, 0x001);    // Internal SRAM0 and SRAM1, original address, cachable, bufferable
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_MMU_UNMAPPED,    0x111, 0x111, 0x01F);    // Unmapped area
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x130, 0x130, 0x0D0);    // Peripherals, NON cachable, NON bufferable
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_MMU_UNMAPPED,    0x200, 0x200, 0x100);    // Unmapped area
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,   0x300, 0x300, 0x040);    // 64MB SDRAM, original address, cachable, bufferable
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_MMU_UNMAPPED,    0x340, 0x340, 0x2C0);    // Unmapped area
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x600, 0x600, 0x001);    // Interrupt controller, NON cachable, NON bufferable
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_MMU_UNMAPPED,    0x601, 0x601, 0x0FF);    // Unmapped area
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB, 0x700, 0x700, 0x001);    // NAND buffer, NON cachable, NON bufferable
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_MMU_UNMAPPED,    0x701, 0x701, 0x8FF);    // Unmapped area
  OS_ARM_MMU_Enable      (&_TranslationTable[0]);
  OS_ARM_ICACHE_Enable();
  OS_ARM_DCACHE_Enable();
  return 1;  // Always return 1 to enable segment initialization!
}

/*************************** End of file ****************************/
