/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose: Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Configuration
*/

#ifndef   OS_USE_MS_TIMER
  #define OS_USE_MS_TIMER  (0)  // The MS timer does not allow high precision time measurement
#endif

#ifndef   OS_INIT_MMU
  #define OS_INIT_MMU  (1)
#endif

#ifndef   OS_IGNORE_UNDEFINED_INTERRUPT
  #define OS_IGNORE_UNDEFINED_INTERRUPT  (0)
#endif

/*********************************************************************
*
*       Clock settings
*/
#define OS_SYSTEM_FREQ     (208000000u)

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ     (OS_SYSTEM_FREQ / 16)
#define OS_TICK_FREQ      (1000u)
#define OS_INT_FREQ       (OS_TICK_FREQ)
#define OS_TIMER_RELOAD   (OS_TIMER_FREQ / OS_INT_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
//
// Master Interrupt Controller
//
#define MIC_BASE_ADDR    (0x40008000u)
#define MIC_ER           (*(volatile unsigned int*)(MIC_BASE_ADDR + 0x00u))
#define MIC_SR           (*(volatile unsigned int*)(MIC_BASE_ADDR + 0x08u))
#define MIC_APR          (*(volatile unsigned int*)(MIC_BASE_ADDR + 0x0Cu))
#define MIC_ATR          (*(volatile unsigned int*)(MIC_BASE_ADDR + 0x10u))
#define MIC_ITR          (*(volatile unsigned int*)(MIC_BASE_ADDR + 0x14u))
//
// Slave Interrupt Controller 1
//
#define SIC1_BASE_ADDR   (0x4000C000u)
#define SIC1_ER          (*(volatile unsigned int*)(SIC1_BASE_ADDR + 0x00u))
#define SIC1_SR          (*(volatile unsigned int*)(SIC1_BASE_ADDR + 0x08u))
#define SIC1_ITR         (*(volatile unsigned int*)(SIC1_BASE_ADDR + 0x14u))
//
// Slave Interrupt Controller 2
//
#define SIC2_BASE_ADDR   (0x40010000u)
#define SIC2_ER          (*(volatile unsigned int*)(SIC2_BASE_ADDR + 0x00u))
#define SIC2_SR          (*(volatile unsigned int*)(SIC2_BASE_ADDR + 0x08u))
#define SIC2_ITR         (*(volatile unsigned int*)(SIC2_BASE_ADDR + 0x14u))

#if (OS_USE_MS_TIMER == 1)
//
// Millisecond Timer
//
#define MSTIM_BASE_ADDR  (0x40034000u)
#define MSTIM_INT        (*(volatile unsigned int*)(MSTIM_BASE_ADDR + 0x00u))
#define MSTIM_CTRL       (*(volatile unsigned int*)(MSTIM_BASE_ADDR + 0x04u))
#define MSTIM_CNT        (*(volatile unsigned int*)(MSTIM_BASE_ADDR + 0x08u))
#define MSTIM_MCTRL      (*(volatile unsigned int*)(MSTIM_BASE_ADDR + 0x14u))
#define MSTIM_MATCH0     (*(volatile unsigned int*)(MSTIM_BASE_ADDR + 0x18u))
#else
//
// High speed timer
//
#define HSTIM_BASE_ADDR  (0x40038000u)
#define HSTIM_INT        (*(volatile unsigned int*)(HSTIM_BASE_ADDR + 0x00u))
#define HSTIM_CTRL       (*(volatile unsigned int*)(HSTIM_BASE_ADDR + 0x04u))
#define HSTIM_COUNTER    (*(volatile unsigned int*)(HSTIM_BASE_ADDR + 0x08u))
#define HSTIM_PMATCH     (*(volatile unsigned int*)(HSTIM_BASE_ADDR + 0x0Cu))
#define HSTIM_PCOUNT     (*(volatile unsigned int*)(HSTIM_BASE_ADDR + 0x10u))
#define HSTIM_MCTRL      (*(volatile unsigned int*)(HSTIM_BASE_ADDR + 0x14u))
#define HSTIM_MATCH0     (*(volatile unsigned int*)(HSTIM_BASE_ADDR + 0x18u))
//
// Timer clock control register
//
#define TIMCLK_CTRL      (*(volatile unsigned int*)(0x400040BCu))
#endif
//
// HCLK Divider Control Register
//
#define HCLKDIV_CTRL     (*(volatile unsigned int*)(0x40004040u))
//
// HCLK PLL Control Register
//
#define HCLKPLL_CTRL     (*(volatile unsigned int*)(0x40004058u))
//
// Power Control Register
//
#define PWR_CTRL         (*(volatile unsigned int*)(0x40004044u))
//
// Timer interrupt assignment
//

#if (OS_USE_MS_TIMER == 1)
  #define OS_TIMER_INT_ID (27u)        // MS timer is source 27 on first controller
  #define OS_TIM_INT_CLR  (1 << 0)     // Clear match0 interrupt
  #define OS_TIM_INT      (MSTIM_INT)  // MS timer interrupt control register
#else
  #define OS_TIMER_INT_ID (5u)         // HS timer is source 5 on first controller
  #define OS_TIM_INT_CLR  (1 << 0)     // Clear match0 interrupt
  #define OS_TIM_INT      (HSTIM_INT)  // HS timer interrupt control register
#endif

#define NUM_INT_SOURCES   (3 * 32)     // 3 interrupt controller

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

#if (OS_INIT_MMU != 0)
/*********************************************************************
*
*       MMU and cache configuration
*
*  Notes
*    The translation table has to be aligned on 16384 bytes boundary
*    The translation table has to be located in un-initialized memory,
*    because it may be filled from startup, before segments and variables
*    are initialized.
*/
#if   (defined(__ICCARM__))  // IAR
  #pragma data_alignment=16384
  static __no_init OS_U32 _TranslationTable[0x1000];
#elif (defined(__CC_ARM))
  #pragma arm section zidata = "NO_INIT"  // Has to be declared as separate section UNINIT in the scatter file
  __align(16384) static OS_U32 _TranslationTable[0x1000];
  #pragma arm section zidata              // back to default (.bss section)
#elif (defined(__GNUC__))    // GCC
  static OS_U32 _TranslationTable[0x1000] __attribute__ ((aligned (16384), section (".no_init")));
#endif
#endif // OS_INIT_MMU

/*********************************************************************
*
*       Interrupt vector table
*
*  Notes
*    The translation table has to be located in un-initialized memory,
*    because it may be filled from startup, before segments and variables
*    are initialized.
*/
#if   (defined(__ICCARM__))  // IAR
  #pragma data_alignment=4
  static __no_init OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES];
#elif (defined(__CC_ARM))
  #pragma arm section zidata = "NO_INIT"  // Has to be declared as separate section UNINIT in the scatter file
  __align(4) static OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES];
  #pragma arm section zidata              // back to default (.bss section)
#elif (defined(__GNUC__))    // GCC
  static OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES] __attribute__ ((aligned (4), section (".no_init")));
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _InitPLL()
*
*  Function description
*    Initialize PLL
*    WARNING: Improper settings may lock CPU
*    Please examine whether init sequence fits your hardware configuration
*    We assume a CPU running with external oscillator which is multiplied by four
*    which is allowed for a maximum external frequency of 15MHz
*/
static void _InitPLL(void) {
  //
  // Set Clk dividers
  //
  HCLKDIV_CTRL  = 0
                | (1u       << 0)     // HCLK (AHB clock) = 1/2  Pll_clk_out. MUST not excced 104 MHz
                | ((16 - 1) << 2)     // Peripheral clock divider: 1/16 Pll_clk_out
                ;
  //
  // PLL Init - OSC * 16 = 208MHz     // External crystal is 13MHz
  //
  HCLKPLL_CTRL  = 0
                | ( 0u << 15)         // OSC connected to PLL input
                | ( 1u << 14)         // CCO connected to PLL_Clk output
                | ( 0u << 13)         // CC0 connected to N divider input
                | ( 0u <<  9)         // Set divider
                | (16u <<  1)         // Set multiplier
                ;
  HCLKPLL_CTRL |= (1u << 16);         // Enable Pll
  while ((HCLKPLL_CTRL & 1) == 0) {   // Wait until PLL locks
    ;
  }
  PWR_CTRL     |= (1u << 2);          // Switch to PLL clock
}

/*********************************************************************
*
*       _InitMIC()
*
*  Function description
*    Clears and disables all interrut sources in Interrupt controller.
*/
static void _InitMIC(void) {
  MIC_ER   = 0x00;    // Disable all interrupts of main controller
  SIC1_ER  = 0x00;    // Disable all interrupts of slave controller 1
  SIC2_ER  = 0x00;    // Disable all interrupts of slave controller 2
  MIC_ITR  = 0x00;    // Assign all inputs to IRQ
  SIC1_ITR = 0x00;    // Assign all inputs to IRQ
  SIC2_ITR = 0x00;    // Assign all inputs to IRQ
  MIC_ER   = (1u << 31) | (1u << 30) | (1u << 1) | (1u << 0);  // Enable interrupts from slave controller
}

/*********************************************************************
*
*       _ISRMaskToIndex()
*
*  Function description
*    Calculates index of highest pending interrupt from Mask
*/
static OS_U32 _ISRMaskToIndex(OS_U32 Mask) {
  OS_U32 Index;

  Index = 0;
  if (Mask & 0xFFFF0000u) {
    Index += 16u;
    Mask >>= 16u;
  }
  if (Mask & 0xFF00u) {
    Index += 8u;
    Mask >>= 8u;
  }
  if (Mask & 0xF0u) {
    Index += 4u;
    Mask >>= 4u;
  }
  if (Mask & 0xCu) {
    Index += 2u;
    Mask >>= 2u;
  }
  if (Mask & 0x2u) {
    Index += 1u;
  }
  return Index;
}

/*********************************************************************
*
*       _OS_ISR_Undefined()
*
*  Function description
*    Is called when an not installed interrupt was detected.
*    As interrupt pending condition of peripherals has to be reset,
*    program will not continue when interrupt is ignored.
*/
static void _OS_ISR_Undefined(int Index) {
#if (OS_IGNORE_UNDEFINED_INTERRUPT == 0)
  volatile int Dummy;

  Dummy = Index;
  while (Dummy > 0) {
    ;  // You may set a breakpoint here to detect non-installed ISR-handler
  }
#endif
  OS_USEPARA(Index);
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
* Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
#if (OS_USE_MS_TIMER == 0)
  return HSTIM_COUNTER;
#else
  #warning Needs to be implemented!
  return 0;
#endif
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
* Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
#if (OS_USE_MS_TIMER == 0)
  return HSTIM_INT & (1uL << 0);
#else
  #warning Needs to be implemented!
  return 0;
#endif
}

/*********************************************************************
*
*       _OS_ISR_Tick()
*
*  Function description
*    This is the hardware timer exception handler.
*/
static void _OS_ISR_Tick(void) {
  OS_TIM_INT |= OS_TIM_INT_CLR;  // Clear interrupt flag
  OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_Process();
#endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  OS_INT_IncDI();
  //
  // Initialize timer for OS
  //
#if OS_USE_MS_TIMER
  //
  // Init MS Timer
  //
  MSTIM_CTRL  = 2;                       // Reset and stop timer counters
  MSTIM_MCTRL = (1u << 1)                // Reset timer counter on match 0
              | (1u << 0)                // Enable interrupt by match 0
              ;
  MSTIM_MATCH0 = 32768u / 1000;          // Set timer period
  MSTIM_CTRL   = 1;                      // Enable milisecond timer and release reset
#else
  //
  // Init HS Timer running at peripheral clock
  //
  TIMCLK_CTRL   |= (1u << 1);            // Enable clock for High speed timer peripheral
  HSTIM_CTRL     = 0;                    // Stop timer
  HSTIM_COUNTER  = 0;                    // Initialize couter
  HSTIM_MATCH0   = OS_TIMER_RELOAD;      // Initialize Compare match register
  HSTIM_MCTRL    = 0
                 | (1u << 0)             // Enable interrupt on match 0
                 | (1u << 1)             // Clear Timer on match 0
                 ;
  HSTIM_CTRL     = 1;                    // Enable clock, start timer
#endif
  //
  // Setup interrupt controller for timer interrupt
  //
  MIC_APR |=  (1u << OS_TIMER_INT_ID);    // On a low level signal
  MIC_ATR &= ~(1u << OS_TIMER_INT_ID);    // Set level sensetive interrupt
  MIC_ITR &= ~(1u << OS_TIMER_INT_ID);    // Assign to IRQ line
  //
  // Enable timer interrupt in interrupt controller
  //
  OS_ARM_InstallISRHandler (OS_TIMER_INT_ID, _OS_ISR_Tick);
  OS_ARM_EnableISR(OS_TIMER_INT_ID);
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       embOS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;
  OS_U32          ISRIndex;
  OS_U32          ISRMask;
  OS_U32          ISRPending;

  ISRMask = MIC_SR & 0x3FFFFFFFu;       // Examine current interrupt source, ignore FIQs
  if (ISRMask & (1u << 0)) {            // Slave 1 pending ?
    ISRMask    = SIC1_SR;
    ISRIndex   = _ISRMaskToIndex(ISRMask) + 32;
    ISRPending = (1u << 0);
  } else if (ISRMask & (1u << 1)) {     // Slave 2 pending ?
    ISRMask    = SIC2_SR;
    ISRIndex   = _ISRMaskToIndex(ISRMask) + 64;
    ISRPending = (1u << 1);
  } else {
    ISRIndex   = _ISRMaskToIndex(ISRMask);
    ISRPending = (1u << ISRIndex);
  }
  MIC_ER &= ~ISRPending;                // Disable pending interrupt source to avoid recursion
  OS_INT_EnterNestable();
  pISR = _apIsrHandler[ISRIndex];       // Read interrupt vector from Table
  if (pISR != NULL) {
    pISR();                             // Call installed interrupt service routine
  } else {
    _OS_ISR_Undefined(ISRIndex);
  }
  MIC_ER |= ISRPending;                 // Re-enable interrupt source
  OS_INT_LeaveNestable();
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Used to install a specific interrupt vector when ARM
*    CPUs with vectored interrupt controller are used.
*
*  Parameters
*    ISRIndex:    Index of the interrupt source, normally the interrupt vector number.
*    pISRHandler: Address of the interrupt handler function.
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler(int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  pOldHandler             = _apIsrHandler[ISRIndex];
  _apIsrHandler[ISRIndex] = pISRHandler;
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Used to enable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be enabled.
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  if (ISRIndex < 32) {         // Main interrupt controller affected ?
    MIC_ER |= (1u << ISRIndex);
  } else if (ISRIndex < 64) {  // Slave 1 affected?
    ISRIndex -= 32;
    SIC1_ER |= (1u << ISRIndex);
  } else {                     // Slave 2
    ISRIndex -= 64;
    SIC2_ER |= (1u << ISRIndex);
  }
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Used to disable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be disabled.
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  if (ISRIndex < 32) {         // Main interrupt controller affected ?
    MIC_ER &= ~(1u << ISRIndex);
  } else if (ISRIndex < 64) {  // Slave 1 affected?
    ISRIndex -= 32;
    SIC1_ER &= ~(1u << ISRIndex);
  } else {                     // Slave 2
    ISRIndex -= 64;
    SIC2_ER &= ~(1u << ISRIndex);
  }
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Function description
*    Used to set or modify the priority of a specific interrupt
*    source by programming the interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be modified.
*    Prio:     The priority which should be set for the specific interrupt.
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  //
  // Individual interrupt priorities are not supported by MIC
  //
  OS_USEPARA(ISRIndex);  // Suppress warning
  OS_USEPARA(Prio);      // Suppress warning
  return 0;
}

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);  // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
OS_INTERWORK int __low_level_init(void) {
  _InitPLL();
  //
  // Initialize MMU and caches
  //
#if (OS_INIT_MMU == 1)
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,     0x000, 0x080, 0x001);    // Internal SRAM mapped to zero to map vectors to correct address
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB,   0x001, 0x001, 0x07F);    // Range from first bank to internal SRAM: non cacheable, non bufferable
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_C_B,     0x080, 0x008, 0x001);    // Internal SRAM, original addresss, cahceable, bufferable.
  OS_ARM_MMU_AddTTEntries(&_TranslationTable[0], OS_ARM_CACHEMODE_NC_NB,   0x081, 0x081, 0x7FF);    // Remaining area, memory, sfrs... non cacheable, non bufferable
  OS_ARM_MMU_Enable      (&_TranslationTable[0]);
#endif
  //
  // Always enable I-Cache
  //
  OS_ARM_ICACHE_Enable();
  //
  // D-Cache must not be enabled as long as MMU is not initialized
  //
#if OS_INIT_MMU
  OS_ARM_DCACHE_Enable();
#endif
  //
  // Initialize interrupt controller
  //
  _InitMIC();
  return 1;  // Always return 1 to enable segment initialization!
}

/*************************** End of file ****************************/
