/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : BSP for Segger J-Trace Cortex-M Pro V1 hardware
*/

#include "BSP.h"
#include "Zynq7010.h"

#define U32 unsigned int

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/
#define LED0_BIT      (50u - 32u)       // LED0: Target power RED   => MIO 50
#define LED1_BIT      (49u - 32u)       // LED1: Target power GREEN => MIO 49
#define LED2_BIT      (48u - 32u)       // LED2: Trace RED          => MIO 48
#define LED3_BIT      (47u - 32u)       // LED3: Trace GREEN        => MIO 47
#define LED4_BIT      (46u - 32u)       // LED4: USB RED            => MIO 46
#define LED5_BIT      (45u - 32u)       // LED5: USB GREEN          => MIO 45
#define LED6_BIT      (44u - 32u)       // LED6: RED                => MIO 44
#define LED0_MASK     (1u << LED0_BIT)
#define LED1_MASK     (1u << LED1_BIT)
#define LED2_MASK     (1u << LED2_BIT)
#define LED3_MASK     (1u << LED3_BIT)
#define LED4_MASK     (1u << LED4_BIT)
#define LED5_MASK     (1u << LED5_BIT)
#define LED6_MASK     (1u << LED6_BIT)
#define LED_MASK_ALL  (LED0_MASK | LED1_MASK | LED2_MASK | LED3_MASK | LED4_MASK | LED5_MASK | LED6_MASK)
#define NUM_LEDS      (7)

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_Init()
*/
void BSP_Init(void) {
  U32 v;
  U32 v1;

  //
  //  LEDs on J-Trace are special as our LED pins are in a 1.8V domain
  //  Meaning we only drive LOW (LED active) and to switch them off, we disable our output and let the 3.3V pull-up do its work
  //
  REG_SLCR_UNLOCK   = SCLR_UNLOCK_KEY;
  //
  // Init LEDs
  // LEDs are in 1.8V domain
  //
  REG_SLCR_MIO_PIN_44 = MIO_PIN_OPT_IO_CMOS_1V8 | MIO_PIN_OPT_LVL30_MUX;  // LED6
  REG_SLCR_MIO_PIN_45 = MIO_PIN_OPT_IO_CMOS_1V8 | MIO_PIN_OPT_LVL30_MUX;  // LED5
  REG_SLCR_MIO_PIN_46 = MIO_PIN_OPT_IO_CMOS_1V8 | MIO_PIN_OPT_LVL30_MUX;  // LED4
  REG_SLCR_MIO_PIN_47 = MIO_PIN_OPT_IO_CMOS_1V8 | MIO_PIN_OPT_LVL30_MUX;  // LED3
  REG_SLCR_MIO_PIN_48 = MIO_PIN_OPT_IO_CMOS_1V8 | MIO_PIN_OPT_LVL30_MUX;  // LED2
  REG_SLCR_MIO_PIN_49 = MIO_PIN_OPT_IO_CMOS_1V8 | MIO_PIN_OPT_LVL30_MUX;  // LED1
  REG_SLCR_MIO_PIN_50 = MIO_PIN_OPT_IO_CMOS_1V8 | MIO_PIN_OPT_LVL30_MUX;  // LED0
  v = LED_MASK_ALL;
  //
  // CAUTION: Writing the DATA registers before setting DIRM to output results in chip to ignore write data
  //          Correct sequence to configure pin for output without glitches:
  //          Disable output driver via OEN
  //          Set direction to output via DIRM
  //          Write output data via DATA register
  //          Enable output driver via OEN
  //
  REG_MIO_OEN_1  &= ~v;  // Disable output of LED pins (LEDs switched off by default)
  REG_MIO_DIRM_1 |=  v;  // 1 == Configure pin as output
  v1 = ~((v >> 0) & 0xFFFF);
  REG_MIO_MASK_DATA_1_LSW = (v1 << 16) | 0;      // Higher 16 bits specify a mask for which pins the write to the MD register has an effect. 0-bit means that the pin is affected
  v1 = ~((v >> 16) & 0xFFFF);
  REG_MIO_MASK_DATA_1_MSW = (v1 << 16) | 0;      // Higher 16 bits specify a mask for which pins the write to the MD register has an effect. 0-bit means that the pin is affected
}

/*********************************************************************
*
*       BSP_SetLED()
*/
void BSP_SetLED(int Index) {
  if (Index < NUM_LEDS) {
    REG_MIO_OEN_1 |= (1u << (LED0_BIT - Index));
  }
}

/*********************************************************************
*
*       BSP_ClrLED()
*/
void BSP_ClrLED(int Index) {
  if (Index < NUM_LEDS) {
    REG_MIO_OEN_1 &= ~(1uL << (LED0_BIT - Index));
  }
}

/*********************************************************************
*
*       BSP_ToggleLED()
*/
void BSP_ToggleLED(int Index) {
  if (Index < NUM_LEDS) {
    if (REG_MIO_OEN_1 & (1uL << (LED0_BIT - Index))) {
      REG_MIO_OEN_1 &= ~(1uL << (LED0_BIT - Index));
    } else {
      REG_MIO_OEN_1 |= (1u << (LED0_BIT - Index));
    }
  }
}

/*************************** End of file ****************************/
