/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
File    : RTOSInit_XC7Z010.c
Purpose : Initializes and handles the hardware for embOS

Literature:
  [1]  Generic Interrupt Controller (GIC-390) Errata Notice
       (\\fileserver\Techinfo\Company\ARM\GIC_GenericInterruptController\GIC_390_Errata_Notice_v6.pdf)
  [2]  Zynq-7000 All Programmable SoC Technical Reference Manual
       (\\fileserver\Techinfo\Company\Xilinx\Zynq_7000\ug585-Zynq-7000-TRM-v1.10.pdf)
*/

#include "RTOS.h"
#include "Zynq7010.h"

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/

#ifndef   DDR_DISABLE_WC
  #define DDR_DISABLE_WC    (1)  // Disable write combine of DDR controller
#endif

#ifndef   DDR_USE_INT_VREF
  #define DDR_USE_INT_VREF  (0)  // Force DDR controller to use internal VREF
#endif

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ    (666666687u / 2)
#define OS_TICK_FREQ     (1000u)
#define OS_INT_FREQ      (OS_TICK_FREQ)
#define OS_TIMER_RELOAD  ((OS_TIMER_FREQ / OS_INT_FREQ) - 1)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Device specific SFRs
*/
//
// L2 Cache controller
//
#define L2CACHE_BASE_ADDR            (0xF8F02000u)

//
// Private timer
//
#define SCU_BASE_ADDR                (0xF8F00000u)
#define CORE_PRIV_TIMER_BASE_ADDR    (SCU_BASE_ADDR + 0x0600u)
#define CORE_PRIV_TIMER_LOAD         (*(volatile unsigned int*)(CORE_PRIV_TIMER_BASE_ADDR + 0x00u))
#define CORE_PRIV_TIMER_COUNT        (*(volatile unsigned int*)(CORE_PRIV_TIMER_BASE_ADDR + 0x04u))
#define CORE_PRIV_TIMER_CONTROL      (*(volatile unsigned int*)(CORE_PRIV_TIMER_BASE_ADDR + 0x08u))
#define CORE_PRIV_TIMER_STATUS       (*(volatile unsigned int*)(CORE_PRIV_TIMER_BASE_ADDR + 0x0Cu))
#define CORE_PRIV_TIMER_INT_ID       (29u)

//
// Generic interrupt controller
//
#define GIC_BASE_ADDR                (0xF8F00000u)
#define GICC_BASE_ADDR               (GIC_BASE_ADDR + 0x0100u)                             // CPU interface
#define GICD_BASE_ADDR               (GIC_BASE_ADDR + 0x1000u)                             // Distributor interface
#define GICC_HPPIR                   (*(volatile unsigned int*)(GICC_BASE_ADDR + 0x018u))  // Highest Priority Pending Interrupt Register
#define GICD_IPRIORITY               (*(volatile unsigned int*)(GICD_BASE_ADDR + 0x400u))  // Interrupt Priority Registers
#define NUM_INT_SOURCES              (96u)

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/
extern unsigned int __vector;

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

/*********************************************************************
*
*       MMU and cache configuration
*
*  Notes
*    The translation table has to be aligned on 16384 bytes boundary
*    The translation table has to be located in un-initialized memory,
*    because it may be filled from startup, before segments and variables
*    are initialized.
*/
#if   (defined(__ICCARM__))  // IAR
  #pragma data_alignment=16384
  static __no_init OS_U32 _TranslationTable[0x1000];
#elif (defined(__CC_ARM))    // Keil
  #pragma arm section zidata = "NO_INIT"  // Has to be declared as separate section UNINIT in the scatter file
  __align(16384) static OS_U32 _TranslationTable[0x1000];
  #pragma arm section zidata              // back to default (.bss section)
#elif (defined(__GNUC__))    // GCC
  static OS_U32 _TranslationTable[0x1000] __attribute__ ((aligned (16384), section (".non_init._TranslationTable")));
#endif

/*********************************************************************
*
*       Interrupt vector table
*
*  Notes
*    The translation table has to be located in un-initialized memory,
*    because it may be filled from startup, before segments and variables
*    are initialized.
*/
#if   (defined(__ICCARM__))  // IAR
  #pragma data_alignment=4
  static __no_init OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES];
#elif (defined(__CC_ARM))    // Keil
  #pragma arm section zidata = "NO_INIT"  // Has to be declared as separate section UNINIT in the scatter file
  __align(4) static OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES];
  #pragma arm section zidata              // back to default (.bss section)
#elif (defined(__GNUC__))    // GCC
  static OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES] __attribute__ ((aligned (4), section (".non_init._apIsrHandler")));
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USE_PARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USE_PARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USE_PARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USE_PARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _OS_ISR_Undefined()
*
*  Function description
*    Is called when a non-installed interrupt was detected.
*    As interrupt pending condition of peripherals has to be reset,
*    program will not continue when interrupt is ignored.
*/
static void _OS_ISR_Undefined(void) {
  volatile int Dummy;

  Dummy = 1;
  while (Dummy > 0) {
    ;  // You may set a breakpoint here to detect non-installed ISR-handler
  }
}

/*********************************************************************
*
*       _SpuriousIrqHandler()
*
*  Function description
*    default spurious IRQ handler
*
*  Notes
*    A spurious interrupt is called under one of the following
*    circumstances:
*    1. An external interrupt source is programmed in level-sensitive mode
*       and an active level occurs for only a short time.
*    2. An internal interrupt source is programmed in level sensitive
*       and the output signal of the corresponding embedded peripheral
*       is activated for a short time. (As in the case for the Watchdog.)
*    3. An interrupt occurs just a few cycles before the software begins
*       to mask it, thus resulting in a pulse on the interrupt source.
*
*    The spurious handler is called to detect these errors.
*/
static void _SpuriousIrqHandler(void) {
#if (OS_DEBUG != 0)
  static volatile OS_U32 _SpuriousIrqCnt;

  _SpuriousIrqCnt++;
  (void)_SpuriousIrqCnt;
#endif
}

/*********************************************************************
*
*       _InitGIC()
*
*  Function description
*    Initialize interrupt controller by setting default vectors
*    and clearing all interrupts
*/
static void _InitGIC(void) {
  OS_U32 i;

  //
  // Set GIC base address
  //
  OS_GIC_SetCPUInterfaceBaseAddr(GICC_BASE_ADDR);
  OS_GIC_SetDistributionBaseAddr(GICD_BASE_ADDR);
  //
  // Configure all interrupts
  //
  for (i = 0u; i < NUM_INT_SOURCES; i++) {
    _apIsrHandler[i] = _OS_ISR_Undefined;  // Set default interrupt handler
    OS_GIC_SetIntSecure(i);                // Set to secure
    OS_GIC_DisableInt(i);                  // Disable interrupts
    OS_GIC_ClrPending(i);                  // Clear pending flags
    //
    // Initialize all Interrupt pointer target registers to address
    // the first core. The first 8 registers are fixed.
    //
    if (i > 8u) {
      OS_GIC_SetTargetProc(i, 0x01u);
    }
  }
  //
  // Configure priority mask to lowest value (highest number) - enables all priorities
  //
  OS_GIC_SetPrioThreshold(0xFFu);
  //
  // This register defines the point at which the priority value fields split
  // into two parts, the group priority field and the sub-priority field.
  //
  OS_GIC_SetPrioGroup(0x00u);
  //
  // Enable the GIC
  //
  OS_GIC_Enable();
}

/*********************************************************************
*
*       _InitPLL()
*
*  Function description
*    Initialize main clock and PLLs
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called!
*/
static void _InitPLL(void) {
  OS_U32 v;

  //
  // The Zynq 7000 provides 3 PLLs:
  // ARM PLL => Can be used to feed the CPU and some peripherals
  // DDR PLL => Exclusively for DDR RAM
  // I/O PLL => Mainly used for I/O peripherals (can be used but does not have to...)
  // Clock tree:
  //   PS_CLK => ARM PLL (ARM_PLL_CTRL) => 6-bit div (ARM_CLK_CTRL[DIVISOR]) => CPU_1x, CPU_2x, CPU_3x2x, CPU_6x4x
  //   PS_CLK => DDR PLL (DDR_PLL_CTRL) => 6-bit div (DDR_CLK_CTRL[DDR_3XCLK_DIVISOR]) => DDR_3x
  //   PS_CLK => DDR PLL (DDR_PLL_CTRL) => 6-bit div (DDR_CLK_CTRL[DDR_2XCLK_DIVISOR]) => DDR_2x
  //   PS_CLK => IO PLL  (IO_PLL_CTRL)                                                 => I/Os
  //
  // Explanation:
  // CPU_1x   == PLLOut / Div * 1/6
  // CPU_2x   == (PLLOut / Div) * 2/6
  // CPU_3x2x == (PLLOut / Div) * 2/6   or   (PLLOut / Div) * 3/6 in case CLK_621_TRUE [0] == 1
  // CPU_6x4x == (PLLOut / Div) * 4/6   or   (PLLOut / Div) * 6/6 in case CLK_621_TRUE [0] == 1
  //
  // CLK_621_TRUE is set on J-Trace
  // PS_CLK on the J-Trace is 33.33 MHz
  // ARM PLLOut is 1333 MHz on J-Trace
  // Div is 2 on J-Trace
  //
  REG_SLCR_UNLOCK = SCLR_UNLOCK_KEY;
  //
  // Init ARM PLL
  // We can change the register values with no risk as they only become active on a PLL reset
  //
  //
  // Small code exception here:
  // To make porting to macro file easier, we always write the value into temp variable v first
  //
  v = 0
    | (   2u <<  4)                  // PLL_RES
    | (   2u <<  8)                  // PLL_CP
    | (0xFAu << 12)                  // LOCK_CNT
    ;
  REG_SLCR_ARM_PLL_CFG = v;
  v = 0
    | (   0u <<  1)                  // PLL_PWRDWN: 0 == PLL powered up
    | (   0u <<  3)                  // PLL_BYPASS_QUAL: 0 == Bypass is controlled by PLL_BYPASS_FORCE
    | (   1u <<  4)                  // PLL_BYPASS_FORCE: 1 == Bypass PLL
    | (0x28u << 12)                  // PLL_FDIV: Multiplier for PLL
    ;
  REG_SLCR_ARM_PLL_CTRL = v;
  v = 0                              // Set dividers before PLL gets active so we never get out of spec. with the clocks
    | (0u <<  4)                     // SRCSEL: 0 ==
    | (2u <<  8)                     // DIVISOR
    | (1u << 24)                     // CPU_6OR4XCLKACT: CPU_6x4x domain active
    | (1u << 25)                     // CPU_3OR2XCLKACT: CPU_3x2x domain active
    | (1u << 26)                     // CPU_2XCLKACT:    CPU_2x domain active
    | (1u << 27)                     // CPU_1XCLKACT:    CPU_1x domain active
    | (1u << 28)                     // CPU_PERI_CLKACT:
    ;
  REG_SLCR_ARM_CLK_CTRL = v;
  v  = REG_SLCR_ARM_PLL_CTRL;
  v |=  (1u << 0);                   // PLL_RESET => Assert reset, new PLL values become active
  REG_SLCR_ARM_PLL_CTRL = v;
  v  = REG_SLCR_ARM_PLL_CTRL;
  v &= ~(1u << 0);                   // PLL_RESET: De-assert
  REG_SLCR_ARM_PLL_CTRL = v;
  do {
    v = REG_SLCR_PLL_STATUS;
  } while ((v & (1u << 0)) == 0);    // Wait until locked
  v = REG_SLCR_ARM_PLL_CTRL;
  v &= ~(1u << 4);                   // PLL_BYPASS_FORCE: 0 == Remove bypass and switch to PLL
  REG_SLCR_ARM_PLL_CTRL = v;
  //
  // Make sure that we enable CPU_6x and CPU_3x (See clock tree explanation above)
  //
  v  = REG_SLCR_CLK_621_TRUE;
  v |= (1u << 0);                    // CLK_621_TRE CLK_621_TRUE (CPU Clock Ratio Mode Select)
  REG_SLCR_CLK_621_TRUE = v;
  //
  // Init DDR PLL
  //
  v = 0
    | (   4u <<  4)                  // PLL_RES
    | (   2u <<  8)                  // PLL_CP
    | (0xFAu << 12)                  // LOCK_CNT
    ;
  REG_SLCR_DDR_PLL_CFG = v;
  v = 0
    | (   0u <<  1)                  // PLL_PWRDWN: 0 == PLL powered up
    | (   0u <<  3)                  // PLL_BYPASS_QUAL: 0 == Bypass is controlled by PLL_BYPASS_FORCE
    | (   1u <<  4)                  // PLL_BYPASS_FORCE: 1 == Bypass PLL
    | (0x30u << 12)                  // PLL_FDIV: Multiplier for PLL
    ;
  REG_SLCR_DDR_PLL_CTRL = v;
  v = 0                              // Set dividers before PLL gets active so we never get out of spec. with the clocks
    | (1u <<  0)                     // DDR_3XCLKACT
    | (1u <<  1)                     // DDR_2XCLKACT
    | (4u << 20)                     // DDR_3XCLK_DIVISOR
    | (5u << 26)                     // DDR_2XCLK_DIVISOR
    ;
  REG_SLCR_DDR_CLK_CTRL = v;
  v  = REG_SLCR_DDR_PLL_CTRL;
  v |=  (1u << 0);                   // PLL_RESET => Assert reset, new PLL values become active
  REG_SLCR_DDR_PLL_CTRL = v;
  v  = REG_SLCR_DDR_PLL_CTRL;
  v &= ~(1u << 0);                   // PLL_RESET: De-assert
  REG_SLCR_DDR_PLL_CTRL = v;
  do {
    v = REG_SLCR_PLL_STATUS;
  } while ((v & (1u << 1)) == 0);    // Wait until locked
  v  = REG_SLCR_DDR_PLL_CTRL;
  v &= ~(1u << 4);                   // PLL_BYPASS_FORCE: 0 == Remove bypass and switch to PLL
  REG_SLCR_DDR_PLL_CTRL = v;
  //
  // Init IO_PLL
  //
  v = 0
    | (0x00Cu <<  4)                 // PLL_RES
    | (    2u <<  8)                 // PLL_CP
    | (0x145u << 12)                 // LOCK_CNT
    ;
  REG_SLCR_IO_PLL_CFG = v;
  v = 0
    | (   0u <<  1)                  // PLL_PWRDWN: 0 == PLL powered up
    | (   0u <<  3)                  // PLL_BYPASS_QUAL: 0 == Bypass is controlled by PLL_BYPASS_FORCE
    | (   1u <<  4)                  // PLL_BYPASS_FORCE: 1 == Bypass PLL
    | (0x1Eu << 12)                  // PLL_FDIV: Multiplier for PLL
    ;
  REG_SLCR_IO_PLL_CTRL = v;
  v  = REG_SLCR_IO_PLL_CTRL;
  v |=  (1u << 0);                   // PLL_RESET => Assert reset, new PLL values become active
  REG_SLCR_IO_PLL_CTRL = v;
  v  = REG_SLCR_IO_PLL_CTRL;
  v &= ~(1u << 0);                   // PLL_RESET: De-assert
  REG_SLCR_IO_PLL_CTRL = v;
  do {
    v = REG_SLCR_PLL_STATUS;
  } while ((v & (1u << 2)) == 0);    // Wait until locked
  v  = REG_SLCR_IO_PLL_CTRL;
  v &= ~(1u << 4);                   // PLL_BYPASS_FORCE: 0 == Remove bypass and switch to PLL
  REG_SLCR_IO_PLL_CTRL = v;
  //
  // Init clock sources for FPGA / programmable logic
  //
  v = 0
    | (1u << 0)                      // CLKACT
    | (0u << 4)                      // SRCSEL
    | (5u << 8)                      // DIVISOR
    ;
  REG_SLCR_PCAP_CLK_CTRL = v;
  v = 0                              // FPGA0 => 50 MHz
    | (   0u <<  4)                  // SRCSEL
    | (0x14u <<  8)                  // DIVISOR0
    | (   1u << 20)                  // DIVISOR1
    ;
  REG_SLCR_FPGA0_CLK_CTRL = v;
  v = 0                              // FPGA1=> 100 MHz
    | (  0u <<  4)                   // SRCSEL
    | (0xAu <<  8)                   // DIVISOR0
    | (  1u << 20)                   // DIVISOR1
    ;
  REG_SLCR_FPGA1_CLK_CTRL = v;
  v = 0                              // FPGA2 => 200 MHz
    | (0u <<  4)                     // SRCSEL
    | (5u <<  8)                     // DIVISOR0
    | (1u << 20)                     // DIVISOR1
    ;
  REG_SLCR_FPGA2_CLK_CTRL = v;
  v = 0                              // FPGA3 => 50 MHz
    | (   0u <<  4)                  // SRCSEL
    | (0x14u <<  8)                  // DIVISOR0
    | (   1u << 20)                  // DIVISOR1
    ;
  REG_SLCR_FPGA3_CLK_CTRL = v;
  REG_SLCR_LOCK = SCLR_LOCK_KEY;
}

/*********************************************************************
*
*       _InitQSPI()
*
*  Function description
*    Initialize QSPI controller
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*/
static void _InitQSPI(void) {
  REG_SLCR_UNLOCK         = SCLR_UNLOCK_KEY;
  REG_SLCR_APER_CLK_CTRL |= (1u << 23);  // LQSPI_CPU_1XCLKACT: 1 == Enable clock for QSPI controller
  REG_SLCR_LQSPI_CLK_CTRL = 0
                          | (1u <<  0)   // CLKACT: 1 == Clock enabled
                          | (0u <<  4)   // SRCSEL: 0 == IO_PLL, 2 == ARM_PLL, 3 == DDR_PLL
                          | (5u <<  8)   // DIVISOR: Controller clock == SRCSEL / DIVISOR
                          ;
  //
  // Init QSPI for memory mapped mode
  // We do not directly execute from flash but copy all stuff at boot time, so flash speed is only partially important
  // We only use single pin SPI mode, so we set SPI speed to 50 MHz so we can use the basic read commands that does not need dummy cycles (which may differ from flash to flash / vendor)
  //
  REG_QSPIC_CONFIG_REG = 0
                       | (1u << 31)  // leg_flsh: 1 == Memory mapped or I/O mode (not legacy SPI)
                       | (0u << 26)  // endian: 0 == little endian access when reading/writing from/to Rx/Tx regs
                       | (1u << 19)  // Holdb_dr: 1 == Holdb and WPn pins are actively driven by the qspi controller in 1-bit and 2-bit modes
                       | (1u << 17)  // Reserved, must be 1
                       | (0u << 15)  // Man_start_en: 0 == Auto mode, data + command is send as soon as Tx regs are accessed
                       | (0u << 14)  // Manual_CS: 0 == Auto mode
                       | (0u << 10)  // PCS: Controls state of nCS in case Manual_CS == 1
                       | (0u <<  8)  // REF_CLK: Must be 0
                       | (3u <<  6)  // FIFO_WIDTH: Must be 3 (32-bit)
                       | (1u <<  3)  // BAUD_RATE_DIV: Baudrate = 2 ^ (BAUD_RATE_DIV + 1) => I/O PLL (200 MHz) is used as source clock for QSPI baudrate generation
                       | (0u <<  2)  // CLK_PH: SCK is active between transfers
                       | (0u <<  1)  // CLK_POL: 1 == SCK is HIGH between transfers
                       | (1u <<  0)  // MODE_SEL: 1 == SPI is in master mode
                       ;
  //
  // Configure and enable linear QSPI mode (memory mapped mode)
  // Need to take care of the following before entering linear QSPI mode:
  // TXFIFO and RXFIFO are empty
  // REG_QSPIC_CR[Manual_CS] and REG_QSPIC_CR[PCS] must be 0
  //
  REG_QSPIC_LQSPI_CFG    = 0
                         | (1u << 31)  // LQ_MODE: 1 == Linear quad SPI mode
                         | (0u << 30)  // TWO_MEM: 1 == Lower and upper memories are active
                         | (0u << 29)  // SEP_BUS: 1 == Separate memory bus for upper and lower memories
                         | (0u << 28)  // U_PAGE: 1 == Select upper memory page as visible. Don't care if LQ_MODE == 0 AND TWO_MEM == 1 AND SEP_BUS == 0
                         | (1u << 26)  // Reserved. Must be 1
                         | (0u << 25)  // MODE_EN: Enable MODE_BITS[23:16] to be sent. Must be set for dual/quad mode. Must be clear for all other modes
                         | (0u << 24)  // MODE_ON: 1 ==  instruction code is only sent for the very first read transfer
                         | (0u << 16)  // MODE_BITS: If bit 25 is set, this value is required for both dual I/O read and quad I/O read. See vendor's datasheet for more information.
                         | (0u <<  8)  // DUMMY_BYTE: Number of dummy bytes between address and return read data
                         | (3u <<  0)  // INST_CODE: Read instruction code (0x03: Read, 0xEB: Fast read quad I/O)
                         ;
  REG_QSPIC_TX_THRES_REG = 1u;         // Report Tx-FIFO full as soon as min. 1 byte is in the FIFO
  REG_QSPIC_RX_THRES_REG = 1u;         // Report Rx-FIFO full as soon as min. 1 byte is in the FIFO
  REG_QSPIC_LPBK_DLY_ADJ = 0
                         | (0 << 5)    // USE_LPBK: 1 == Use internal loopback master clock for read data capturing when baud rate divisor (reg 0x00) is 2
                         | (0 << 3)    // DLY1: Must be set to 0 if Loopback clk used
                         | (0 << 0)    // DLY0: Must be set to 0 if Loopback clk used
                         ;
  REG_QSPIC_EN_REG       = 1u;         // SPI_EN: Enable SPI
  //
  // Init PIOs needed for QSPI operation after controller init is done
  //
  REG_SLCR_APER_CLK_CTRL |= (1u << 22);  // GPIO_CPU_1XCLKACT: 1 == Enable for GPIO controller
  REG_SLCR_MIO_PIN_01     = MIO_PIN_OPT_IO_CMOS_3V3 | MIO_PIN_OPT_ENA_SPEED_FAST | MIO_PIN_OPT_LVL01_MUX;  // MIO_PIN_01: QSPI_#CS
  REG_SLCR_MIO_PIN_02     = MIO_PIN_OPT_IO_CMOS_3V3 | MIO_PIN_OPT_ENA_SPEED_FAST | MIO_PIN_OPT_LVL01_MUX;  // MIO_PIN_02: QSPI_IO0
  REG_SLCR_MIO_PIN_03     = MIO_PIN_OPT_IO_CMOS_3V3 | MIO_PIN_OPT_ENA_SPEED_FAST | MIO_PIN_OPT_LVL01_MUX;  // MIO_PIN_03: QSPI_IO1
  REG_SLCR_MIO_PIN_04     = MIO_PIN_OPT_IO_CMOS_3V3 | MIO_PIN_OPT_ENA_SPEED_FAST | MIO_PIN_OPT_LVL01_MUX;  // MIO_PIN_04: QSPI_IO2
  REG_SLCR_MIO_PIN_05     = MIO_PIN_OPT_IO_CMOS_3V3 | MIO_PIN_OPT_ENA_SPEED_FAST | MIO_PIN_OPT_LVL01_MUX;  // MIO_PIN_05: QSPI_IO3
  REG_SLCR_MIO_PIN_06     = MIO_PIN_OPT_IO_CMOS_3V3 | MIO_PIN_OPT_ENA_SPEED_FAST | MIO_PIN_OPT_LVL01_MUX;  // MIO_PIN_06: QSPI_CLK
  REG_SLCR_MIO_PIN_08     = MIO_PIN_OPT_IO_CMOS_3V3 | MIO_PIN_OPT_ENA_SPEED_FAST | MIO_PIN_OPT_LVL01_MUX;  // MIO_PIN_08: QSPI_FBCLK
  REG_SLCR_LOCK           = SCLR_LOCK_KEY;
}

/*************************************************************************
*
*       _InitDDR()
*
*  Function description
*    Initialize DDR controller
*
*  Notes
*    (1) This function may be called from __low_level_init() and therefore
*        must not use or call any function which relies on any variable,
*        because variables are not initialized before __low_level_init()
*        is called !
*/
static void _InitDDR(void) {
  OS_U32 v;

  //
  // Small code exception here:
  // To make porting to macro file easier, we always write the value into temp variable v first
  //
  REG_SLCR_UNLOCK = SCLR_UNLOCK_KEY;
  //
  // Enable pins for DDR
  //
  v = 0
    | (0u <<  1)                             // INP_TYPE
    | (0u <<  3)                             // DCI_UPDATE_B
    | (0u <<  4)                             // TERM_EN
    | (1u <<  5)                             // DCI_TYPE
    | (0u <<  7)                             // IBUF_DISABLE_MODE
    | (0u <<  8)                             // TERM_DISABLE_MODE
    | (3u <<  9)                             // OUTPUT_EN
    | (0u << 11)                             // PULLUP_EN
    ;
  REG_SLCR_DDRIOB_ADDR0 = v;
  v = 0
    | (0u <<  0)                             // reserved_INP_POWER
    | (0u <<  1)                             // INP_TYPE
    | (0u <<  3)                             // DCI_UPDATE_B
    | (0u <<  4)                             // TERM_EN
    | (1u <<  5)                             // DCI_TYPE
    | (0u <<  7)                             // IBUF_DISABLE_MODE
    | (0u <<  8)                             // TERM_DISABLE_MODE
    | (3u <<  9)                             // OUTPUT_EN
    | (0u << 11)                             // PULLUP_EN
    ;
  REG_SLCR_DDRIOB_ADDR1 = v;
  v = 0
    | (0u <<  0)                             // reserved_INP_POWER
    | (1u <<  1)                             // INP_TYPE
    | (0u <<  3)                             // DCI_UPDATE_B
    | (0u <<  4)                             // TERM_EN
    | (1u <<  5)                             // DCI_TYPE
    | (0u <<  7)                             // IBUF_DISABLE_MODE
    | (0u <<  8)                             // TERM_DISABLE_MODE
    | (3u <<  9)                             // OUTPUT_EN
    | (0u << 11)                             // PULLUP_EN
    ;
  REG_SLCR_DDRIOB_DATA0 = v;
  v = 0
    | (0u << 11)                             // PULLUP_EN
    | (3u <<  9)                             // OUTPUT_EN
    | (0u <<  8)                             // TERM_DISABLE_MODE
    | (0u <<  7)                             // IBUF_DISABLE_MODE
    | (1u <<  5)                             // DCI_TYPE
    | (0u <<  4)                             // TERM_EN
    | (0u <<  3)                             // DCI_UPDATE_B
    | (1u <<  1)                             // INP_TYPE
    ;
  REG_SLCR_DDRIOB_DATA1 = v;
  v = 0
    | (0u << 11)                             // PULLUP_EN
    | (3u <<  9)                             // OUTPUT_EN
    | (0u <<  8)                             // TERM_DISABLE_MODE
    | (0u <<  7)                             // IBUF_DISABLE_MODE
    | (1u <<  5)                             // DCI_TYPE
    | (0u <<  4)                             // TERM_EN
    | (0u <<  3)                             // DCI_UPDATE_B
    | (2u <<  1)                             // INP_TYPE
    ;
  REG_SLCR_DDRIOB_DIFF0 = v;
  v = 0
    | (0u <<  0)                             // reserved_INP_POWER
    | (2u <<  1)                             // INP_TYPE
    | (0u <<  3)                             // DCI_UPDATE_B
    | (0u <<  4)                             // TERM_EN
    | (1u <<  5)                             // DCI_TYPE
    | (0u <<  7)                             // IBUF_DISABLE_MODE
    | (0u <<  8)                             // TERM_DISABLE_MODE
    | (3u <<  9)                             // OUTPUT_EN
    | (0u << 11)                             // PULLUP_EN
    ;
  REG_SLCR_DDRIOB_DIFF1 = v;
  v = 0
    | (0u <<  0)                             // reserved_INP_POWER
    | (0u <<  1)                             // INP_TYPE
    | (0u <<  3)                             // DCI_UPDATE_B
    | (0u <<  4)                             // TERM_EN
    | (1u <<  5)                             // DCI_TYPE
    | (0u <<  7)                             // IBUF_DISABLE_MODE
    | (0u <<  8)                             // TERM_DISABLE_MODE
    | (3u <<  9)                             // OUTPUT_EN
    | (0u << 11)                             // PULLUP_EN
    ;
  REG_SLCR_DDRIOB_CLOCK = v;
  v = 0
    | (0x2cu <<  0)                          // reserved_DRIVE_P
    | (0x28u <<  7)                          // reserved_DRIVE_N
    | (0x03u << 14)                          // reserved_SLEW_P
    | (0x03u << 19)                          // reserved_SLEW_N
    | (0x00u << 24)                          // reserved_GTL
    | (0x00u << 27)                          // reserved_RTERM
    ;
  REG_SLCR_DDRIOB_DRIVE_SLEW_ADDR = v;
  v = 0
    | (0x2cu <<  0)                          // reserved_DRIVE_P
    | (0x28u <<  7)                          // reserved_DRIVE_N
    | (0x07u << 14)                          // reserved_SLEW_P
    | (0x1fu << 19)                          // reserved_SLEW_N
    | (0x00u << 24)                          // reserved_GTL
    | (0x00u << 27)                          // reserved_RTERM
    ;
  REG_SLCR_DDRIOB_DRIVE_SLEW_DATA = v;
  v = 0
    | (0x2cu <<  0)                          // reserved_DRIVE_P
    | (0x28u <<  7)                          // reserved_DRIVE_N
    | (0x07u << 14)                          // reserved_SLEW_P
    | (0x1fu << 19)                          // reserved_SLEW_N
    | (0x00u << 24)                          // reserved_GTL
    | (0x00u << 27)                          // reserved_RTERM
    ;
  REG_SLCR_DDRIOB_DRIVE_SLEW_DIFF = v;
  v = 0
    | (0x2cu <<  0)                          // reserved_DRIVE_P
    | (0x28u <<  7)                          // reserved_DRIVE_N
    | (0x07u << 14)                          // reserved_SLEW_P
    | (0x1fu << 19)                          // reserved_SLEW_N
    | (0x00u << 24)                          // reserved_GTL
    | (0x00u << 27)                          // reserved_RTERM
    ;
  REG_SLCR_DDRIOB_DRIVE_SLEW_CLOCK = v;
#if (DDR_USE_INT_VREF != 0)
  v = 0
    | (0x1u <<  0)                           // VREF_INT_EN
    | (0x1u <<  1)                           // VREF_SEL
    | (0x0u <<  5)                           // VREF_EXT_EN
    | (0x0u <<  7)                           // reserved_VREF_PULLUP_EN
    | (0x1u <<  9)                           // REFIO_EN
    | (0x3u << 10)                           // reserved_REFIO_TEST
    | (0x0u << 12)                           // reserved_REFIO_PULLUP_EN
    | (0x0u << 13)                           // reserved_DRST_B_PULLUP_EN
    | (0x0u << 14)                           // reserved_CKE_PULLUP_EN
    ;
#else  // DDR_USE_INT_VREF
  v = 0
    | (0x0u <<  0)                           // VREF_INT_EN
    | (0x0u <<  1)                           // VREF_SEL
    | (0x3u <<  5)                           // VREF_EXT_EN
    | (0x0u <<  7)                           // reserved_VREF_PULLUP_EN
    | (0x1u <<  9)                           // REFIO_EN
    | (0x3u << 10)                           // reserved_REFIO_TEST
    | (0x0u << 12)                           // reserved_REFIO_PULLUP_EN
    | (0x0u << 13)                           // reserved_DRST_B_PULLUP_EN
    | (0x0u << 14)                           // reserved_CKE_PULLUP_EN
    ;
#endif  // DDR_USE_INT_VREF
  REG_SLCR_DDRIOB_DDR_CTRL = v;
  v  = REG_SLCR_DDRIOB_DCI_CTRL;
  v |=  (1u << 0);                           // Assert RESET
  REG_SLCR_DDRIOB_DCI_CTRL = v;
  v  = REG_SLCR_DDRIOB_DCI_CTRL;
  v &= ~(1u << 0);                           // Deassert RESET
  REG_SLCR_DDRIOB_DCI_CTRL = v;
  v = REG_SLCR_DDRIOB_DCI_CTRL;
  v |=  (1u << 5);                           // reserved_VRN_OUT
  REG_SLCR_DDRIOB_DCI_CTRL = v;
  v = 0
    | (1u <<  0)                             // RESET
    | (1u <<  1)                             // ENABLE
    | (0u <<  2)                             // reserved_VRP_TRI
    | (0u <<  3)                             // reserved_VRN_TRI
    | (0u <<  4)                             // reserved_VRP_OUT
    | (1u <<  5)                             // reserved_VRN_OUT
    | (2u <<  6)                             // NREF_OPT1
    | (0u <<  8)                             // NREF_OPT2
    | (1u << 11)                             // NREF_OPT4
    | (2u << 14)                             // PREF_OPT1
    | (0u << 17)                             // PREF_OPT2
    | (0u << 20)                             // UPDATE_CONTROL
    | (0u << 21)                             // reserved_INIT_COMPLETE
    | (0u << 22)                             // reserved_TST_CLK
    | (0u << 23)                             // reserved_TST_HLN
    | (0u << 24)                             // reserved_TST_HLP
    | (0u << 25)                             // reserved_TST_RST
    | (0u << 26)                             // reserved_INT_DCI_EN
    ;
  REG_SLCR_DDRIOB_DCI_CTRL = v;
  //
  // Enable clock for DDR module
  //
  v = 0
    | (0x01u <<  0)                          // CLKACT
    | (0x35u <<  8)                          // DIVISOR0
    | (0x03u << 20)                          // DIVISOR1
    ;
  REG_SLCR_DCI_CLK_CTRL = v;
  REG_SLCR_LOCK         = SCLR_LOCK_KEY;
  //
  //  LOCK DDR
  //
  v = 0
    | (0u <<  0)                             // reg_ddrc_soft_rstb
    | (0u <<  1)                             // reg_ddrc_powerdown_en
    | (0u <<  2)                             // reg_ddrc_data_bus_width
    | (0u <<  4)                             // reg_ddrc_burst8_refresh
    | (1u <<  7)                             // reg_ddrc_rdwr_idle_gap
    | (0u << 14)                             // reg_ddrc_dis_rd_bypass
    | (0u << 15)                             // reg_ddrc_dis_act_bypass
    | (0u << 16)                             // reg_ddrc_dis_auto_refresh
    ;
  REG_DDRC_CTRL = v;
  v = 0
    | (0x30u <<  0)                          // reg_ddrc_t_rfc_nom_x32
    | (0x01u << 12)                          // reserved_reg_ddrc_active_ranks
    | (0x00u << 14)                          // reg_ddrc_addrmap_cs_bit0
    | (0x01u << 19)                          // Reserved. Set after reset, do not change
    ;
  REG_DDRC_TWO_RANK_CFG = v;
  //
  // HPR_reg
  //
  v = 0
    | (0xFu << 22)                           // hpr_xact_run_length: Number of transactions that will be serviced once the HPR queue goes critical is the smaller of this number and the number of transactions available.
    | (0xFu << 11)                           // hpr_max_starve_x32: Number of clocks that the HPR queue can be starved before it goes critical. Unit: 32 clocks
    | (0xFu <<  0)                           // hpr_min_non_critical_x32: Number of counts that the HPR queue is guaranteed to be non-critical. Unit: 32 clocks
    ;
  REG_DDRC_HPR_REG = v;
  //
  // LPR_reg
  //
  v = 0
    | (8u << 22)                             // lpr_xact_run_length: Number of transactions that will be serviced once the LPR queue goes critical is the smaller of this number and the number of transactions available.
    | (2u << 11)                             // lpr_max_starve_x32: Number of clocks that the LPR queue can be starved before it goes critical. Unit: 32 clocks
    | (1u <<  0)                             // lpr_min_non_critical_x32: Number of counts that the LPR queue is guaranteed to be non-critical. Unit: 32 clocks
    ;
  REG_DDRC_LPR_REG = v;
  //
  // WR_reg
  //
  v = 0
    | (2u << 15)                             // w_max_starve_x32: Number of clocks that the Write queue can be starved before it goes critical. Unit: 32 clocks. FOR PERFORMANCE ONLY.
    | (8u << 11)                             // w_xact_run_length: Number of transactions that will be serviced once the WR queue goes critical is the smaller of this number and the number of transactions available
    | (1u <<  0)                             // w_min_non_critical_x32: Number of clock cycles that the WR queue is guaranteed to be non-critical.
    ;
  REG_DDRC_WR_REG = v;
  v = 0x00040D27u;
  REG_DDRC_DRAM_PARAM_REG0 = v;
  v = 0x745B50CEu;
  REG_DDRC_DRAM_PARAM_REG1 = v;
  v = 0xA305AD63u;
  REG_DDRC_DRAM_PARAM_REG2 = v;
  v = 0x2648B288u;
  REG_DDRC_DRAM_PARAM_REG3 = v;
  v = 0
    | (0x0u <<  0)                           // reg_ddrc_en_2t_timing_mode
    | (0x0u <<  1)                           // reg_ddrc_prefer_write
    | (0xFu <<  2)                           // Reserved. Set after reset, do not change
    | (0x0u <<  6)                           // reg_ddrc_mr_wr
    | (0x0u <<  7)                           // reg_ddrc_mr_addr
    | (0x0u <<  9)                           // reg_ddrc_mr_data
    | (0x0u << 25)                           // ddrc_reg_mr_wr_busy
    | (0x0u << 26)                           // reg_ddrc_mr_type
    | (0x0u << 27)                           // reg_ddrc_mr_rdata_valid
    ;
  REG_DDRC_DRAM_PARAM_REG4 = v;
  v = 0
    | (7u <<  0)                             // reg_ddrc_final_wait_x32
    | (0u <<  7)                             // reg_ddrc_pre_ocd_x32
    | (4u <<  11)                            // reg_ddrc_t_mrd
    ;
  REG_DDRC_DRAM_INIT_PARAM = v;
  v = 0
    | (2u <<  0)                             // reg_ddrc_emr2
    | (0u <<  16)                            // reg_ddrc_emr3
    ;
  REG_DDRC_DRAM_EMR_REG = v;
  v = 0x00040083u;
  REG_DDRC_DRAM_EMR_MR_REG = v;
  v = 0x004F3FF4u;
  REG_DDRC_DRAM_BURST8_RDWR = v;
  v = 0
    | (0 <<  0)                              // reg_ddrc_force_low_pri_n
    | (0 <<  1)                              // reg_ddrc_dis_dq
    ;
  REG_DDRC_DRAM_DISABLE_DQ = v;
  v = 0
    | (6u <<  0)                              // reg_ddrc_addrmap_bank_b0
    | (6u <<  4)                              // reg_ddrc_addrmap_bank_b1
    | (6u <<  8)                              // reg_ddrc_addrmap_bank_b2
    | (0u << 12)                              // reg_ddrc_addrmap_col_b5
    | (0u << 16)                              // reg_ddrc_addrmap_col_b6
    ;
  REG_DDRC_DRAM_ADDR_MAP_BANK = v;
  v = 0
    | (0x0u <<  0)                            // reg_ddrc_addrmap_col_b2
    | (0x0u <<  4)                            // reg_ddrc_addrmap_col_b3
    | (0x0u <<  8)                            // reg_ddrc_addrmap_col_b4
    | (0x0u << 12)                            // reg_ddrc_addrmap_col_b7
    | (0xFu << 16)                            // reg_ddrc_addrmap_col_b8
    | (0xFu << 20)                            // reg_ddrc_addrmap_col_b9
    | (0xFu << 24)                            // reg_ddrc_addrmap_col_b10
    | (0xFu << 28)                            // reg_ddrc_addrmap_col_b11
    ;
  REG_DDRC_DRAM_ADDR_MAP_COL = v;
  v = 0
    | (0x5u <<  0)                            // reg_ddrc_addrmap_row_b0
    | (0x5u <<  4)                            // reg_ddrc_addrmap_row_b1
    | (0x5u <<  8)                            // reg_ddrc_addrmap_row_b2_11
    | (0x5u << 12)                            // reg_ddrc_addrmap_row_b12
    | (0xFu << 16)                            // reg_ddrc_addrmap_row_b13
    | (0xFu << 20)                            // reg_ddrc_addrmap_row_b14
    | (0xFu << 24)                            // reg_ddrc_addrmap_row_b15
    ;
  REG_DDRC_DRAM_ADDR_MAP_ROW = v;
  v = 0
    | (1u <<  0)                              // reg_ddrc_rank0_rd_odt: Unused...
    | (1u <<  3)                              // reg_ddrc_rank0_wr_odt: Indicates which remote ODT's must be turned on during a write to rank 0.
    | (1u <<  6)                              // Reserved. Set after reset, do not change
    | (1u <<  9)                              // Reserved. Set after reset, do not change
    | (0u << 12)                              // reg_phy_rd_local_odt
    | (3u << 14)                              // reg_phy_wr_local_odt
    | (3u << 16)                              // reg_phy_idle_local_odt
    ;
  REG_DDRC_DRAM_ODT_REG = v;
  v = 0
    | (0u <<  0)                              // reg_phy_rd_cmd_to_data
    | (0u <<  4)                              // reg_phy_wr_cmd_to_data
    | (8u <<  8)                              // reg_phy_rdc_we_to_re_delay
    | (0u << 15)                              // reg_phy_rdc_fifo_rst_disable
    | (1u << 16)                              // reg_phy_use_fixed_re
    | (0u << 17)                              // reg_phy_rdc_fifo_rst_err_cnt_clr
    | (0u << 18)                              // reg_phy_dis_phy_ctrl_rstn
    | (0u << 19)                              // reg_phy_clk_stall_level
    | (7u << 24)                              // reg_phy_gatelvl_num_of_dq0
    | (7u << 28)                              // reg_phy_wrlvl_num_of_dq0
    ;
  REG_DDRC_PHY_CMD_TIMEOUT_RDD_ATA_CPT = v;
  v = 0
    | (1u <<  0)                              // Reserved. Set after reset, do not change
    | (1u <<  8)                              // Reserved. Set after reset, do not change
    | (0u << 16)                              // reg_ddrc_dis_dll_calib
    ;
  REG_DDRC_DLL_CALIB = v;
  v = 0
    | (3u <<  0)                              // reg_ddrc_rd_odt_delay
    | (2u <<  4)                              // reg_ddrc_wr_odt_delay
    | (0u <<  8)                              // reg_ddrc_rd_odt_hold
    | (4u << 12)                              // reg_ddrc_wr_odt_hold
    ;
  REG_DDRC_ODT_DELAY_HOLD = v;
  v = 0
    | (0x00u <<  0)                           // reg_ddrc_pageclose
    | (0x1fu <<  1)                           // reg_ddrc_lpr_num_entries
    | (0x00u <<  7)                           // reg_ddrc_auto_pre_en
    | (0x00u <<  8)                           // reg_ddrc_refresh_update_level
    | (DDR_DISABLE_WC << 9)                   // reg_ddrc_dis_wc
    | (0     << 10)                           // reg_ddrc_dis_collision_page_opt
    | (0     << 12)                           // reg_ddrc_selfref_en
    ;
  REG_DDRC_CTRL_REG1 = v;
  v = 0
    | (0u <<  5)                              // reg_ddrc_go2critical_hysteresis
    | (1u << 17)                              // reg_arb_go2critical_en
    ;
  REG_DDRC_CTRL_REG2 = v;
  v = 0
    | (0x40u <<  0)                           // reg_ddrc_wrlvl_ww
    | (0x40u <<  8)                           // reg_ddrc_rdlvl_rr
    | (0x28u << 16)                           // reg_ddrc_dfi_t_wlmrd
    ;
  REG_DDRC_CTRL_REG3 = v;
  v = 0
    | (0x10u << 0)                            // dfi_t_ctrlupd_interval_min_x1024
    | (0x16u << 8)                            // dfi_t_ctrlupd_interval_max_x1024
    ;
  REG_DDRC_CTRL_REG4 = v;
  v = 0
    | (1u <<  0)                              // reg_ddrc_dfi_t_ctrl_delay
    | (1u <<  4)                              // reg_ddrc_dfi_t_dram_clk_disable
    | (1u <<  8)                              // reg_ddrc_dfi_t_dram_clk_enable
    | (5u << 12)                              // reg_ddrc_t_cksre
    | (5u << 16)                              // reg_ddrc_t_cksrx
    | (6u << 20)                              // reg_ddrc_t_ckesr
    ;
  REG_DDRC_CTRL_REG5 = v;
  v = 0
    | (2u <<  0)                              // reg_ddrc_t_ckpde
    | (2u <<  4)                              // reg_ddrc_t_ckpdx
    | (2u <<  8)                              // reg_ddrc_t_ckdpde
    | (2u << 12)                              // reg_ddrc_t_ckdpdx
    | (5u << 16)                              // reg_ddrc_t_ckcsx
    ;
  REG_DDRC_CTRL_REG6 = v;
  v = 0
    | (0      <<  0)                          // reg_ddrc_dis_auto_zq
    | (0      <<  1)                          // reg_ddrc_ddr3
    | (0x200u <<  2)                          // reg_ddrc_t_mod
    | (0x200u << 12)                          // reg_ddrc_t_zq_long_nop
    | (0x040u << 22)                          // reg_ddrc_t_zq_short_nop
    ;
  REG_DDRC_CHE_T_ZQ = v;
  v = (0x4Fu << 20) | (0x9896u << 0);         // dram_rstn_x1024, t_zq_short_interval_x1024
  REG_DDRC_CHE_T_ZQ_SHORT_INTERVAL_REG = v;
  v = (0xC4u << 1) | (0 << 0);                // deeppowerdown_to_x1024, deeppowerdown_en
  REG_DDRC_DEEP_PWRDWN_REG = v;
  v = 0
    | (0xfffu  <<  0)                         // dfi_wrlvl_max_x1024
    | (0xfffu  << 12)                         // dfi_rdlvl_max_x1024
    | (0x0u    << 24)                         // ddrc_reg_twrlvl_max_error
    | (0x0u    << 25)                         // ddrc_reg_trdlvl_max_error
    | (0x0u    << 26)                         // reg_ddrc_dfi_wr_level_en
    | (0x1u    << 27)                         // reg_ddrc_dfi_rd_dqs_gate_level
    | (0x1u    << 28)                         // reg_ddrc_dfi_rd_data_eye_train
    ;
  REG_DDRC_REG_2C = v;
  v = (1u << 9);                              // reg_ddrc_skip_ocd
  REG_DDRC_REG_2D = v;
  v = 0
    | (0x05u <<  0)                           // reg_ddrc_dfi_t_rddata_en
    | (0x03u <<  5)                           // reg_ddrc_dfi_t_ctrlup_min
    | (0x40u << 15)                           // reg_ddrc_dfi_t_ctrlup_max
    ;
  REG_DDRC_DFI_TIMING = v;
  v = REG_DDRC_CHE_ECC_CONTROL_REG_OFFSET;
  v |=  3u;  // Clear_Uncorrectable_DRAM_ECC_error, Clear_Correctable_DRAM_ECC_error
  REG_DDRC_CHE_ECC_CONTROL_REG_OFFSET = v;
  v = REG_DDRC_CHE_ECC_CONTROL_REG_OFFSET;
  v &= ~3u;  // Clear_Uncorrectable_DRAM_ECC_error, Clear_Correctable_DRAM_ECC_error
  REG_DDRC_CHE_ECC_CONTROL_REG_OFFSET = v;
  v = 0
    | (0 << 0)                                // CORR_ECC_LOG_VALID
    | (0 << 1)                                // ECC_CORRECTED_BIT_NUM
    ;
  REG_DDRC_CHE_CORR_ECC_LOG_REG_OFFSET = v;
  v = REG_DDRC_CHE_UNCORR_ECC_LOG_REG_OFFSET;
  v &= ~(1u << 0);                            // UNCORR_ECC_LOG_VALID
  REG_DDRC_CHE_UNCORR_ECC_LOG_REG_OFFSET = v;
  v = 0
    | (0 << 8)                                // STAT_NUM_CORR_ERR
    | (0 << 0)                                // STAT_NUM_UNCORR_ERR
    ;
  REG_DDRC_CHE_ECC_STATS_REG_OFFSET = v;
  v = 0
    | (0u << 0)                               // reg_ddrc_ecc_mode
    | (1u << 3)                               // reg_ddrc_dis_scrub
    ;
  REG_DDRC_ECC_SCRUB = v;
  v = 0
    | (0 << 0)                                // reg_phy_dif_on
    | (0 << 4)                                // reg_phy_dif_off
    ;
  REG_DDRC_PHY_RCVR_ENABLE = v;
  v = 0
    | (0x1u  <<  0)                           // reg_phy_data_slice_in_use
    | (0x0u  <<  1)                           // reg_phy_rdlvl_inc_mode
    | (0x0u  <<  2)                           // reg_phy_gatelvl_inc_mode
    | (0x0u  <<  3)                           // reg_phy_wrlvl_inc_mode
    | (0x0u  <<  6)                           // reg_phy_bist_shift_dq
    | (0x0u  << 15)                           // reg_phy_bist_err_clr
    | (0x40u << 24)                           // reg_phy_dq_offset
    ; // 0x40000001
  REG_DDRC_PHY_CONFIG0 = v;
  REG_DDRC_PHY_CONFIG1 = v;
  REG_DDRC_PHY_CONFIG2 = v;
  REG_DDRC_PHY_CONFIG3 = v;
  v = (0x222u << 10);                         // reg_phy_gatelvl_init_ratio
  REG_DDRC_PHY_INIT_RATIO0 = v;
  REG_DDRC_PHY_INIT_RATIO1 = v;
  REG_DDRC_PHY_INIT_RATIO2 = v;
  REG_DDRC_PHY_INIT_RATIO3 = v;
  v = 0
    | (0x35u <<  0)                           // reg_phy_rd_dqs_slave_ratio
    | (0     << 10)                           // reg_phy_rd_dqs_slave_force
    | (0     << 11)                           // reg_phy_rd_dqs_slave_delay
    ;
  REG_DDRC_PHY_RD_DQS_CFG0 = v;
  REG_DDRC_PHY_RD_DQS_CFG1 = v;
  REG_DDRC_PHY_RD_DQS_CFG2 = v;
  REG_DDRC_PHY_RD_DQS_CFG3 = v;
  v = 0
    | (0 <<  0)                               // reg_phy_wr_dqs_slave_ratio
    | (0 << 10)                               // reg_phy_wr_dqs_slave_force
    | (0 << 11)                               // reg_phy_wr_dqs_slave_delay
    ;
  REG_DDRC_PHY_WR_DQS_CFG0 = v;
  REG_DDRC_PHY_WR_DQS_CFG1 = v;
  REG_DDRC_PHY_WR_DQS_CFG2 = v;
  REG_DDRC_PHY_WR_DQS_CFG3 = v;
  v = 0
    | (0x204u <<  0)                          // reg_phy_fifo_we_slave_ratio
    | (0      << 11)                          // reg_phy_fifo_we_in_force
    | (0      << 12)                          // reg_phy_fifo_we_in_delay
    ;
  REG_DDRC_PHY_WE_CFG0 = v;
  REG_DDRC_PHY_WE_CFG1 = v;
  REG_DDRC_PHY_WE_CFG2 = v;
  REG_DDRC_PHY_WE_CFG3 = v;
  v = 0
    | (0x40u <<  0)                           // reg_phy_wr_data_slave_ratio
    | (0     << 10)                           // reg_phy_wr_data_slave_force
    | (0     << 11)                           // reg_phy_wr_data_slave_delay
    ;
  REG_DDRC_WR_DATA_SLV0 = v;
  REG_DDRC_WR_DATA_SLV1 = v;
  REG_DDRC_WR_DATA_SLV2 = v;
  REG_DDRC_WR_DATA_SLV3 = v;
  v = 0
    | (0     <<  1)                           // reg_phy_bl2
    | (0     <<  2)                           // reg_phy_at_spd_atpg
    | (0     <<  3)                           // reg_phy_bist_enable
    | (0     <<  4)                           // reg_phy_bist_force_err
    | (0     <<  5)                           // reg_phy_bist_mode
    | (0     <<  7)                           // reg_phy_invert_clkout
    | (0     <<  9)                           // reg_phy_sel_logic
    | (0x40u << 10)                           // reg_phy_ctrl_slave_ratio
    | (0     << 20)                           // reg_phy_ctrl_slave_force
    | (0     << 21)                           // reg_phy_ctrl_slave_delay
    | (1u    << 28)                           // Reserved. Set after reset, do not change
    | (0     << 29)                           // reg_phy_lpddr
    | (0     << 30)                           // reg_phy_cmd_latency
    ;
  REG_DDRC_REG_64 = v;
  v = 0x0001BC61u;
  REG_DDRC_REG_65 = v;
  v = 0; // reg_arb_page_addr_mask
  REG_DDRC_PAGE_MASK = v;
  v = 0
    | (0x3ffu <<  0)                          // reg_arb_pri_wr_portn
    | (0      << 16)                          // reg_arb_disable_aging_wr_portn
    | (0      << 17)                          // reg_arb_disable_urgent_wr_portn
    | (0      << 18)                          // reg_arb_dis_page_match_wr_portn
    | (1u     << 19)                          // reg_arb_set_hpr_rd_portn
    ;
  REG_DDRC_AXI_PRIORITY_WR_PORT0 = v;
  REG_DDRC_AXI_PRIORITY_WR_PORT1 = v;
  REG_DDRC_AXI_PRIORITY_WR_PORT2 = v;
  REG_DDRC_AXI_PRIORITY_WR_PORT3 = v;
  v = 0
    | (0x3ffu <<  0)                          // reg_arb_pri_rd_portn
    | (0      << 16)                          // reg_arb_disable_aging_rd_portn
    | (0      << 17)                          // reg_arb_disable_urgent_rd_portn
    | (0      << 18)                          // reg_arb_dis_page_match_rd_portn
    ;
  REG_DDRC_AXI_PRIORITY_RD_PORT0 = v;
  REG_DDRC_AXI_PRIORITY_RD_PORT1 = v;
  REG_DDRC_AXI_PRIORITY_RD_PORT2 = v;
  REG_DDRC_AXI_PRIORITY_RD_PORT3 = v;
  v = 0
    | (1u << 0)                               // reg_ddrc_lpddr2
    | (0u << 2)                               // reg_ddrc_derate_enable
    | (0u << 4)                               // reg_ddrc_mr4_margin
    ;
  REG_DDRC_LPDDR_CTRL0 = v;
  v = 0; // reg_ddrc_mr4_read_interval
  REG_DDRC_LPDDR_CTRL1 = v;
  v = 0
    | (0x5u <<  0)                            // reg_ddrc_min_stable_clock_x1
    | (0xEu <<  4)                            // reg_ddrc_idle_after_reset_x32
    | (0x5u << 12)                            // reg_ddrc_t_mrw
    ;
  REG_DDRC_LPDDR_CTRL2 = v;
  v = (0x7Eu << 0) | (0xEu << 8);             // reg_ddrc_max_auto_init_x1024, reg_ddrc_dev_zqinit_x32
  REG_DDRC_LPDDR_CTRL3 = v;
  do {
    v = REG_SLCR_DDRIOB_DCI_STATUS;
  } while ((v & (1u << 13)) == 0);            // DONE
  v = 0
    | (1u <<  0)                              // reg_ddrc_soft_rstb
    | (0u <<  1)                              // reg_ddrc_powerdown_en
    | (0u <<  2)                              // reg_ddrc_data_bus_width
    | (0u <<  4)                              // reg_ddrc_burst8_refresh
    | (1u <<  7)                              // reg_ddrc_rdwr_idle_gap
    | (0u << 14)                              // reg_ddrc_dis_rd_bypass
    | (0u << 15)                              // reg_ddrc_dis_act_bypass
    | (0u << 16)                              // reg_ddrc_dis_auto_refresh
    ;
  REG_DDRC_CTRL = v;
  do {
    v = REG_DDRC_MODE_STS_REG;
  } while ((v & (7 << 0)) == 0);
  // Init DDR peripherals
  //
  //
  // DDR TERM/IBUF_DISABLE_MODE SETTINGS
  //
  REG_SLCR_UNLOCK = SCLR_UNLOCK_KEY;
  v  = REG_SLCR_DDRIOB_DATA0;
  v |=  (1u << 7);                            // IBUF_DISABLE_MODE: 0 == ibuf is enabled
  v &= ~(1u << 8);                            // TERM_DISABLE_MODE: 0 == termination always enabled
  REG_SLCR_DDRIOB_DATA0 = v;
  v  = REG_SLCR_DDRIOB_DATA1;
  v |=  (1u << 7);                            // IBUF_DISABLE_MODE: 0 == ibuf is enabled
  v &= ~(1u << 8);                            // TERM_DISABLE_MODE: 0 == termination always enabled
  REG_SLCR_DDRIOB_DATA1 = v;
  v  = REG_SLCR_DDRIOB_DIFF0;
  v |=  (1u << 7);                            // IBUF_DISABLE_MODE: 0 == ibuf is enabled
  v &= ~(1u << 8);                            // TERM_DISABLE_MODE: 0 == termination always enabled
  REG_SLCR_DDRIOB_DIFF0 = v;
  v  = REG_SLCR_DDRIOB_DIFF1;
  v |=  (1u << 7);                            // IBUF_DISABLE_MODE: 0 == ibuf is enabled
  v &= ~(1u << 8);                            // TERM_DISABLE_MODE: 0 == termination always enabled
  REG_SLCR_DDRIOB_DIFF1 = v;
  REG_SLCR_LOCK = SCLR_LOCK_KEY;
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return CORE_PRIV_TIMER_COUNT;
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return CORE_PRIV_TIMER_STATUS & (1u << 0);
}

/*********************************************************************
*
*       _OS_ISR_Tick()
*
*  Function description
*    This is the hardware timer exception handler.
*/
static void _OS_ISR_Tick(void) {
  OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_Process();
#endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
void OS_InitHW(void) {
  OS_U32 TickPrio;

  OS_INT_IncDI();
  //
  // Ensure caches are synchronized
  //
  OS_ARM_CACHE_Sync();
  //
  // Set vector table base address
  //
  OS_ARM_SetVBAR((OS_U32)&__vector);
  //
  // Initialize interrupt controller
  //
  _InitGIC();
  //
  // Initialize timer for embOS
  //
  CORE_PRIV_TIMER_CONTROL = 0u;               // Stop Timer, set Prescaler to 1
  CORE_PRIV_TIMER_LOAD    = OS_TIMER_RELOAD;  // Set timer interval
  CORE_PRIV_TIMER_CONTROL = (1u << 0)         // Enable Timer
                          | (1u << 1)         // Set auto reload mode
                          | (1u << 2);        // Enable timer interrupt
  OS_ARM_InstallISRHandler(CORE_PRIV_TIMER_INT_ID, _OS_ISR_Tick);
  OS_ARM_ISRSetPrio(CORE_PRIV_TIMER_INT_ID, 0xFFu);              // Set to lowest level, ALL BITS set
  TickPrio  = OS_ARM_ISRSetPrio(CORE_PRIV_TIMER_INT_ID, 0xFFu);  // Read back priority to examine relevant preemption-level-bits
  TickPrio -= 1;                                                 // Set to low preemption level, 1 level higher than lowest
  OS_ARM_ISRSetPrio(CORE_PRIV_TIMER_INT_ID, TickPrio);
  OS_ARM_EnableISR(CORE_PRIV_TIMER_INT_ID);
  //
  // Inform embOS about the timer settings
  //
  {
    OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
    OS_TIME_ConfigSysTimer(&SysTimerConfig);
  }
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {  // Idle loop: No task is ready to execute
  while (1) {         // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                      // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USE_PARA(c);          // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       embOS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;
  volatile OS_U32 Dummy;
  OS_U32          int_id;
  OS_U32          icdipr0_value;

  Dummy = GICC_HPPIR;                              // Erratum 801120 [1] - ensure subsequent ICCIAR data will be correct
  OS_USE_PARA(Dummy);                              // Avoid compiler warning about set but unused variable
  int_id = OS_GIC_AckInt();                        // Acknowledge the interrupt and read the interrupt ID
  //
  // Read data of 0x0 (SGI ID0 from CPU0) is possibly affected by 733075
  // Interrupt IDs 0x3FE and 0x3FF are also possibly affected
  //
  if ((int_id == 0u) || (int_id  >= 0x3FEu)) {
    //
    // Erratum 733075 - prevent possible CPU interface deadlock (ARM Errata Notice, PR258-GENC-008719 v6.0).
    // Assumes SGI ID0 from CPU0 is not used as suggested by ARM.
    //
    icdipr0_value  = GICD_IPRIORITY;               // One way to get the value for the workaround write is to read it first
    GICD_IPRIORITY = icdipr0_value;
    __asm volatile("dsb");                         // Ensure write completes before any future ICCHPIR or ICCIAR reads
  }
  if ((int_id == 0x3FEu) || (int_id == 0x3FFu)) {  // Check for spurious interrupts
     _SpuriousIrqHandler();
  } else {
    if (int_id < NUM_INT_SOURCES) {
      pISR = _apIsrHandler[int_id];                // Get interrupt vector handler
      OS_INT_Enter();
      pISR();                                      // Call interrupt service routine
      OS_INT_Leave();
    }
    OS_GIC_EOI(int_id);                            // End interrupt
  }
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Installs an interrupt handler
*
*  Parameters
*    ISRIndex:    Index of the interrupt source, usually the interrupt vector number
*    pISRHandler: Address of the interrupt handler function
*
*  Return value
*    The address of the interrupt handler that was previously installed with the addressed interrupt source
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler(int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER* pOldHandler;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  pOldHandler             = _apIsrHandler[ISRIndex];
  _apIsrHandler[ISRIndex] = pISRHandler;
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Enables an interrupt
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be enabled
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_GIC_EnableInt(ISRIndex);
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Disables an interrupt
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be disabled
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_GIC_DisableInt(ISRIndex);
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Function description
*    Sets the priority of an interrupt
*
*  Parameters
*    ISRIndex: Index of the interrupt source for which to configure the priority
*    Prio    : The priority which should be set for the specific interrupt
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  OS_U32 OldPrio;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  OldPrio = OS_GIC_SetPrio(ISRIndex, Prio);
  OS_INT_EnableConditional();
  return (int)OldPrio;
}

/*********************************************************************
*
*       OS_ARM_ClearPendingFlag()
*
*  Function description
*    Clears an interrupt pending flag
*
*  Parameters
*    ISRIndex: Index of the interrupt source for which to clear the flag
*/
void OS_ARM_ClearPendingFlag(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  OS_GIC_ClrPending(ISRIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_IsPending()
*
*  Function description
*    Checks if an interrupt is pending
*
*  Parameters
*    ISRIndex: Index of the interrupt source that shall be checked
*
*  Return value
*    == 0; Interrupt is not pending
*    != 0: Interrupt is pending
*/
unsigned int OS_ARM_IsPending(int ISRIndex) {
  OS_U32 Result;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  Result = OS_GIC_IsPending(ISRIndex);
  OS_INT_EnableConditional();
  return Result;
}

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL.
*
*    Has to be modified if another CPU clock frequency should be
*    used. This function is called during startup and has to
*    return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
OS_INTERWORK int __low_level_init(void);
OS_INTERWORK int __low_level_init(void) {
  REG_SLCR_UNLOCK  = SCLR_UNLOCK_KEY;
  REG_SLCR_OCM_CFG = (1u << 4);  // [2] seems to be set and must kept set it seems. Doc does not say anything about this bit...
  REG_SLCR_LOCK    = SCLR_LOCK_KEY;
  //
  // Initialize PLL, DDR RAM and QSPI flash
  //
  _InitPLL();
  _InitDDR();
  _InitQSPI();
  //
  // Init MMU and caches. This defines the virtual memory map, which is used during execution.
  // Code below fills all 4096 entries in ascending VAddr order.
  //
  OS_ARM_MMU_InitTT(_TranslationTable);
  //                                         Mode                               VAddr  PAddr  Size[MB]
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_WRITE_BACK_ALLOC, 0x000, 0x000, 0x001);  // 256 KB iRAM
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_WRITE_BACK_ALLOC, 0x001, 0x001, 0x004);  // First 4 MB of DDR are reserved for FW execution, therefore cacheable, bufferable. Rest is non cacheable, non bufferable
  OS_ARM_MMU_AddTTEntries(_TranslationTable, OS_ARM_CACHEMODE_STRONGLY_ORDERED, 0x005, 0x005, 0xFFB);  // Everything else is non cacheable, non bufferable
  //
  // Set API functions and base address for L2 Cache
  //
  OS_ARM_AddL2Cache(&OS_L2CACHE_XilinxZynq7000, (void*)L2CACHE_BASE_ADDR);
  //
  // Now enable MMU, Caches and branch prediction unit
  //
  OS_ARM_MMU_Enable(_TranslationTable);
  OS_ARM_ICACHE_Enable();
  OS_ARM_DCACHE_Enable();
  return 1;
}

/*************************** End of file ****************************/
