/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : BSP.c
Purpose : BSP for TI OMAP AM3517 EVM Main board (OMAP3517)
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "BSP.h"
#include "RTOS.h"

/*********************************************************************
*
*       Prototypes
*
**********************************************************************
*/
static OS_U32 I2C_read(OS_U32 i2c, OS_U16 in_addr, OS_U8 *dest_buffer, OS_U16 in_length);
static OS_U32 I2C_write(OS_U32 i2c, OS_U16 in_addr, OS_U8 *src_buffer, OS_U16 in_length);

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/
typedef enum
{
  I2C_CLK_100K = 0,
  I2C_CLK_400K
} i2c_clk_e;

typedef struct PIN_CONFIG
{
   OS_U32  reg_addr;
   OS_U8   mode;
   OS_U8   pull_en;
   OS_U8   pull_type;
} PIN_CONFIG;

#define SYS_CLOCK_SPEED_MHZ                 (26)

#define LED_1                               (0)
#define LED_2                               (1)
#define MAX_LED_NUM                         (2)

#define IN_REGB(offSet)                     (*(volatile OS_U8*) (offSet))
#define IN_REGS(offSet)                     (*(volatile OS_U16*)(offSet))
#define IN_REGL(offSet)                     (*(volatile OS_U32*)(offSet))
#define OUT_REGB(offSet, value)             (*(volatile OS_U8*) (offSet) = (OS_U8) (value))
#define OUT_REGS(offSet, value)             (*(volatile OS_U16*)(offSet) = (OS_U16)(value))
#define OUT_REGL(offSet, value)             (*(volatile OS_U32*)(offSet) = (OS_U32)(value))
#define SETBIT_REGL(dest, mask)             (OUT_REGL((dest), IN_REGL((dest)) | (mask)))
#define CLRBIT_REGL(dest, mask)             (OUT_REGL((dest), IN_REGL((dest)) & (OS_U32)~(mask)))

#define SETBIT(dest,mask)                   (dest |= mask)
#define CLRBIT(dest,mask)                   (dest &= ~mask)

#define ERR_NO_ERROR                        (0)
#define ERR_INIT_FAIL                       (2)
#define ERR_I2C_ERROR                       (14)

#define TIOCP_CFG                           0x0010
#define TISTAT                              0x0014
#define TISR                                0x0018
#define TIER                                0x001C
#define TWER                                0x0020
#define TCLR                                0x0024
#define TCRR                                0x0028
#define TLDR                                0x002C
#define TTGR                                0x0030
#define TWPS                                0x0034
#define TMAR                                0x0038
#define TCAR1                               0x003C
#define TSICR                               0x0040
#define TCAR2                               0x0044
#define TPIR                                0x0048
#define TNIR                                0x004C
#define TCVR                                0x0050
#define TOCR                                0x0054
#define TOWR                                0x0058

// bitmask defines for TCLR
#define GPO_CFG                             (0x00004000)   // bit 14
#define CAPT_MODE                           (0x00002000)   // bit 13
#define PT                                  (0x00001000)   // bit 12
#define TRG                                 (0x00000C00)   // bit 10,11
#define TCM                                 (0x00000300)   // bit 8,9
#define SCPWM                               (0x00000040)   // bit 7
#define CE                                  (0x00000020)   // bit 6
#define PRE                                 (0x00000020)   // bit 5
#define PTV                                 (0x0000001C)   // bit 2,3,4
#define AR                                  (0x00000002)   // bit 1
#define ST                                  (0x00000001)   // bit 0

#define GPTIMER10                           (GPTIMER10_REG_BASE)
#define GPTIMER11                           (GPTIMER11_REG_BASE)

#define GPTIMER10_REG_BASE                  (0x48086000)
#define GPTIMER11_REG_BASE                  (0x48088000)

#define GPTIMER_FREE                        GPTIMER10
#define GPTIMER_DELAY                       GPTIMER11

// bitmask defines for TSICR
#define POSTED                              (0x00000004)   // bit 2
#define SFT                                 (0x00000002)   // bit 1

#define I2C_TIMEOUT                         (2000000) // in us
#define MAX_TIMEOUT_COUNT                   (10000)
#define MAX_NUM_ATTEMPTS                    (1)

// I2C register offsets
#define I2C_IE                              (0x04)
#define I2C_STAT                            (0x08)
#define I2C_WE                              (0x0C)
#define I2C_SYSS                            (0x10)
#define I2C_BUF                             (0x14)
#define I2C_CNT                             (0x18)
#define I2C_DATA                            (0x1C)
#define I2C_SYSC                            (0x20)
#define I2C_CON                             (0x24)
#define I2C_OA0                             (0x28)
#define I2C_SA                              (0x2C)
#define I2C_PSC                             (0x30)
#define I2C_SCLL                            (0x34)
#define I2C_SCLH                            (0x38)
#define I2C_SYSTEST                         (0x3C)
#define I2C_BUFSTAT                         (0x40)
#define I2C_OA1                             (0x44)
#define I2C_OA2                             (0x48)
#define I2C_OA3                             (0x4C)
#define I2C_ACTOA                           (0x50)
#define I2C_SBLOCK                          (0x54)

// defines for I2C_CON
#define I2C_CON_I2C_EN                      (0x00008000)  // bit 15
#define I2C_CON_MST                         (0x00000400)  // bit 10
#define I2C_CON_TRX                         (0x00000200)  // bit 9
#define I2C_CON_STP                         (0x00000002)  // bit 1
#define I2C_CON_STT                         (0x00000001)  // bit 0

// defines for I2C_STAT.
#define I2C_STAT_SBD                        (0x00008000)  // bit 15
#define I2C_STAT_XDR                        (0x00004000)  // bit 14
#define I2C_STAT_RDR                        (0x00002000)  // bit 13
#define I2C_STAT_BB                         (0x00001000)  // bit 12
#define I2C_STAT_ROVR                       (0x00000800)  // bit 11
#define I2C_STAT_XUDF                       (0x00000400)  // bit 10
#define I2C_STAT_AAS                        (0x00000200)  // bit 9
#define I2C_STAT_BF                         (0x00000100)  // bit 8
#define I2C_STAT_GC                         (0x00000020)  // bit 5
#define I2C_STAT_XRDY                       (0x00000010)  // bit 4
#define I2C_STAT_RRDY                       (0x00000008)  // bit 3
#define I2C_STAT_ARDY                       (0x00000004)  // bit 2
#define I2C_STAT_NACK                       (0x00000002)  // bit 1
#define I2C_STAT_AL                         (0x00000001)  // bit 0

// defines for I2C_SYSS
#define I2C_SYSS_RDONE                      (0x00000001)  // bit 0

// defines for I2C_SYSC
#define I2C_SYSC_SRST                       (0x00000002)  // bit 1

// bitmask defines for SYSC.
#define IDLEMODE                            (0x00000018)   // bit 3,4
#define ENAWAKEUP                           (0x00000004)   // bit 2
#define SOFTRESET                           (0x00000002)   // bit 1
#define AUTOIDLE                            (0x00000001)   // bit 0

// bitmask defines for TIOCP_CFG
#define CLOCKACTIVITY                       (0x00000300)   // bit 8,9
#define RESERVED                            (0x000000C0)   // bit 6,7
#define EMUFREE                             (0x00000020)   // bit 5
#define IDLEMODE                            (0x00000018)   // bit 3,4
#define ENAWAKEUP                           (0x00000004)   // bit 2
#define SOFTRESET                           (0x00000002)   // bit 1
#define AUTOIDLE                            (0x00000001)   // bit 0

#define I2C_GPIO_EX_LED2                    (6)
#define I2C_GPIO_EX_LED1                    (7)

#define CORE_CM                             (0x48004A00)
#define CM_CLKSEL_CORE                      (CORE_CM + 0x40)

// PRM_CLCKSRC_CTRL SYSCLKDIV Bitmask
#define PRM_CLCKSRC_CTRL_SYSCLKDIV          (0x3uL << 6)

// PRM_CLCKSRC_CTRL SYSCLKDIV: Syst_clk is external clock / 2
#define PRM_CLCKSRC_CTRL_SYSCLKDIV_2        (0x2uL << 6)

#define GLOBAL_REGISTERS_PRM                (0x48307200)
#define PRM_CLCKSRC_CTRL                    (GLOBAL_REGISTERS_PRM + 0x70)

// defines for I2C_OA0
#define I2C_OA0_ADDRESS                     (0x0000000E)
#define I2C_OA0_MCODE                       (0x00004000)  // bit 14

// i2c addresses for AM-35xx EVM.
#define I2C_ADDR_AIC3106                    (0x0018)
#define I2C_ADDR_GPIO_UI1                   (0x0020)
#define I2C_ADDR_GPIO_UI2                   (0x0021)
#define I2C_ADDR_GPIO_EX                    (0x0021)
#define I2C_ADDR_PMIC                       (0x0048)
#define I2C_ADDR_SVIDEODEC                  (0x005C)
#define I2C_ADDR_CVIDEODEC                  (0x005D)
#define I2C_ADDR_CDCE913                    (0x0065)
#define I2C_ADDR_CVIDEOENC                  (0x002A)
#define I2C_ADDR_RTC                        (0x0030)

#define I2C1_REG_BASE                       (0x48070000)
#define I2C2_REG_BASE                       (0x48072000)
#define I2C3_REG_BASE                       (0x48060000)

#define I2C1                                (OS_U32)(I2C1_REG_BASE)
#define I2C2                                (OS_U32)(I2C2_REG_BASE)
#define I2C3                                (OS_U32)(I2C3_REG_BASE)

#define I2C_PORT_GPIO_EX                    (I2C2)
#define I2C_PORT_GPIO_UI                    (I2C3)

#define I2C_GPIO_PIN_MAX                    (16)
// TCA6416 command byte defines.
#define CMD_BYTE_INPUT0                     (0x00)
#define CMD_BYTE_INPUT1                     (0x01)
#define CMD_BYTE_OUTPUT0                    (0x02)
#define CMD_BYTE_OUTPUT1                    (0x03)
#define CMD_BYTE_POLARITY0                  (0x04)
#define CMD_BYTE_POLARITY1                  (0x05)
#define CMD_BYTE_CONFIG0                    (0x06)
#define CMD_BYTE_CONFIG1                    (0x07)

/*pinmuxing base address sHs*/
#define PAD_CONFIG_BASE                     (0x48002000)

#define CONTROL_PADCONF_hsusb0_data7        (PAD_CONFIG_BASE + 0x01B8)
#define CONTROL_PADCONF_i2c1_sda            (PAD_CONFIG_BASE + 0x01BC)

#define CONTROL_PADCONF_i2c1_sda_HI         (PAD_CONFIG_BASE + 2 + 0x01BC)
#define CONTROL_PADCONF_i2c2_sda            (PAD_CONFIG_BASE + 0x01C0)

#define CONTROL_PADCONF_i2c2_sda_HI         (PAD_CONFIG_BASE + 2 + 0x01C0)
#define CONTROL_PADCONF_i2c3_sda            (PAD_CONFIG_BASE + 0x01C4)

#define MODE0                               (00)

#define PUPD_EN                             (01)
#define PUPD_DIS                            (00)

#define PULL_UP                             (01)
#define PULL_DOWN                           (00)
#define INPUT_EN                            (02)
#define INPUT_DIS                           (00)

#define LOAD_PINMUX(x)                      do { _pin_config_mux(x, ARRAY_ENTRIES(x)); } while(0)
#define ARRAY_ENTRIES(x)                    (sizeof(x) / sizeof(x[0]))

#define I2C_GPIO_CONFIG0_EX                 (0x3F)
#define I2C_GPIO_CONFIG1_EX                 (0xFF)
#define I2C_GPIO_CONFIG0_UI1                (0x0F)
#define I2C_GPIO_CONFIG1_UI1                (0xFF)
#define I2C_GPIO_CONFIG0_UI2                (0xBF)
#define I2C_GPIO_CONFIG1_UI2                (0xFF)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static OS_U8 ticks_per_us =    SYS_CLOCK_SPEED_MHZ;

// Pinmux definitions
const PIN_CONFIG  i2c1_pin_mux[] = {
  { CONTROL_PADCONF_hsusb0_data7,  MODE0,  PUPD_EN, PULL_DOWN | INPUT_EN },  // i2c1_scl
  { CONTROL_PADCONF_i2c1_sda,      MODE0,  PUPD_EN, PULL_UP   | INPUT_EN }   // i2c1_sda
};

const PIN_CONFIG  i2c2_pin_mux[] = {
  { CONTROL_PADCONF_i2c1_sda_HI,   MODE0,  PUPD_EN, PULL_UP   | INPUT_EN },  // i2c2_scl
  { CONTROL_PADCONF_i2c2_sda,      MODE0,  PUPD_EN, PULL_UP   | INPUT_EN }   // i2c2_sda
};

const PIN_CONFIG  i2c3_pin_mux[] = {
  { CONTROL_PADCONF_i2c2_sda_HI,   MODE0,  PUPD_EN, PULL_UP   | INPUT_EN },  // i2c3_scl
  { CONTROL_PADCONF_i2c3_sda,      MODE0,  PUPD_EN, PULL_UP   | INPUT_EN }   // i2c3_sda
};

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

static void _config_pad(OS_U32 reg,OS_U8 mode,OS_U8 pull_en,OS_U8 pull_type) {
  OS_U16 tmp;

  tmp = ((mode & 0x7) |     // pin mode
    ((pull_en & PUPD_EN  )  << 3) | // pull enable
    ((pull_type & PULL_UP)  << 4) | // pull type select
    ((pull_type & INPUT_EN) << 7)); // input enable (bit 8)
  OUT_REGS(reg, tmp);
}

static void _pin_config_mux(const PIN_CONFIG pin_mux[], const OS_U32 no_of_pins) {
  OS_U32 i;
  for(i=0;i<no_of_pins;i++) {
    _config_pad(pin_mux[i].reg_addr,
         pin_mux[i].mode,
         pin_mux[i].pull_en,
         pin_mux[i].pull_type);
  }
}

static OS_U32 sys_clock_divide_value(void) {
   return IN_REGL(PRM_CLCKSRC_CTRL) & PRM_CLCKSRC_CTRL_SYSCLKDIV;
}

static OS_U32 USTIMER_get_for_timer(OS_U32 timer) {
  return (IN_REGL(timer + TCRR) / ticks_per_us);
}

static void USTIMER_set_for_timer(OS_U32 timer, OS_U32 in_time) {
  OUT_REGL(timer + TLDR, ticks_per_us * in_time);
  OUT_REGL(timer + TTGR, 0x1);
}

static void USTIMER_start(OS_U32 timer) {
  SETBIT_REGL(timer + TCLR, ST);
}

static void USTIMER_stop(OS_U32 timer) {
  CLRBIT_REGL(timer + TCLR, ST);
}

static void USTIMER_reset_for_timer(OS_U32 timer) {
  USTIMER_set_for_timer(timer, 0);
}

static void USTIMER_delay(OS_U32 in_delay) {
  USTIMER_stop(GPTIMER_DELAY);
  USTIMER_reset_for_timer(GPTIMER_DELAY);

  USTIMER_start(GPTIMER_DELAY);

  while (USTIMER_get_for_timer(GPTIMER_DELAY) < in_delay) {
     asm volatile("nop");
  }
}

static void USTIMER_soft_reset_and_free(OS_U32 timer) {
  switch(timer) {
  case GPTIMER10_REG_BASE:
    SETBIT_REGL(CM_CLKSEL_CORE, (1 << 6)); // use 26 Mhz system clock
    break;
  case GPTIMER11_REG_BASE:
    SETBIT_REGL(CM_CLKSEL_CORE, (1 << 7)); // use 26 Mhz system clock
    break;
  default:
    // invalid timer selected
    break;
  }

  // Check the sys clock divide setting to adjust tick counting
  if (sys_clock_divide_value() == PRM_CLCKSRC_CTRL_SYSCLKDIV_2) {
    ticks_per_us = SYS_CLOCK_SPEED_MHZ / 2; // 13 Mhz
  } else { // PRM_CLCKSRC_CTRL_SYSCLKDIV_1
     ticks_per_us = SYS_CLOCK_SPEED_MHZ; // 26 Mhz
  }

  SETBIT_REGL(timer + TIOCP_CFG, SOFTRESET);
  while (!IN_REGL(timer + TISTAT));

  SETBIT_REGL(timer + TSICR, SFT);
  while (!IN_REGL(timer + TISTAT));

  SETBIT_REGL(timer + TIOCP_CFG, EMUFREE);
}

static void USTIMER_init(void) {
  // configure a free-running timer.
  USTIMER_soft_reset_and_free(GPTIMER_FREE);
  USTIMER_start(GPTIMER_FREE);

  // configure a timer to use for delays.
  USTIMER_soft_reset_and_free(GPTIMER_DELAY);
}

static void I2C_clear(OS_U32 i2c) {
  // put i2c in reset.
  OUT_REGS(i2c + I2C_CON, 0);
}

static void I2C_enable(OS_U32 i2c, OS_U16 i2c_con_value, OS_U8 check) {
   OS_U8 count = 100;

   // Enable I2C and wait for it to come out of reset
   OUT_REGS(i2c + I2C_CON, (I2C_CON_I2C_EN | i2c_con_value));
   while (check && ((IN_REGS(i2c + I2C_SYSS) & I2C_SYSS_RDONE) == 0)) {
      USTIMER_delay(10);
      count--;
      if (!count) {
         break;
      }
   }
}

static OS_U32 I2C_read_with_command(OS_U32 i2c, OS_U16 in_addr, OS_U8 command_addr, OS_U8 *dest_buffer, OS_U16 in_length) {
  // write the command (sub-address) to the device
  I2C_write(i2c, in_addr, &command_addr, 1);
  return I2C_read(i2c, in_addr, dest_buffer, in_length);
}

static void I2CGPIO_getInput(OS_U32 i2c, OS_U16 in_addr, OS_U8 in_pin_num, OS_U8 *data) {
  // check address and pin number.
  if (
     (
        (
           (I2C_ADDR_GPIO_EX == in_addr)
           &&
           (I2C_PORT_GPIO_EX == i2c)
        )
        ||
        (
           ((I2C_ADDR_GPIO_UI1 == in_addr) || (I2C_ADDR_GPIO_UI2 == in_addr))
           &&
           (I2C_PORT_GPIO_UI == i2c)
        )
     )
     &&
     (in_pin_num < I2C_GPIO_PIN_MAX)
     &&
     (data != NULL)
  )
  {
    OS_U8 i2c_data;
    OS_U8 gpio_bit = 0;
    OS_U8 i2c_command;
    // set command byte to read appropriate input.
    if (in_pin_num < 8) {
      i2c_command = CMD_BYTE_INPUT0;
       gpio_bit = 1 << in_pin_num;
    } else {
       i2c_command = CMD_BYTE_INPUT1;
       gpio_bit = 1 << (in_pin_num - 8);
    }
     // read the gpio data.
    I2C_read_with_command(i2c, in_addr, i2c_command, &i2c_data, 1);
     // check the input pin value and set var.
    if (i2c_data & gpio_bit) {
      *data = 1;
    } else {
      *data = 0;
    }
  }
}

static OS_U32 I2C_check_status(OS_U32 i2c) {
  OS_U32 rtn = ERR_NO_ERROR;
  OS_U16 i2c_status = IN_REGS(i2c + I2C_STAT);

  // Arbitration lost
  if (i2c_status & I2C_STAT_AL) {
    // Clear the bit
    //debug_print("I2C - Arbitration lost: 0x%x\r\n", i2c_status);
    OUT_REGS(i2c + I2C_STAT, I2C_STAT_AL);
    rtn = ERR_I2C_ERROR;
  }

  // No-acknowledge
  if (i2c_status & I2C_STAT_NACK) {
    // Clear the bit
    OUT_REGS(i2c + I2C_STAT, I2C_STAT_NACK);
    rtn = ERR_I2C_ERROR;
  }

  // Transmit under-run
  if (i2c_status & I2C_STAT_XUDF) {
    //debug_print("I2C - Under-run: 0x%x\r\n", i2c_status);
    OUT_REGS(i2c + I2C_STAT, I2C_STAT_XUDF);
    rtn = ERR_I2C_ERROR;
  }

  // Receiving over-run
  if (i2c_status & I2C_STAT_ROVR) {
    //debug_print("I2C - Over-run: 0x%x\r\n", i2c_status);
    OUT_REGS(i2c + I2C_STAT, I2C_STAT_ROVR);
    rtn = ERR_I2C_ERROR;
  }

  return rtn;
}

static OS_U32 I2C_read(OS_U32 i2c, OS_U16 in_addr, OS_U8 *dest_buffer, OS_U16 in_length) {
  OS_U16 i2c_con_val = 0, i2c_stat_val = 0, rxstat = 0, op_mode;
  OS_U32 i2c_ret = ERR_NO_ERROR, tmpcnt = 0;
  OS_U32 timeout = 0, attempts = 0;
  OS_U16 fifodepth, transmission_finished = 0;
  OS_U8 index = 0, subindex = 0;

  while ((attempts++ < MAX_NUM_ATTEMPTS) && !transmission_finished) {
    timeout = 0;
    // poll bus busy bit in status register
    while (IN_REGS(i2c + I2C_STAT) & I2C_STAT_BB) {
       if (timeout++ > MAX_TIMEOUT_COUNT) {
          I2C_clear(i2c);
          return ERR_I2C_ERROR;
       }
       USTIMER_delay(1);
    }

    // write the slave address
    OUT_REGS(i2c + I2C_SA, (0x007f & in_addr));
    // write the payload size
    OUT_REGS(i2c + I2C_CNT, in_length);

    // Disable all interrupts
    OUT_REGS(i2c + I2C_IE, 0);

    // I2C_BUFSTAT_OFFSET[15:14] contains fifodepth
    fifodepth = (IN_REGS(i2c + I2C_BUFSTAT) & 0xC000) >> 14;
    // I2C_BUFSTAT_FIFODEPTH:   0x0=8byte, 0x1=16 bytes, 0x2=32bytes, 0x3=64bytes
    fifodepth = 8 << fifodepth;
    OUT_REGS(i2c + I2C_BUF, ((((fifodepth / 2 - 1) << 8) & 0x3F00) | 0x4000));

    op_mode = 0;

    // This is a suggested work around..
    i2c_con_val = (I2C_CON_I2C_EN | op_mode | I2C_CON_MST | I2C_CON_STT);
    I2C_enable(i2c, i2c_con_val, 0);
    timeout = 0;
    while (IN_REGS(i2c + I2C_CON) & I2C_CON_STT) {
       if (timeout++ > MAX_TIMEOUT_COUNT) {
          I2C_clear(i2c);
          goto dead_i2c;
       }
    }
    i2c_con_val = (I2C_CON_I2C_EN | op_mode | I2C_CON_MST | I2C_CON_STP);
    I2C_enable(i2c, i2c_con_val, 1);

    tmpcnt = 0;
    while (tmpcnt++ < MAX_TIMEOUT_COUNT) {
       i2c_stat_val = IN_REGS(i2c + I2C_STAT);
       // in case of our CPU running at a high speed and i2c is slow, we'd see this
       if (0 == i2c_stat_val) {
          USTIMER_delay(1);
          continue;
       }
       else if ((i2c_stat_val & I2C_STAT_NACK)) {
          // Considered to be a transmission error in this design
          break;
       }
       else if (i2c_stat_val & I2C_STAT_AL) {
          // Not relevant - not a Multimaster device
          OUT_REGS(i2c + I2C_STAT, I2C_STAT_AL);
       }
       else if (i2c_stat_val & I2C_STAT_ARDY) {
          OUT_REGS(i2c + I2C_STAT, I2C_STAT_ARDY);
          transmission_finished = 1;
          break;
       }
       else if (i2c_stat_val & (I2C_STAT_RDR | I2C_STAT_RRDY)) {
          if (index < in_length) {
            // Note: "index" equals "rtrsh" multiplied with a positive
            // whole number Read I2C_BUFSTAT[13:8] RXSTAT to check the
            // amount of data left to be transmitted
            rxstat = i2c_stat_val & (I2C_STAT_RDR) ? (0x3F00 & IN_REGS(i2c + I2C_BUFSTAT)) >> 8 : fifodepth / 2;

            // Read from I2C_DATA register for RXSTAT times
            for (subindex = 0; (subindex < rxstat) && (index < in_length); index++) {
               ++subindex;
               dest_buffer[index] = IN_REGB(i2c + I2C_DATA);
            }
            OUT_REGS(i2c + I2C_STAT, (i2c_stat_val & (I2C_STAT_RDR | I2C_STAT_RRDY)));
            tmpcnt = 0;
          }
          continue;
       }
       else if (i2c_stat_val & (I2C_STAT_BF | I2C_STAT_BB)) {
          OUT_REGS(i2c + I2C_STAT, (i2c_stat_val & (I2C_STAT_BF | I2C_STAT_BB)));
          continue;
       } else {
          // (NACK + no AL +) no ARDY + no RDR + no RRDY
          return ERR_I2C_ERROR;
       }
    }
    if (tmpcnt >= MAX_TIMEOUT_COUNT) {
       I2C_clear(i2c);
       continue;
    }
dead_i2c:

    i2c_ret = I2C_check_status(i2c);
    if (i2c_ret != ERR_NO_ERROR) {
      // Reset I2C module
      I2C_clear(i2c);
    }
      timeout = 0;
  } // End of while

  if ((attempts > MAX_NUM_ATTEMPTS) && (!transmission_finished)) {
    i2c_ret = ERR_I2C_ERROR;
  }
  I2C_clear(i2c);

  return i2c_ret;
}

static OS_U32 I2C_write(OS_U32 i2c, OS_U16 in_addr, OS_U8 *src_buffer, OS_U16 in_length) {
  OS_U16 i2c_con_val = 0, i2c_stat_val = 0, txstat = 0, op_mode;
  OS_U32 i2c_ret = ERR_NO_ERROR, tmpcnt = 0;
  OS_U32 timeout = 0, attempts = 0;
  OS_U16 fifodepth, transmission_finished = 0;
  OS_U8 index = 0;
  OS_U8 subindex = 0;

  while ((attempts++ < MAX_NUM_ATTEMPTS) && !transmission_finished) {
    OUT_REGS(i2c + I2C_CON, 0);

    // I2C_BUFSTAT_OFFSET[15:14] contains fifodepth
    // I2C_BUFSTAT_FIFODEPTH: 0x0=8byte, 0x1=16 bytes, 0x2=32bytes, 0x3=64bytes
    fifodepth = (IN_REGS(i2c + I2C_BUFSTAT) & 0xC000) >> 14;
    fifodepth = 8 << fifodepth;
    OUT_REGS(i2c + I2C_BUF, (fifodepth - 1) | 0x0040);

    // Disable all interrupts
    OUT_REGS(i2c + I2C_IE, 0);
    // Dummy read
    i2c_stat_val = IN_REGS(i2c + I2C_STAT);
    OUT_REGS(i2c + I2C_SA, i2c_stat_val);
    // configure slave address
    OUT_REGS(i2c + I2C_SA, (0x007f & in_addr));

    // configure data count register
    OUT_REGS(i2c + I2C_CNT, in_length);

    op_mode = 0x0000;
    timeout = 0;

    // poll bus busy bit in status register
    while (IN_REGS(i2c + I2C_STAT) & I2C_STAT_BB) {
       if (timeout++ >MAX_TIMEOUT_COUNT) {
          goto dead_mode;
       }
       USTIMER_delay(1);
    }

    // set master mode, transmit mode, Enable start, stop conditions
    // This is a suggested work around..
    i2c_con_val = (I2C_CON_I2C_EN | I2C_CON_MST | I2C_CON_TRX | I2C_CON_STT);
    I2C_enable(i2c, i2c_con_val, 0);
    while (IN_REGS(i2c + I2C_CON) & I2C_CON_STT) {
       if (timeout++ > MAX_TIMEOUT_COUNT) {
          I2C_clear(i2c);
          goto dead_mode;
       }
    }
    i2c_con_val = (I2C_CON_I2C_EN | op_mode | I2C_CON_MST | I2C_CON_TRX | I2C_CON_STP);
    I2C_enable(i2c, i2c_con_val, 1);

    tmpcnt = 0;
    i2c_stat_val = 0;
    while (tmpcnt++ < MAX_TIMEOUT_COUNT) {
       i2c_stat_val = IN_REGS(i2c + I2C_STAT);
       // in case of our CPU running at a high speed and i2c is slow, we'd see this
       if (0 == i2c_stat_val) {
          USTIMER_delay(1);
          continue;
       }
       else {
          // Considered to be a transmission error in this design
          if ((i2c_stat_val & I2C_STAT_NACK)) {
             break;
          }
          else if (i2c_stat_val & I2C_STAT_AL) {
             // Not relevant - not a Multimaster device
             break;
          }
          else if (i2c_stat_val & I2C_STAT_ARDY) {
             OUT_REGS(i2c + I2C_STAT, I2C_STAT_ARDY);
             transmission_finished = 1;
             break;
          }
          else if (i2c_stat_val & (I2C_STAT_XDR | I2C_STAT_XRDY)) {
            // Note: "index" is equal to "xtrsh" multiplied with a positive
            // whole number Read I2C_BUFSTAT[5:0] TXSTAT to check the amount
            // of data left to be transmitted
            txstat = (i2c_stat_val & I2C_STAT_XRDY) ? fifodepth : 0x003F & IN_REGS(i2c + I2C_BUFSTAT);

            txstat = (0x003F & IN_REGS(i2c + I2C_BUFSTAT));

            // Write I2C_DATA register for TXSTAT times
            for (subindex = 0; (subindex < txstat) && (index < in_length); index++) {
               ++subindex;
               OUT_REGB(i2c + I2C_DATA, src_buffer[index]);

            }
            OUT_REGS(i2c + I2C_STAT, (i2c_stat_val & (I2C_STAT_XDR | I2C_STAT_XRDY)));
            tmpcnt = 0;
            continue;
          }
          else if (i2c_stat_val & (I2C_STAT_BF | I2C_STAT_BB)) {
             OUT_REGS(i2c + I2C_STAT, (i2c_stat_val & (I2C_STAT_BF | I2C_STAT_BB)));
             continue;
          }
          else {
             // (NACK + no AL +) no ARDY + no XDR + no XRDY
             return ERR_I2C_ERROR;
          }
       }
    } // while
    if (tmpcnt >= MAX_TIMEOUT_COUNT) {
       I2C_clear(i2c);
       continue;
    }

dead_mode:
    i2c_ret = I2C_check_status(i2c);
    if (i2c_ret != ERR_NO_ERROR) {
       // Reset I2C module
       I2C_clear(i2c);
    }
    timeout = 0;
  }  // end of while

  if ((attempts > MAX_NUM_ATTEMPTS) && (!transmission_finished)) {
    i2c_ret = ERR_I2C_ERROR;
  }
  I2C_clear(i2c);

  return i2c_ret;
}

static void I2CGPIO_init(OS_U32 i2c, OS_U16 in_addr) {
  if (
     (
        (I2C_ADDR_GPIO_EX == in_addr)
        &&
        (I2C_PORT_GPIO_EX == i2c)
     )
     ||
     (
        ((I2C_ADDR_GPIO_UI1 == in_addr) || (I2C_ADDR_GPIO_UI2 == in_addr))
        &&
        (I2C_PORT_GPIO_UI == i2c)
     )
  )
  {
    OS_U8 i2c_data[3];

    // make sure polarity is not inverted.
    i2c_data[0] = CMD_BYTE_POLARITY0;
    i2c_data[1] = 0;
    i2c_data[2] = 0;
    I2C_write(i2c, in_addr, i2c_data, 3);

    // set config regs on I/O expander.
    if (I2C_ADDR_GPIO_EX == in_addr && (I2C_PORT_GPIO_EX == i2c)) {
      i2c_data[0] = CMD_BYTE_CONFIG0;
      i2c_data[1] = I2C_GPIO_CONFIG0_EX;
      i2c_data[2] = I2C_GPIO_CONFIG1_EX;
    } else if (I2C_ADDR_GPIO_UI1 == in_addr && (I2C_PORT_GPIO_UI == i2c)) {
      i2c_data[0] = CMD_BYTE_CONFIG0;
      i2c_data[1] = I2C_GPIO_CONFIG0_UI1;
      i2c_data[2] = I2C_GPIO_CONFIG1_UI1;
    } else {
      i2c_data[0] = CMD_BYTE_CONFIG0;
      i2c_data[1] = I2C_GPIO_CONFIG0_UI2;
      i2c_data[2] = I2C_GPIO_CONFIG1_UI2;
    }

    I2C_write(i2c, in_addr, i2c_data, 3);
  }
}

static OS_U32 I2C_init(OS_U32 i2c, i2c_clk_e in_clock_rate) {
  OS_U16 resetDone = 0;
  // set the pinmux for the given i2c port.
  switch ((OS_U32)i2c) {
     case I2C1_REG_BASE:
        LOAD_PINMUX(i2c1_pin_mux);
        break;

     case I2C2_REG_BASE:
        LOAD_PINMUX(i2c2_pin_mux);
        break;

     case I2C3_REG_BASE:
        LOAD_PINMUX(i2c3_pin_mux);
        break;

     default:
        return (ERR_INIT_FAIL);
  }

  OUT_REGS(i2c + I2C_CON, 0x0);
  OUT_REGS(i2c + I2C_SYSC, I2C_SYSC_SRST);
  OUT_REGS(i2c + I2C_CON, I2C_CON_I2C_EN);
  do {
     USTIMER_delay(10);
     resetDone = IN_REGS(i2c + I2C_SYSS);
     resetDone &= I2C_SYSS_RDONE;
  } while (!resetDone);

  I2C_clear(i2c);

  // configure clocks.
  switch (in_clock_rate) {
    // set prescaler and clock dividers to precomputed
    // values for input clock rate.
    case I2C_CLK_100K:
      OUT_REGS(i2c + I2C_PSC, 0x7);
      OUT_REGS(i2c + I2C_SCLL, 0x35);
      OUT_REGS(i2c + I2C_SCLH, 0x37);
      break;

    case I2C_CLK_400K:
      OUT_REGS(i2c + I2C_PSC, 0x4);
      OUT_REGS(i2c + I2C_SCLL, 0x12);
      OUT_REGS(i2c + I2C_SCLH, 0x12);
      break;
  }

  OUT_REGS(i2c + I2C_OA0, (I2C_OA0_ADDRESS | I2C_OA0_MCODE));

  // release i2c from reset.
  I2C_enable(i2c, I2C_CON_I2C_EN, 1);

  return (ERR_NO_ERROR);
}

static void I2CGPIO_setOutput(OS_U32 i2c, OS_U16 in_addr, OS_U8 in_pin_num, OS_U16 in_val) {
  if (
       (
         (
            (I2C_ADDR_GPIO_EX == in_addr)
            &&
            (I2C_PORT_GPIO_EX == i2c)
         )
         ||
         (
            ((I2C_ADDR_GPIO_UI1 == in_addr) || (I2C_ADDR_GPIO_UI2 == in_addr))
            &&
            (I2C_PORT_GPIO_UI == i2c)
         )
      )
      &&
      (in_pin_num < I2C_GPIO_PIN_MAX)
  )
  {
    OS_U8 i2c_data[2];
    OS_U32 gpio_bit = 0;
    OS_U8 i2c_command;

    // set command byte to read appropriate output, so we do not change
    // any data that we do not want to.
    if (in_pin_num < 8) {
       i2c_command = CMD_BYTE_OUTPUT0;
       gpio_bit = 1 << in_pin_num;
    } else {
      i2c_command = CMD_BYTE_OUTPUT1;
      gpio_bit = 1 << (in_pin_num - 8);
    }

    // read the gpio data.
    I2C_read_with_command(i2c, in_addr, i2c_command, &i2c_data[1], 1);

    // update the data to set/clr bit for pin num.
    i2c_data[0] = i2c_command;
    if (in_val) {
      SETBIT(i2c_data[1], gpio_bit);
    } else {
      CLRBIT(i2c_data[1], gpio_bit);
    }

    // write the gpio data back to the I/O expander.
    I2C_write(i2c, in_addr, i2c_data, 2);
  }
}

static void convertLedToI2CGpio(OS_U8 in_led_num, OS_U16 *in_pin_num) {
  switch (in_led_num) {
    case LED_1:
      *in_pin_num = I2C_GPIO_EX_LED1;
      break;

    case LED_2:
      *in_pin_num = I2C_GPIO_EX_LED2;
      break;
  }
}

static void LED_init(void) {
  // init experimenter i/o expander.
  I2CGPIO_init(I2C_PORT_GPIO_EX, I2C_ADDR_GPIO_EX);

  // turn all leds off.
  I2CGPIO_setOutput(I2C_PORT_GPIO_EX, I2C_ADDR_GPIO_EX, I2C_GPIO_EX_LED1, 1);
  I2CGPIO_setOutput(I2C_PORT_GPIO_EX, I2C_ADDR_GPIO_EX, I2C_GPIO_EX_LED2, 1);
}


static void LED_turnOn(OS_U8 in_led_num) {
  OS_U16 pin_num;

  convertLedToI2CGpio(in_led_num, &pin_num);
  // leds are active low...set to 0 to turn on.
  I2CGPIO_setOutput(I2C_PORT_GPIO_EX, I2C_ADDR_GPIO_EX, pin_num, 0);
}

static void LED_turnOff(OS_U8 in_led_num) {
  OS_U16 pin_num;
  convertLedToI2CGpio(in_led_num, &pin_num);
  // leds are active low...set to 1 to turn off.
  I2CGPIO_setOutput(I2C_PORT_GPIO_EX, I2C_ADDR_GPIO_EX, pin_num, 1);
}

static void LED_isOn(OS_U8 in_led_num, OS_U8 *led_status) {
  if (in_led_num < MAX_LED_NUM) {
    OS_U16 pin_num;

    convertLedToI2CGpio(in_led_num, &pin_num);

    I2CGPIO_getInput(I2C_PORT_GPIO_EX, I2C_ADDR_GPIO_EX, pin_num, led_status);
    if (*led_status) {
     // i2c gpio pin is high, led is off.
     *led_status = 0;
    } else {
      // i2c gpio pin is low, led is on.
      *led_status = 1;
    }
  }
}

static void LED_toggle(OS_U8 in_led_num) {
  OS_U8 led_status;

  LED_isOn(in_led_num, &led_status);

  if (led_status) {
    LED_turnOff(in_led_num);
  } else {
    LED_turnOn(in_led_num);
  }
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_Init()
*/
void BSP_Init(void) {
  USTIMER_init();
  I2C_init(I2C2, I2C_CLK_100K);
  LED_init();
}

/*********************************************************************
*
*       BSP_SetLED()
*/
void BSP_SetLED(int Index) {
  if (Index < MAX_LED_NUM) {
    LED_turnOn(Index);
  }
}

/*********************************************************************
*
*       BSP_ClrLED()
*/
void BSP_ClrLED(int Index) {
  if (Index < MAX_LED_NUM) {
    LED_turnOff(Index);
  }
}

/*********************************************************************
*
*       BSP_ToggleLED()
*/
void BSP_ToggleLED(int Index) {
  if (Index < MAX_LED_NUM) {
    LED_toggle(Index);
  }
}

/****** End Of File *************************************************/

