/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose : UART implementation

Additional information:

  Device : AM3517
  Board  : TMDSEVM3517C

  Unit | UART   | Board connector
  ===============================
  0    | UART3  | DEBUG SERIAL

*/

#include "BSP_UART.h"
#include "RTOS.h"      // For OS_ARM_* and OS_INT_* functions.

/*********************************************************************
*
*       Defines, configurable
*
**********************************************************************
*/

#define UART_FREQ                    (48000000uL)

/*********************************************************************
*
*       Defines, fixed
*
**********************************************************************
*/

#define UART3                        (0)

//
// UART 3
//
#define UART3_BASE_ADDR                      (0x49020000uL)
#define UART3_DLL                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x000uL))
#define UART3_RHR                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x000uL))
#define UART3_THR                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x000uL))
#define UART3_DLH                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x004uL))
#define UART3_IER                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x004uL))
#define UART3_IIR                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x008uL))
#define UART3_EFR                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x008uL))
#define UART3_LCR                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x00CuL))
#define UART3_MCR                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x010uL))
#define UART3_SPR_TLR                        (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x01CuL))
#define UART3_SCR                            (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x040uL))
#define UART3_SYSC                           (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x054uL))
#define UART3_SYSS                           (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x058uL))
#define UART3_MDR1                           (*(volatile unsigned short*)(UART3_BASE_ADDR + 0x020uL))

#define CONTROL_PADCONF_MCBSP1_CLK_16_31     (*(volatile unsigned short*)0x4800219AuL)
#define CONTROL_PADCONF_UART3_RTS_SD_00_15   (*(volatile unsigned short*)0x4800219CuL)
#define CONTROL_PADCONF_UART3_RTS_SD_16_31   (*(volatile unsigned short*)0x4800219EuL)
#define CONTROL_PADCONF_UART3_TX_IRTX_00_15  (*(volatile unsigned short*)0x480021A0uL)

#define UART_BRR_VALUE(Baudrate) (UART_FREQ / 16 / (Baudrate))

#define OVF_IT_ENA                          (1uL << 1)            // Overflow interrupt enable
#define OVF_IT_FLAG                         (1uL << 1)            // Overflow interrupt flag
#define LCR_CONF_MODE_A                     0x80                  // Switch device to configuration mode A
#define LCR_CONF_MODE_B                     0xBF                  // Switch device to configuration mode B
#define MDR1_MODE_SELECT_UART16             0x00                  // UART mode selection UART 16x mode
#define FCR_FIFO_EN                         (1uL << 0)            // Set the FIFO enable bit, enables TX/RX FIFO
#define FCR_RX_FIFO_CLEAR                   (1uL << 11)           // Clears the receive FIFO buffer
#define FCR_TX_FIFO_CLEAR                   (1uL << 2)            // Clears the transmit FIFO buffer
#define FCR_DMA_MODE                        (1uL << 3)            // Enables DMA mode
#define FCR_TXFIFO_TRIG_LSB(x)              (x << 4)              // Two LSB bits of transmit FIFO trigger level
#define FCR_RXFIFO_TRIG_LSB(x)              (x << 6)              // Two LSB bits of receive FIFO trigger level
#define TLR_TX_FIFO_TRIG_DMA(x)             (x & 0x0FL )          // Two LSB bits of transmit FIFO trigger level
#define TLR_RX_FIFO_TRIG_DMA(x)             (x << 4)              // Two LSB bits of receive FIFO trigger level
#define MCR_TCR_TLR                         6
#define SCR_TX_TRIG_GRANU1                  6                     // Set Granularity of TX FIFO trigger level to 1
#define SCR_RX_TRIG_GRANU1                  7                     // Set Granularity of RX FIFO trigger level to 1
#define IER_RHR_IT                          0                     // Enable transmit holding register interrupt
#define IER_THR_IT                          1                     // Enable transmit holding register interrupt
#define IIR_IT_TYPE_MASK                    (0x1FuL << 1)         // Mask Bit for reading out interrupt type
#define IIR_IT_TYPE_THR                     (0x01uL << 1)         // Transmit holding register interrupt
#define IIR_IT_TYPE_RHR                     (0x02uL << 1)         // Receive holding register interrupt
#define IIR_IT_TYPE_RXTIMEOUT               (0x06uL << 1)         // Receive Timeout interrupt

#define UART3_IRQ_ID                        (74)
#define UART3_IRQ_PRIO                      (63)

#define PER_CM                              (0x48005000)
#define CM_FCLKEN_PER                       (PER_CM)
#define CM_ICLKEN_PER                       (PER_CM + 0x10)

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

static BSP_UART_TX_CB* _pfWriteCB;
static BSP_UART_RX_CB* _pfReadCB;

/*********************************************************************
*
*       Prototypes
*
*  Declare ISR handler here to avoid "no prototype" warning.
*  They are not declared in any CMSIS header.
*
**********************************************************************
*/

#if defined(__cplusplus)
  extern "C" {                // Make sure we have C-declarations in C++ programs.
#endif

static void UART3_RX_TX_IRQHandler(void);

#if defined(__cplusplus)
}                             // Make sure we have C-declarations in C++ programs.
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _SetBaudrate()
*
*  Function description
*    Configures the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
static void _SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  BSP_UART_USE_PARA(Unit);
  BSP_UART_USE_PARA(Baudrate);
}

/*********************************************************************
*
*       Global functions, IRQ handler
*
**********************************************************************
*/

/*********************************************************************
*
*       UART3_RX_TX_IRQHandler()
*
*  Function description
*    UART Rx & Tx interrupt handler.
*/
static void UART3_RX_TX_IRQHandler(void) {
  unsigned int  Unit;

  Unit = UART3;
  switch (UART3_IIR & IIR_IT_TYPE_MASK) {
    case IIR_IT_TYPE_THR:
      if ((_pfWriteCB != 0) && (_pfWriteCB(Unit) != 0)) {  // No more characters to send ?
        UART3_IER &= ~(1uL << IER_THR_IT);                   // Disable Tx Interrupt
      }
      break;
    case IIR_IT_TYPE_RHR:
      if (_pfReadCB != 0) {
        _pfReadCB(Unit, UART3_RHR);
      }
      break;
    default:
      break;
  }
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_UART_Init()
*
*  Function description
*    Initializes the UART peripheral.
*
*  Parameters
*    Unit       : Unit number (typically zero-based).
*    Baudrate   : Baud rate to configure [Hz].
*    NumDataBits: Number of data bits to use.
*    Parity     : One of the following values:
*                   * BSP_UART_PARITY_NONE
*                   * BSP_UART_PARITY_ODD
*                   * BSP_UART_PARITY_EVEN
*    NumStopBits: Number of stop bits to use.
*/
void BSP_UART_Init(unsigned int Unit, unsigned long Baudrate, unsigned char NumDataBits, unsigned char Parity, unsigned char NumStopBits) {
  volatile int i;

  BSP_UART_USE_PARA(NumDataBits);
  BSP_UART_USE_PARA(Parity);
  BSP_UART_USE_PARA(NumStopBits);
  OS_INT_IncDI();
  switch (Unit) {
  case UART3:
    //
    // Enable clocks
    //
    (*(volatile unsigned long*)CM_FCLKEN_PER) |= (1uL << 11);
    (*(volatile unsigned long*)CM_ICLKEN_PER) |= (1uL << 11);
    //
    // Initialize uart
    //
    for (i = 0; i < 0xFFFF; i++);                  // Wait a short time
    UART3_SYSC          |= (1uL << 1);               // Initiate a software reset
    while ((UART3_SYSS & (1uL << 0)) == 0);          // Wait for the end of the reset operation
    UART3_MDR1           = MDR1_MODE_SELECT_UART16;
    UART3_LCR            = LCR_CONF_MODE_B;          // Switch to configuration mode b
    UART3_EFR           |= (1uL << 4);               // Enable hardware flow control if desired
    UART3_LCR            = LCR_CONF_MODE_A;          // Switch to configuration mode a
    UART3_MCR           |= (1uL << MCR_TCR_TLR);     // Enable access to TCR and TLR
    UART3_SCR           |= (1uL << SCR_TX_TRIG_GRANU1) | (1uL << SCR_RX_TRIG_GRANU1);
    UART3_IIR            = FCR_FIFO_EN | FCR_RX_FIFO_CLEAR | FCR_TX_FIFO_CLEAR | FCR_TXFIFO_TRIG_LSB(0) | FCR_RXFIFO_TRIG_LSB(1);
    UART3_SPR_TLR        = TLR_TX_FIFO_TRIG_DMA(2) | TLR_RX_FIFO_TRIG_DMA(0);
    UART3_MDR1           = MDR1_MODE_SELECT_UART16;
    UART3_DLL            = (unsigned char) (UART_BRR_VALUE(Baudrate)       & 0xFFuL);
    UART3_DLH            = (unsigned char)((UART_BRR_VALUE(Baudrate) >> 8) & 0xFFuL);
    UART3_LCR            = 0x13uL;                   // Switch to operation mode, Set uart to 8N1
    UART3_IER           |= (1uL << IER_RHR_IT);      // Enable Rx interrupt
    //
    // Set pad configuration
    //
    CONTROL_PADCONF_MCBSP1_CLK_16_31    = 0x0118uL;
    CONTROL_PADCONF_UART3_RTS_SD_00_15  = 0x0118uL;
    CONTROL_PADCONF_UART3_RTS_SD_16_31  = 0x0118uL;
    CONTROL_PADCONF_UART3_TX_IRTX_00_15 = 0x0118uL;
    //
    // Install uart isr
    //
    OS_ARM_InstallISRHandler(UART3_IRQ_ID, (OS_ISR_HANDLER*)UART3_RX_TX_IRQHandler);
    OS_ARM_ISRSetPrio       (UART3_IRQ_ID, UART3_IRQ_PRIO);
    OS_ARM_EnableISR        (UART3_IRQ_ID);
    break;
  }
  OS_INT_DecRI();
}

/*********************************************************************
*
*       BSP_UART_DeInit()
*
*  Function description
*    De-initializes the UART peripheral.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*/
void BSP_UART_DeInit(unsigned int Unit) {
  BSP_UART_USE_PARA(Unit);
}

/*********************************************************************
*
*       BSP_UART_SetBaudrate()
*
*  Function description
*    Configures/changes the UART baud rate.
*
*  Parameters
*    Unit    : Unit number (typically zero-based).
*    Baudrate: Baud rate to configure [Hz].
*/
void BSP_UART_SetBaudrate(unsigned int Unit, unsigned long Baudrate) {
  _SetBaudrate(Unit, Baudrate);
}

/*********************************************************************
*
*       BSP_UART_SetReadCallback()
*
*  Function description
*    Sets the callback to execute upon an Rx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetReadCallback(unsigned Unit, BSP_UART_RX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfReadCB = pf;
}

/*********************************************************************
*
*       BSP_UART_SetWriteCallback()
*
*  Function description
*    Sets the callback to execute upon a Tx interrupt.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    pf  : Callback to execute.
*/
void BSP_UART_SetWriteCallback(unsigned int Unit, BSP_UART_TX_CB* pf) {
  BSP_UART_USE_PARA(Unit);
  _pfWriteCB = pf;
}

/*********************************************************************
*
*       BSP_UART_Write1()
*
*  Function description
*    Sends one byte via UART.
*
*  Parameters
*    Unit: Unit number (typically zero-based).
*    Data: (First) data byte to send.
*
*  Additional information
*    The first byte of a transfer is typically sent from application
*    context. Further bytes of the transfer are then sent from the
*    Tx interrupt handler by also calling this function from interrupt
*    context.
*/
void BSP_UART_Write1(unsigned int Unit, unsigned char Data) {
  switch (Unit) {
  case UART3:
    UART3_IER |= (1uL << IER_THR_IT);  // Enable Tx interrupt
    UART3_THR  = Data;
    break;
  }
}

/*************************** End of file ****************************/
