/*********************************************************************
*                    SEGGER Microcontroller GmbH                     *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*            (c) 1995 - 2019 SEGGER Microcontroller GmbH             *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system for microcontrollers      *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: Internal                                         *
*                                                                    *
*       Current version number will be inserted here                 *
*       when shipment is built.                                      *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose   : Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Configuration
*/

#ifndef   ALLOW_NESTED_INTERRUPTS
  #define ALLOW_NESTED_INTERRUPTS  (0)  // Caution: Nesting interrupts will cause higher stack load on system stack CSTACK
#endif

/*********************************************************************
*
*       Clock settings
*/
#define OS_SYSTEM_FREQ     (220000000uL)

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ      (OS_SYSTEM_FREQ / 4)
#define OS_TICK_FREQ       (1000u)
#define OS_INT_FREQ        (OS_TICK_FREQ)
#define OS_TIMER_RELOAD    (OS_TIMER_FREQ / OS_INT_FREQ)

/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/
#define PLLCTL1       (*(volatile unsigned int*)(0xFFFFFF70))   // PLL Control 1 Register
#define PLLCTL2       (*(volatile unsigned int*)(0xFFFFFF74))   // PLL Control 2 Register
#define PLLCTL3       (*(volatile unsigned int*)(0xFFFFE100))   // PLL Control 3 Register
#define CSDIS         (*(volatile unsigned int*)(0xFFFFFF30))   // Clock Source Disable Register
#define CLKCNTL       (*(volatile unsigned int*)(0xFFFFFFD0))   // Clock Control Register
#define CLK2CNTRL     (*(volatile unsigned int*)(0xFFFFE13C))   // Clock 2 Control Register
#define PENA_MASK     (1u << 8)
#define PSPWRDWNCLR0  (*(volatile unsigned int*)(0xFFFFE0A0))   // Peripheral Power-Down Clear Register 0
#define PSPWRDWNCLR1  (*(volatile unsigned int*)(0xFFFFE0A4))   // Peripheral Power-Down Clear Register 1
#define PSPWRDWNCLR2  (*(volatile unsigned int*)(0xFFFFE0A8))   // Peripheral Power-Down Clear Register 2
#define PSPWRDWNCLR3  (*(volatile unsigned int*)(0xFFFFE0AC))   // Peripheral Power-Down Clear Register 3
#define FRDCNTL       (*(volatile unsigned int*)(0xFFF87000))   // Flash Option Control Register
#define FSM_WR_ENA    (*(volatile unsigned int*)(0xFFF87288))   // FSM Register Write Enable
#define EEPROM_CONFIG (*(volatile unsigned int*)(0xFFF872B8))   // EEPROM Emulation Configuration Register
#define FBFALLBACK    (*(volatile unsigned int*)(0xFFF87040))   // Flash Bank Fallback Power Register
#define CDDIS         (*(volatile unsigned int*)(0xFFFFFF3C))   // Clock Domain Disable Register
#define CSVSTAT       (*(volatile unsigned int*)(0xFFFFFF54))   // Clock Source Valid Status Register
#define GHVSRC        (*(volatile unsigned int*)(0xFFFFFF48))   // GCLK, HCLK, VCLK, and VCLK2 Source Register
#define RCLKSRC       (*(volatile unsigned int*)(0xFFFFFF50))   // RTI Clock Source Register
#define VCLKASRC      (*(volatile unsigned int*)(0xFFFFFF4C))   // Peripheral Asynchronous Clock Source Register
#define VCLKACON1     (*(volatile unsigned int*)(0xFFFFE140))   // Peripheral Asynchronous Clock Configuration 1 Register
#define VCLKR_MASK    (1u << 16)
#define VCLK2R_MASK   (1u << 24)
#define VCLK3R_MASK   (1u << 0)
#define TRUE          (1u)
#define FALSE         (0u)

enum flashWPowerModes {
  SYS_SLEEP   = 0U, // Alias for flash bank power mode sleep
  SYS_STANDBY = 1U, // Alias for flash bank power mode standby
  SYS_ACTIVE  = 3U  // Alias for flash bank power mode active
};
enum systemClockSource {
  SYS_OSC           = 0,  /**< Alias for oscillator clock Source                */
  SYS_PLL1          = 1,  /**< Alias for Pll1 clock Source                      */
  SYS_EXTERNAL1     = 3,  /**< Alias for external clock Source                  */
  SYS_LPO_LOW       = 4,  /**< Alias for low power oscillator low clock Source  */
  SYS_LPO_HIGH      = 5,  /**< Alias for low power oscillator high clock Source */
  SYS_PLL2          = 6,  /**< Alias for Pll2 clock Source                      */
  SYS_EXTERNAL2     = 7,  /**< Alias for external 2 clock Source                */
  SYS_VCLK          = 9   /**< Alias for synchronous VCLK1 clock Source         */
};

#define VCLKACON1      (*(volatile unsigned int*)(0xFFFFE140))   // Peripheral Asynchronous Clock Configuration 1 Register
#define RTIGCTRL       (*(volatile unsigned int*)(0xFFFFFC00))   // RTI Global Control Register Section
#define RTITBCTRL      (*(volatile unsigned int*)(0xFFFFFC04))   // RTI Timebase Control Register Section
#define RTICAPCTRL     (*(volatile unsigned int*)(0xFFFFFC08))   // RTI Capture Control Register Section
#define RTICOMPCTRL    (*(volatile unsigned int*)(0xFFFFFC0C))   // RTI Compare Control Register Section
#define RTIFRC0        (*(volatile unsigned int*)(0xFFFFFC10))   // RTI Free Running Counter 0 Register Section
#define RTIUC0         (*(volatile unsigned int*)(0xFFFFFC14))   // RTI Up Counter 0 Register Section
#define RTICPUC0       (*(volatile unsigned int*)(0xFFFFFC18))   // RTI Compare Up Counter 0 Register Section
#define RTICAFRC0      (*(volatile unsigned int*)(0xFFFFFC20))   // RTI Capture Free Running Counter 0 Register Section
#define RTICAUC0       (*(volatile unsigned int*)(0xFFFFFC24))   // RTI Capture Up Counter 0 Register Section
#define RTICOMP0       (*(volatile unsigned int*)(0xFFFFFC50))   // RTI Compare 0 Register Section
#define RTIUDCP0       (*(volatile unsigned int*)(0xFFFFFC54))   // RTI Update Compare 0 Register Section
#define RTISETINTENA   (*(volatile unsigned int*)(0xFFFFFC80))   // RTI Set Interrupt Enable Register Section
#define RTIINTFLAG     (*(volatile unsigned int*)(0xFFFFFC88))   // RTI Interrupt Flag Register Section

#define RTI_INT_ID           (2u)
#define RTIINTFLAG_INT0_MASK (1u << 0)
#define VIM_CHANNEL          (96u)
#define VIM_RAM_ADDR         (0xFFF82000u) // Fixed vector table start addr in RAM

#define FIRQPR0        (*(volatile unsigned int*)(0xFFFFFE10))   // FIQ/IRQ Program Control Register 0
#define FIRQPR1        (*(volatile unsigned int*)(0xFFFFFE14))   // FIQ/IRQ Program Control Register 1
#define FIRQPR2        (*(volatile unsigned int*)(0xFFFFFE18))   // FIQ/IRQ Program Control Register 2
#define REQENASET0     (*(volatile unsigned int*)(0xFFFFFE30))   // Interrupt Enable Set Register 0
#define REQENASET1     (*(volatile unsigned int*)(0xFFFFFE34))   // Interrupt Enable Set Register 1
#define REQENASET2     (*(volatile unsigned int*)(0xFFFFFE38))   // Interrupt Enable Set Register 2
#define REQENACLR0     (*(volatile unsigned int*)(0xFFFFFE40))   // Interrupt Enable CLR Register 0
#define REQENACLR1     (*(volatile unsigned int*)(0xFFFFFE44))   // Interrupt Enable CLR Register 1
#define REQENACLR2     (*(volatile unsigned int*)(0xFFFFFE48))   // Interrupt Enable CLR Register 2
#define IRQVECREG      (*(volatile unsigned int*)(0xFFFFFE70))   // IRQ Interrupt Vector Register

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value
*
*  Return value
*    Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  OS_U32 CompareValue;

  CompareValue = RTICOMP0;
  return (OS_TIMER_RELOAD + RTIFRC0 - CompareValue);
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set
*
*  Return value
*    == 0; Interrupt pending flag not set
*    != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return (RTIINTFLAG & RTIINTFLAG_INT0_MASK);
}

/*********************************************************************
*
*       _PLL_Init()
*
*  Function description
*    Initialize PLL to speed up system
*    This function is called from __low_level_init() before
*    any variables are initialized.
*    Therefore, it can not rely on any variable content
*    and can not set any variables (except no-init variables).
*/
static void _PLL_Init(void) {
  PLLCTL1 = 0x00000000U
          | 0x20000000U
          | ((0x1F) << 24)      // PLLDIV 0x20 = 32
          | 0x00000000U
          | ((6u - 1u) << 16)   // REFCLKDIV
          | ((165u - 1u) << 8); // PLLMUL

  PLLCTL2 = 0x00000000U
          | (255u << 22)
          | (7u << 12)
          | ((2u - 1u) << 9)    // ODPLL
          | 61u;

  PLLCTL3 = ((2u - 1u) << 29)   // ODPLL
          | ((0x1Fu) << 24)     // PLLDIV
          | ((6u - 1u) << 16)
          | ((165u - 1u) << 8);

  CSDIS = 0x00000000U
        | 0x00000000U
        | 0x00000008U
        | 0x00000080U
        | 0x00000000U
        | 0x00000000U
        | 0x00000000U;
  //
  // Disable Peripherals before peripheral powerup
  //
  CLKCNTL &= ~PENA_MASK;
  //
  // Release peripherals from reset and enable clocks to all peripherals
  // Power-up all peripharals
  //
  PSPWRDWNCLR0 = 0xFFFFFFFFu;
  PSPWRDWNCLR1 = 0xFFFFFFFFu;
  PSPWRDWNCLR2 = 0xFFFFFFFFu;
  PSPWRDWNCLR3 = 0xFFFFFFFFu;
  //
  // Enable Peripherals
  //
  CLKCNTL |= PENA_MASK;
}

/*********************************************************************
*
*       _Flash_Init()
*
* Function description
*   Setup flash
*/
static void _Flash_Init(void) {
  //
  // Setup flash read mode, address wait states and data wait states */
  //
  FRDCNTL = 0x00000000U
          | (3U << 8U)
          | (1U << 4U)
          |  1U;
  //
  // Setup flash access wait states for bank 7
  //
  FSM_WR_ENA    = 0x5;
  EEPROM_CONFIG = 0x00030002;
  FSM_WR_ENA    = 0xA;          // Disable write access to flash state machine registers
  //
  // Setup flash bank power modes
  //
  FBFALLBACK = 0x00000000
             | (SYS_ACTIVE << 14U)
             | (SYS_SLEEP << 12U)
             | (SYS_SLEEP << 10U)
             | (SYS_SLEEP << 8U)
             | (SYS_SLEEP << 6U)
             | (SYS_SLEEP << 4U)
             | (SYS_ACTIVE << 2U)
             |  SYS_ACTIVE;
}

/*********************************************************************
*
*       _MapClocks()
*
* Function description
*   Initialize Clock Tree
*/
static void _MapClocks(void) {
  OS_U32 CSDISValue;

  //
  // Initialize Clock Tree:
  // Disable / Enable clock domain
  //
  CDDIS = (FALSE << 4 )  /* AVCLK 1 OFF */
         |(TRUE << 5 )   /* AVCLK 2 OFF */
         |(FALSE << 8 )  /* VCLK3 OFF   */
         |(FALSE << 10)  /* AVCLK 3 OFF */
         |(FALSE << 11); /* AVCLK 4 OFF */
  //
  // Wait for until clocks are locked
  //
  do {
    CSDISValue = CSDIS;
  } while ((CSVSTAT & ((CSDISValue ^ 0xFF) & 0xFF)) != ((CSDISValue ^ 0xFF) & 0xFF));
  //
  // Now the PLLs are locked and the PLL outputs can be sped up
  // The R-divider was programmed to be 0xF. Now this divider is changed to programmed value
  //
  PLLCTL1 = (PLLCTL1 & 0xE0FFFFFF)|((1u - 1u)<< 24u);
  PLLCTL3 = (PLLCTL3 & 0xE0FFFFFF)|((1u - 1u)<< 24u);

  /** - Map device clock domains to desired sources and configure top-level dividers */
  /** - All clock domains are working off the default clock sources until now */
  /** - The below assignments can be easily modified using the HALCoGen GUI */

  /** - Setup GCLK, HCLK and VCLK clock source for normal operation, power down mode and after wakeup */
  GHVSRC = (SYS_PLL1 << 24U)
         | (SYS_PLL1 << 16U)
         |  SYS_PLL1;
  //
  // Setup synchronous peripheral clock dividers for VCLK1, VCLK2, VCLK3 */
  //
  CLKCNTL   |= VCLKR_MASK | VCLK2R_MASK;
  CLK2CNTRL |= VCLK3R_MASK;
  //
  // Setup RTICLK1 and RTICLK2 clocks
  //
  RCLKSRC = (1U << 24U)
          | (SYS_VCLK << 16U)
          | (1U << 8U)
          |  SYS_VCLK;

  // Setup asynchronous peripheral clock sources for AVCLK1 and AVCLK2
  VCLKASRC  = (SYS_VCLK << 8U) | SYS_VCLK;
  VCLKACON1 = (1U << 24)
            | 1 << 20U
            | (SYS_VCLK << 16)
            | (1U << 8)
            | 1 << 4U
            | SYS_VCLK;
}

/*********************************************************************
*
*       _DefaultInterrupt_Handler()
*
* Function description
*   Default interrupt handler
*/
static void _DefaultInterrupt_Handler(void) {
  //
  // Just enter an endless loop so that user can set a breakpoint here
  //
  while (1) {
  }
}

/*********************************************************************
*
*       _VIM_Init()
*
* Function description
*   Initializes the Vector interrupt manager
*/
static void _VIM_Init(void) {
  OS_ISR_HANDLER** papISR;
  OS_U32 i;
  //
  // Map all interrupts to IRQ
  //
  FIRQPR0 = 0u;
  FIRQPR1 = 0u;
  FIRQPR2 = 0u;
  //
  // Fill vector table with default interrupt handler
  //
  papISR = (OS_ISR_HANDLER**)VIM_RAM_ADDR;
  for (i = 0 ; i < VIM_CHANNEL ; i++) {
    *(papISR + i) = &_DefaultInterrupt_Handler;
  }
}

/*********************************************************************
*
*       _SysTick_Handler()
*
* Function description
*   Systick handler
*/
static void _SysTick_Handler(void) {
  RTIINTFLAG = RTIINTFLAG_INT0_MASK; // Clear the pending interrupts flag
  OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_Process();
#endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*       Initialize the hardware (timer) required for the OS to run.
*       May be modified, if an other timer should be used
*/
void OS_InitHW(void) {
  OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_TICK_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
  OS_INT_IncDI();
  //
  // Initialize Vector interrupt manager
  //
  _VIM_Init();
  //
  // RTI Initialization
  //
  RCLKSRC      = 0x08;             // Clock source 8 i.e VClk, div the clk source 0 by 1
  RTIGCTRL     = 0x00000000;       // Disable RTIUC0 and RTIUC1
  RTIUC0       = 0x00000000;       // Initialize up Counter
  RTIFRC0      = 0x00000000;       // Initialize Free Running Counter
  RTICPUC0     = 0x00000001;       // Set Compare Up Counter value to Prescale RTICLK.
  RTICOMP0     = OS_TIMER_RELOAD;  // Compare value
  RTIUDCP0     = OS_TIMER_RELOAD;  // This value is added to generate new compare value since the timer register continous counting
  RTIINTFLAG   = 0x0000000F;       // Clearing the pending Interrupts Flags
  RTICOMPCTRL  = 0x00000000;       // Initializing the Compare Counter register
  RTISETINTENA = 0x00000001;       // Enabling compare interrupt 0
  RTIGCTRL     = 0x00000001;       // Enable the RTIUC0
  //
  // Set interrupt handler and enable RTI interrupt
  //
  OS_ARM_InstallISRHandler(RTI_INT_ID, _SysTick_Handler);
  OS_ARM_EnableISR(RTI_INT_ID);
  //
  // Setup values for usec precise system time functions
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This is basically the "core" of the idle loop.
*    This core loop can be changed, but:
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved. However, a simple program loop can be programmed
*    (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       OS_GetTime_Cycles()
*
*  Function description
*    This routine is required for task-info via embOSView or high
*    resolution time measurement functions.
*    It returns the system time in timer clock cycles.
*/
OS_U32 OS_GetTime_Cycles(void) {
  OS_U32 Time;
  OS_I32 Cnt;
  OS_U32 CompareValue;

  Time         = OS_TIME_GetTicks32();
  CompareValue = RTICOMP0;
  Cnt          = OS_TIMER_RELOAD + RTIFRC0 - CompareValue;
  //
  // Check if timer interrupt pending ...
  //
  if (RTIINTFLAG & RTIINTFLAG_INT0_MASK) { // Interrupt pending, re-read timer and adjust result
    CompareValue = RTICOMP0;
    Cnt          = OS_TIMER_RELOAD + RTIFRC0 - CompareValue;
    Time++;
  }
  return (OS_TIMER_RELOAD * Time) + Cnt;
}

/*********************************************************************
*
*       OS_ConvertCycles2us()
*
*  Function description
*    Convert Cycles into micro seconds.
*
*    If your clock frequency is not a multiple of 1 MHz,
*    you may have to modify this routine in order to get proper
*    diagnostics.
*
*    This routine is required for profiling or high resolution time
*    measurement only. It does not affect operation of the OS.
*/
OS_U32 OS_ConvertCycles2us(OS_U32 Cycles) {
  return Cycles / (OS_TIMER_FREQ / 1000000);
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       OS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;

  pISR = (OS_ISR_HANDLER*) IRQVECREG;  // Read interrupt vector
#if ALLOW_NESTED_INTERRUPTS
  OS_INT_EnterNestable();              // Now interrupts may be reenabled. If nesting should be allowed
#else
  OS_INT_Enter();                      // Inform OS that interrupt handler is running
#endif
  pISR();                              // Call interrupt service routine
#if ALLOW_NESTED_INTERRUPTS
  OS_INT_LeaveNestable();              // Leave nestable interrupt, perform task switch if required
#else
  OS_INT_Leave();                      // Leave interrupt, perform task switch if required
#endif
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Used to install a specific interrupt vector when ARM
*    CPUs with vectored interrupt controller are used.
*
*  Parameters
*    ISRIndex:    Index of the interrupt source, normally the interrupt vector number.
*    pISRHandler: Address of the interrupt handler function.
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler (int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;
  OS_ISR_HANDLER** papISR;

#if (OS_DEBUG != 0)
  if ((unsigned)ISRIndex >= VIM_CHANNEL) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  papISR = (OS_ISR_HANDLER**)VIM_RAM_ADDR;
  pOldHandler          = *(papISR + (ISRIndex + 1u));
  *(papISR + (ISRIndex + 1)) = pISRHandler;
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Used to enable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be enabled.
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned)ISRIndex >= VIM_CHANNEL) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  if (ISRIndex < 32) {
    REQENASET0 = (1u << ISRIndex);
  } else {
    if (ISRIndex < 64) {
      REQENASET1 = (1u << ISRIndex);
    } else {
      REQENASET2 = (1u << ISRIndex);
    }
  }
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Used to disable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be disabled.
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned)ISRIndex >= VIM_CHANNEL) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif

  OS_INT_Disable();
  if (ISRIndex < 32) {
    REQENACLR0 = (1u << ISRIndex);
  } else {
    if (ISRIndex < 64) {
      REQENACLR1 = (1u << ISRIndex);
    } else {
      REQENACLR2 = (1u << ISRIndex);
    }
  }
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       __low_level_init()
*
*  Function description
*    Initialize memory controller, clock generation and PLL
*
*    Has to be modified, if another CPU clock frequency should be
*    used. This function is called during startup and
*    has to return 1 to perform segment initialization.
*
*    Because variables are not initialized before __low_level_init() is called,
*    no access or call which relies on any variable can be used here.
*/
#ifdef __cplusplus
extern "C" {
#endif
OS_INTERWORK int __low_level_init(void);    // Avoid "no prototype" warning
#ifdef __cplusplus
  }
#endif
int __low_level_init(void) {
  //
  // Initialize PLL
  //
  _PLL_Init();
  //
  // Set Flash wait states
  _Flash_Init();
  //
  // Initialize Clock Tree
  //
  _MapClocks();
  return 1;                    // Always return 1 to enable segment initilization!
}

/*************************** End of file ****************************/
