/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
Purpose: Initializes and handles the hardware for embOS
*/

#include "RTOS.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/

/*********************************************************************
*
*       Configuration
*/

#ifndef   OS_IGNORE_UNDEFINED_INTERRUPT
  #define OS_IGNORE_UNDEFINED_INTERRUPT  (0)   // Do not ignore, undefined interrupts end in an endless loop for diagnostics
#endif

/*********************************************************************
*
*       Clock settings
*/
#define OS_EXTERNAL_FREQ  (7500000u)
#define OS_SYSTEM_FREQ    (OS_EXTERNAL_FREQ * 8)
#define OS_ICLK_DIVIDER   (4u)  // ICLK runs with 1/4 of MCLK

/*********************************************************************
*
*       System tick settings
*/
#define OS_TIMER_FREQ     (OS_SYSTEM_FREQ / 4)
#define OS_TICK_FREQ      (1000u)
#define OS_INT_FREQ       (OS_TICK_FREQ)
#define OS_TIMER_RELOAD   (OS_TIMER_FREQ / OS_INT_FREQ)
#define OS_TIMER_IRQ_ID   CIM_COMP1


/*********************************************************************
*
*       embOSView settings
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #include "BSP_UART.h"
  #define OS_UART      (0u)
  #define OS_BAUDRATE  (38400u)
#endif

/*********************************************************************
*
*       Local defines (sfrs used in RTOSInit.c)
*
**********************************************************************
*/

#define NUM_INT_SOURCES    (32u)
#define INT_CHANNEL_MASK   (0xFFu)
//
// Global control register
//
#define GLBCTRL            (*(volatile OS_U32*)0xFFFFFFDCu)
//
// Peripheral control register
//
#define PCR                (*(volatile OS_U32*)0xFFFFFD30u)
//
// System exception control registers
//
#define SYSECR             (*(volatile OS_U32*)0xffffffe0u)
//
// Flash control registers
//
#define FMREGOPT           (*(volatile OS_U32*)0xffe89c00u)
#define FMMAC1             (*(volatile OS_U16*)0xffe8bc02u)
#define FMMAC2             (*(volatile OS_U16*)0xffe8bc06u)
#define FMBAC1             (*(volatile OS_U16*)0xffe88002u)
#define FMBAC2             (*(volatile OS_U16*)0xffe88006u)

#define BANK0_ENA          ((unsigned int)0x00u << 0)    // 0:2  BANK Enable 0
#define BANK1_ENA          ((unsigned int)0x01u << 0)    // 0:2  BANK Enable 1
#define PMPPWR_ACTIVE      ((unsigned int)0x03u << 3)    // 4:3  Flash pump fallback power mode
#define BNKPWR_ACTIVE      ((unsigned int)0x03u << 0)    // 1:0  Bank power mode
#define PIPELINE_MODE      ((unsigned int)0x11u << 0)    // 0:7  Set Pipeline mode
#define ENPIPE             ((unsigned int)0x01u << 0)    // 0    Flash pipeline mode enable
#define BSTDBY_MAX         ((unsigned int)0x06u << 2)    // 7:2    Bank standby (100nS at 60MHz)
#define BSLEEP_MAX         ((unsigned int)0x72u << 8)    // 14:8   Bank sleep (1.9uS at 60MHz)
#define PSTDBY_MAX         ((unsigned int)0x3Cu << 5)    // 15:5   Pump standby (1.0uS at 60MHz)
#define PSLEEP_MAX         ((unsigned int)0x78u << 0)    // 14:1   Pump sleep (2.0uS at 60MHz)
//
// Other registers
//
#define RTICNTR            (*(volatile OS_U32*)0xFFFFFF00u)
#define RTIPCTL            (*(volatile OS_U32*)0xFFFFFF04u)
#define RTICNTL            (*(volatile OS_U32*)0xFFFFFF08u)
#define RTICMP1            (*(volatile OS_U32*)0xFFFFFF10u)
#define RTICINT            (*(volatile OS_U32*)0xFFFFFF18u)
#define RTICNTEN           (*(volatile OS_U32*)0xFFFFFF1Cu)
#define IRQIVEC            (*(volatile OS_U32*)0xFFFFFF20u)
#define FIRQPR             (*(volatile OS_U32*)0xFFFFFF2Cu)
#define REQMASK            (*(volatile OS_U32*)0xFFFFFF34u)
//
// Bit definitions
//
#define CMP1FLAG           ( 1u << 7)  //   7, Compare 1 flag
#define PENABLE            ( 1u << 0)  //   0, Peripheral enable
#define CNTEN_NOCNT        ( 3u << 0)  // 1:0, Counter enable No Count
#define CMP1ENA            ( 1u << 5)  //   5, Compare 1 enable
#define CNTEN_UP           ( 0u)       // 1:0, Counter enable User / Privilege
#define TX_ACTION_ENA      ( 1u << 3)  //   3, Tx DMA or interrupt enable
#define RX_ACTION_ENA      ( 1u << 4)  //   4, Rx DMA or interrupt enable
#define SW_NRESET          ( 1u << 7)  //   7, Software reset (active low)
#define RXENA              ( 1u)       //   0, Rx enable
#define TXENA              ( 1u)       //   0, Tx enable
#define CLOCK              ( 1u << 5)  //   5, Internal clock enable
#define RX_FUNC            ( 1u << 1)  //   1, SCIRX pin function
#define TX_FUNC            ( 1u << 1)  //   1, SCITX pin function
#define TIMING_MODE_ASYNC  ( 1u << 4)  // Timing Mode
#define CHAR_8             ( 7u)       // Character length control 8
#define CIM_COMP1          ( 2u)       // COMP1 interrupt
#define CIM_SCI1RX         (10u)       // SCI1 receive interrupt
#define CIM_SCI1TX         (20u)       // SCI1 transmit interrupt
#define CIM_SCI2RX         (17u)       // SCI2 receive interrupt
#define CIM_SCI2TX         (26u)       // SCI2 transmit interrupt

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/

/*********************************************************************
*
*       Interrupt vector table
*
*  Notes
*    The translation table has to be located in un-initialized memory,
*    because it may be filled from startup, before segments and variables
*    are initialized.
*/
#if   (defined(__ICCARM__))  // IAR
  #pragma data_alignment=4
  static __no_init OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES];
#elif (defined(__CC_ARM))
  #pragma arm section zidata = "NO_INIT"         //   Has to be declared as separate section UNINIT in the scatter file
  __align(4) static OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES];
  #pragma arm section zidata // back to default (.bss section)
#elif (defined(__GNUC__))    // GCC
  static OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES] __attribute__ ((aligned (4), section (".no_init")));
#endif

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USEPARA(Channel);  // Not supported by this version, avoid compiler warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USEPARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
/*********************************************************************
*
*       _OS_OnRX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static void _OS_OnRX(unsigned int Unit, unsigned char c) {
  OS_USEPARA(Unit);
  OS_COM_OnRx(c);
}

/*********************************************************************
*
*       _OS_OnTX()
*
*  Function description
*    Callback wrapper function for BSP UART module.
*/
static int _OS_OnTX(unsigned int Unit) {
  OS_USEPARA(Unit);
  return (int)OS_COM_OnTx();
}
#endif

/*********************************************************************
*
*       _InitCIM()
*
*  Function description
*    Initialize interrupt controller
*/
static void _InitCIM(void) {
  OS_INT_Disable();
  FIRQPR  = 0;        // Assign all sources to IRQ, clear FIQ priority
  REQMASK = 0;        // Disable all Interrupts
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       _InitFlash()
*
*  Function description
*    Initialize Flash module timing
*/
static void _InitFlash(void) {
  //
  // Setup Flash access timing, wait states, enable pipeline mode
  //
  GLBCTRL  |= (1u << 4);                               // Set Flash configuration mode
  FMMAC1    = PSLEEP_MAX;                              // pump sleep to standby time
  FMMAC2    = BANK0_ENA + PMPPWR_ACTIVE + PSTDBY_MAX;  // bank0
  FMBAC1    = BNKPWR_ACTIVE+ BSTDBY_MAX;               // active
  FMBAC2    = PIPELINE_MODE + BSLEEP_MAX;              // 0 wait states
  FMMAC2    = BANK1_ENA + PMPPWR_ACTIVE + PSTDBY_MAX;  // bank1
  FMBAC1    = BNKPWR_ACTIVE + BSTDBY_MAX;              // active
  FMBAC2    = PIPELINE_MODE + BSLEEP_MAX;              // 0 wait states
  FMREGOPT  = ENPIPE;                                  // ENABLE PIPELINE MODE
  GLBCTRL  &= ~(1u << 4);                              // Disable Flash configuration mode
  //
  // Enable peripherals, setup ICLK divider
  //
  PCR       = 0;                                       // Disable and reset peripherals
  PCR       = ((OS_ICLK_DIVIDER - 1u) & 0xFu) << 1;    // Setup Divider
  PCR      |= 1u;                                      // Enable peripherals
  //
  // Initialize PLL, Multiply input frequency by 8
  //
  GLBCTRL &= ~(0x07u);                                 // Clear divider prescaler => MCLK = PLLCLK
  GLBCTRL &= ~(1u << 3);                               // Switch PLL multiplier to * 8 mode
}

/*********************************************************************
*
*       _OS_ISR_Undefined()
*
*  Function description
*    Is called when an not installed interrupt was detected.
*    As interrupt pending condition of peripherals has to be reset,
*    program will not continue when interrupt is ignored.
*/
static void _OS_ISR_Undefined(int Index) {
#if (OS_IGNORE_UNDEFINED_INTERRUPT == 0)
  volatile unsigned int Dummy;

  Dummy = Index;
  while (Dummy < (INT_CHANNEL_MASK + 1)) {
    ;  // You may set a breakpoint here to detect non-installed ISR-handler
  }
#endif
  OS_USEPARA(Index);
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value.
*
*  Return value
*    Current timer count value.
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  OS_U32 Preload;
  OS_U32 UpCount;
  OS_U32 DownCount;

  Preload   = (RTIPCTL & ((1u << 11) - 1u));  // PRELOAD Value for the downcounting timer (10:0)
  UpCount   = (RTICNTR >> 11);                // Current upcounting Value (31:11)
  DownCount = (RTICNTR & ((1u << 11) - 1u));  // Current downcounting Value (10:0)
  return ((Preload + 1) * UpCount + Preload - DownCount);
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set.
*
*  Return value
*    == 0: Interrupt pending flag not set.
*    != 0: Interrupt pending flag set.
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return (RTICINT & CMP1FLAG);
}

/*********************************************************************
*
*       _OS_ISR_Tick()
*
* Function description
*   This is the code that gets called when the processor
*   receives a hardware timer exception.
*/
static void _OS_ISR_Tick(void) {
  RTICINT &= ~CMP1FLAG;        // Clear timer interrupt pending condition
  RTICMP1 += OS_TIMER_RELOAD;  // Advance comparator to next interrupt
  OS_TICK_Handle();            // Call OS tick handler
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_Process();
#endif
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware required for embOS to run.
*/
const OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};

void OS_InitHW(void) {
  OS_INT_IncDI();               // Initially disable interrupts
  SYSECR   = 0x00004007;        // Disable illegal address reset

  _InitFlash();
  _InitCIM();                   // Initialize interrupt controller
  PCR |= PENABLE;               // enable peripheral clock
  //
  // Setup OS timer: periodic interrupt using RTI with RTICMP1
  //
  RTICNTEN  = CNTEN_NOCNT;      // Stop counting
  RTICNTR   = 0;                // clear 21-bits CNTR
  RTIPCTL   = 0x3u;             // preload 11-bits MOD
  RTICMP1   = OS_TIMER_RELOAD;  // CMP1 used to generate interrupt.
  RTICNTL   = 0;                // clear and disable tap
  RTICINT   = 0;
  RTICINT  |= CMP1ENA;
  RTICNTEN  = CNTEN_UP;         // Start count, CNTR and MOD will count in both USER and SYSTEM mode
  //
  // Enable timer interrupt in interrupt controller
  //
  OS_ARM_InstallISRHandler(OS_TIMER_IRQ_ID, _OS_ISR_Tick);
  OS_ARM_EnableISR(OS_TIMER_IRQ_ID);
  //
  // Inform embOS about the timer settings
  //
  OS_TIME_ConfigSysTimer(&SysTimerConfig);
  //
  // Initialize communication for embOSView
  //
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Init(OS_UART, OS_BAUDRATE, BSP_UART_DATA_BITS_8, BSP_UART_PARITY_NONE, BSP_UART_STOP_BITS_1);
  BSP_UART_SetReadCallback(OS_UART, _OS_OnRX);
  BSP_UART_SetWriteCallback(OS_UART, _OS_OnTX);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This code is executed whenever no task, software timer, or
*    interrupt is ready for execution.
*
*  Additional information
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved.
*/
void OS_Idle(void) {    // Idle loop: No task is ready to execute
  while (1) {           // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                        // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Sends one character.
*/
void OS_COM_Send1(OS_U8 c) {
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_SendCharOnChannelNB(1, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  BSP_UART_Write1(OS_UART, c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_IF_ETHERNET)
  UDP_Process_Send1(c);
#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)
  OS_USEPARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
#endif
}

/*********************************************************************
*
*       OS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;
  int             ISRIndex;

  ISRIndex = (IRQIVEC & 0xFFu) - 1;  // Examine current interrupt source (channel number)
  if (ISRIndex >= 0) {
    REQMASK &= ~(1u << ISRIndex);    // Disable pending interrupt source to avoid recursion
    OS_INT_EnterNestable();
    pISR = _apIsrHandler[ISRIndex];  // Read interrupt vector from Table
    if (pISR != NULL) {
      pISR();                        // Call installed interrupt service routine
    } else {
      _OS_ISR_Undefined(ISRIndex);
    }
    REQMASK |= (1u << ISRIndex);     // Re-enable pending interrupt source
    OS_INT_LeaveNestable();
  }
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Used to install a specific interrupt vector when ARM
*    CPUs with vectored interrupt controller are used.
*
*  Parameters
*    ISRIndex:    Index of the interrupt source, normally the interrupt vector number.
*    pISRHandler: Address of the interrupt handler function.
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler(int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER*  pOldHandler;

  pOldHandler = NULL;
  OS_INT_Disable();
  if ((unsigned int)ISRIndex < NUM_INT_SOURCES) {
    pOldHandler             = _apIsrHandler[ISRIndex];
    _apIsrHandler[ISRIndex] = pISRHandler;
  }
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Used to enable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be enabled.
*/
void OS_ARM_EnableISR(int ISRIndex) {
  OS_INT_Disable();
  if ((unsigned int)ISRIndex < NUM_INT_SOURCES) {
    REQMASK |= (1u << ISRIndex);
  }
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Used to disable interrupt acceptance of a specific interrupt
*    source in a vectored interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be disabled.
*/
void OS_ARM_DisableISR(int ISRIndex) {
  OS_INT_Disable();
  if ((unsigned int)ISRIndex < NUM_INT_SOURCES) {
    REQMASK &= ~(1u << ISRIndex);
  }
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Function description
*    Used to set or modify the priority of a specific interrupt
*    source by programming the interrupt controller.
*
*  Parameters
*    ISRIndex: Index of the interrupt source which should be modified.
*    Prio:     The priority which should be set for the specific interrupt.
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  //
  // individual interrupt priorities are not supported by CIM
  //
  (void)ISRIndex;  // Suppress warning
  (void)Prio;      // Suppress warning
  return 0;
}

/*************************** End of file ****************************/
