/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

-------------------------- END-OF-HEADER -----------------------------
File      : BSP.c
Purpose   : BSP for the Enclustra Mercury XU5 SoC module
Literature:
  [1]  Enclustra User Schematics Mercury XU5 Revision 1.2
       (\\fileserver\Techinfo\Company\Enclustra\Evalboards\Mercury_XU5-R1-2_User_Schematics_V4.pdf)
  [2]  Zynq UltraScale+ Devices Register Reference, UG1087 (v1.7) February 8, 2019
       (\\fileserver\Techinfo\Company\Xilinx\UltraScale\ug1087-zynq-ultrascale-registers.zip and
        https://www.xilinx.com/html_docs/registers/ug1087/ug1087-zynq-ultrascale-registers.html)
*/

#include "BSP.h"

/*********************************************************************
*
*       Defines
*
**********************************************************************
*/
#define GPIO_BASE_ADDR         (0xFF0A0000u)  // [2] UG1087, GPIO Module description
#define GPIO_DATA_0            (*(volatile unsigned int*)(GPIO_BASE_ADDR + 0x040u))
#define GPIO_DIRMODE_0         (*(volatile unsigned int*)(GPIO_BASE_ADDR + 0x204u))
#define GPIO_OUTPUT_ENABLE_0   (*(volatile unsigned int*)(GPIO_BASE_ADDR + 0x208u))

#define IOU_SLCR_BASE_ADDR     (0xFF180000u)  // [2] UG1087, IOP system-level control Module Description
#define IOU_SLCR_MIO_PIN_24    (*(volatile unsigned int*)(IOU_SLCR_BASE_ADDR + 0x060u))
#define IOU_SLCR_MIO_PIN_25    (*(volatile unsigned int*)(IOU_SLCR_BASE_ADDR + 0x064u))
#define IOU_SLCR_BANK0_CTRL0   (*(volatile unsigned int*)(IOU_SLCR_BASE_ADDR + 0x138u))
#define IOU_SLCR_BANK0_CTRL1   (*(volatile unsigned int*)(IOU_SLCR_BASE_ADDR + 0x13Cu))
#define IOU_SLCR_MIO_MST_TRI0  (*(volatile unsigned int*)(IOU_SLCR_BASE_ADDR + 0x204u))

#define LED0_BIT               (24)  // PS_MIO24, [1]
#define LED1_BIT               (25)  // PS_MIO25, [1]

#define LED0_MASK              (1u << LED0_BIT)
#define LED1_MASK              (1u << LED1_BIT)

#define LEDMASK_ALL            (LED0_MASK | LED1_MASK)
#define LED_NUM_LEDS           (2)

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       BSP_Init()
*/
void BSP_Init(void) {
  //
  // Configure MIO pins for GPIO (i.e.: apply reset values)
  //
  IOU_SLCR_MIO_PIN_24    = (0u << 1)      // Level 0 Mux Select, 0 = Select Level 1 output
                         | (0u << 2)      // Level 1 Mux Select, 0 = Select Level 2 output
                         | (0u << 3)      // Level 2 Mux Select, 0 = Select Level 3 output
                         | (0u << 5);     // Level 3 Mux Select, 0 = Select GPIO 24
  IOU_SLCR_MIO_PIN_25    = (0u << 1)      // Level 0 Mux Select, 0 = Select Level 1 output
                         | (0u << 2)      // Level 1 Mux Select, 0 = Select Level 2 output
                         | (0u << 3)      // Level 2 Mux Select, 0 = Select Level 3 output
                         | (0u << 5);     // Level 3 Mux Select, 0 = Select GPIO 25
  IOU_SLCR_MIO_MST_TRI0 &= ~LEDMASK_ALL;  // Disable Tri-State for relevant pins
  //
  // Configure drive strength in tuple [CTRL0],[CTRL1] for each pin (i.e.: apply reset values)
  //
  IOU_SLCR_BANK0_CTRL0   = 0x3FFFFFFu;    // Set all bits =>  8 mA on all pins
  IOU_SLCR_BANK0_CTRL1   = 0x3FFFFFFu;    // Set all bits => 12 mA on all pins
  //
  // Configure pins
  //
  GPIO_DIRMODE_0         = LEDMASK_ALL;   // Set relevant pins to output
  GPIO_OUTPUT_ENABLE_0   = LEDMASK_ALL;   // Enable output for relevant pins
  GPIO_DATA_0            = 0u;            // Initially, clear LEDs
}

/*********************************************************************
*
*       BSP_SetLED()
*/
void BSP_SetLED(int Index) {
  if ((Index >= 0) && (Index < LED_NUM_LEDS)) {
    GPIO_DATA_0 |= (1u << (LED0_BIT + Index));  // Switch on LED
  }
}

/*********************************************************************
*
*       BSP_ClrLED()
*/
void BSP_ClrLED(int Index) {
  if ((Index >= 0) && (Index < LED_NUM_LEDS)) {
    GPIO_DATA_0 &= ~(1u << (LED0_BIT + Index)); // Switch off LED
  }
}

/*********************************************************************
*
*       BSP_ToggleLED()
*/
void BSP_ToggleLED(int Index) {
  if ((Index >= 0) && (Index < LED_NUM_LEDS)) {
    GPIO_DATA_0 ^= (1uL << (LED0_BIT + Index));
  }
}

/*************************** End of file ****************************/
