/*********************************************************************
*                     SEGGER Microcontroller GmbH                    *
*                        The Embedded Experts                        *
**********************************************************************
*                                                                    *
*       (c) 1995 - 2025 SEGGER Microcontroller GmbH                  *
*                                                                    *
*       Internet: segger.com  Support: support_embos@segger.com      *
*                                                                    *
**********************************************************************
*                                                                    *
*       embOS * Real time operating system                           *
*                                                                    *
*       Please note:                                                 *
*                                                                    *
*       Knowledge of this file may under no circumstances            *
*       be used to write a similar product or a real-time            *
*       operating system for in-house use.                           *
*                                                                    *
*       Thank you for your fairness !                                *
*                                                                    *
**********************************************************************
*                                                                    *
*       OS version: V5.20.0.0                                        *
*                                                                    *
**********************************************************************

----------------------------------------------------------------------
File    : RTOSInit_XCZU2EG.c for Xilinx XCZU2EG CPU (Cortex-R5_0)
Purpose : Initializes and handles the hardware for embOS as far
          as required by embOS
          Feel free to modify this file acc. to your target system.
Literature:
  [1]  Generic Interrupt Controller (GIC-390) Errata Notice
       (\\fileserver\Techinfo\Company\ARM\GIC_GenericInterruptController\GIC_390_Errata_Notice_v6.pdf)
  [2]  Zynq UltraScale+ Device TRM, UG1085 (v1.9) January 17, 2019
       (\\fileserver\Techinfo\Company\Xilinx\UltraScale\ug1085-zynq-ultrascale-trm.pdf)
--------  END-OF-HEADER  ---------------------------------------------
*/

#include "RTOS.h"
#include "xscugic.h"
#include "xttcps.h"

/*********************************************************************
*
*       Configuration
*
**********************************************************************
*/
#define ALLOW_NESTED_INTERRUPTS  0   // Caution: Nesting interrupts will cause higher stack load on system stack CSTACK

/*********************************************************************
*
*       Clock frequency settings (configuration)
*/
#define OS_TIMER_FREQ  (_xTimerInstance.Config.InputClockHz)
#define OS_TICK_FREQ   (1000u)
#define OS_INT_FREQ    (OS_TICK_FREQ)

/*********************************************************************
*
*       Configuration of communication to embOSView
*/
#ifndef   OS_VIEW_IFSELECT
  #define OS_VIEW_IFSELECT  OS_VIEW_IF_JLINK
#endif

/*********************************************************************
*
*       UART settings for embOSView
*       If you do not want (or can not due to hardware limitations)
*       to dedicate a UART to embOSView, please define it to be -1
*       Currently no uart is supported.
*       When using DCC for communication, the UART is not used for embOSView,
*       regardless the OS_UART settings
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_UART)
  #ifndef   OS_UART
    #define OS_UART (-1)
  #endif

  #ifndef   OS_BAUDRATE
    #define OS_BAUDRATE (38400)
  #endif
#endif

/****** End of configurable options *********************************/

#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  #include "JLINKDCC.h"
#endif

/*********************************************************************
*
*       Local defines (sfrs and addresses used in RTOSInit.c)
*
**********************************************************************
*/
//
// Triple-Timer counter 0
//
#define TTC0_INT_ID             (68)                        // [2] UG1085, p. 305: [68:70]

//
// Interrupt controller GIC PL390
//
#define GIC_BASE_ADDR           (0xF9000000uL)              // [2] UG1085, p. 226
#define GICD_BASE_ADDR          (GIC_BASE_ADDR + 0x0000uL)  // GIC distributor
#define GICC_BASE_ADDR          (GIC_BASE_ADDR + 0x1000uL)  // GICC interface
#define GICC_HPIR               (*(volatile unsigned int*)(GICC_BASE_ADDR + 0x018uL))
#define GICD_IPRIORITY          (*(volatile unsigned int*)(GICD_BASE_ADDR + 0x400uL))

#define NUM_INT_SOURCES         (148u)                      // [2] UG1085, p. 302

/*********************************************************************
*
*       Static data
*
**********************************************************************
*/
static XTtcPs          _xTimerInstance;  // Triple timer counter instance
static OS_ISR_HANDLER* _apIsrHandler[NUM_INT_SOURCES] __attribute__ ((aligned (4), section (".no_init")));

/*********************************************************************
*
*       Local functions
*
**********************************************************************
*/

/*********************************************************************
*
*       _OS_ISR_Undefined()
*
*  Function description
*    Is called when an uninstalled interrupt was detected
*    As interrupt pending condition of peripherals has to be reset,
*    program will not continue when interrupt is ignored.
*/
static void _OS_ISR_Undefined(void) {
  volatile int Dummy;

  Dummy = 1;
  /* You may set a breakpoint here to detect uninstalled ISR-handler */
  while (Dummy > 0) {
  }
}

/*********************************************************************
*
*       _SpuriousIrqHandler()
*
*  Function description
*    A dummy spurious IRQ handler
*/
static OS_U32 _SpuriousIrqCnt = 0;
static void _SpuriousIrqHandler(void) {
  _SpuriousIrqCnt++;
}

/*********************************************************************
*
*       OS_ISR_Tick()
*
*  Function description
*    The OS system timer interrupt
*/
static void OS_ISR_Tick (void) {
  OS_U32 CurrentState;

  CurrentState = XTtcPs_GetInterruptStatus(&_xTimerInstance);
  XTtcPs_ClearInterruptStatus(&_xTimerInstance, CurrentState);
  OS_TICK_Handle();
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)
  JLINKDCC_Process();
#endif
}

/*********************************************************************
*
*       _GIC_Init()
*
*  Function description
*    Initialize interrupt controller by setting default vectors
*    and clearing all interrupts
*/
static void _GIC_Init(void) {
  OS_U32          i;
  XScuGic_Config* pxGICConfig;
  XScuGic         xInterruptController;  // Generic interrupt controller instance

  pxGICConfig = XScuGic_LookupConfig(XPAR_SCUGIC_SINGLE_DEVICE_ID);                        // Retrieve current GIC configuration
  XScuGic_CfgInitialize(&xInterruptController, pxGICConfig, pxGICConfig->CpuBaseAddress);  // Set appropriate address for GIC CPU interface
  XScuGic_SetPriorityTriggerType(&xInterruptController, XPAR_XTTCPS_0_INTR, 0xF8, 3);      // Configure TTC0 interrupt for rising edge
  XScuGic_Enable(&xInterruptController, XPAR_XTTCPS_0_INTR);                               // Enable TTC0 interrupt in GIC
  //
  // Set GIC base address
  //
  OS_GIC_SetCPUInterfaceBaseAddr(GICC_BASE_ADDR);
  OS_GIC_SetDistributionBaseAddr(GICD_BASE_ADDR);
  //
  // Configure all interrupts
  //
  for (i = 0; i < NUM_INT_SOURCES; i++) {
    _apIsrHandler[i] = _OS_ISR_Undefined;  // Set default interrupt handler
    OS_GIC_SetIntSecure(i);                // Set to secure
    OS_GIC_DisableInt(i);                  // Disable interrupt
    //
    // Initialize all Interrupt pointer target register to address the first R5 core
    // The first 8 registers are fixed
    //
    if (i > 8u) {
      OS_GIC_SetTargetProc(i, 0x01);
    }
  }
  //
  // Configure priority mask to lowest value (highest number) - enables all priorities
  //
  OS_GIC_SetPrioThreshold(0xFF);
  //
  // This register defines the point at which the priority value fields split into two parts,
  // the group priority field and the subpriority field.
  //
  OS_GIC_SetPrioGroup(0x00);
  //
  // Enable the GIC
  //
  OS_GIC_Enable();
}

/*********************************************************************
*
*       _OS_GetHWTimerCycles()
*
*  Function description
*    Returns the current hardware timer count value
*
*  Return value
*    Current timer count value
*/
static unsigned int _OS_GetHWTimerCycles(void) {
  return XTtcPs_GetCounterValue(&_xTimerInstance);
}

/*********************************************************************
*
*       _OS_GetHWTimer_IntPending()
*
*  Function description
*    Returns if the hardware timer interrupt pending flag is set
*
*  Return value
*    == 0; Interrupt pending flag not set
*    != 0: Interrupt pending flag set
*/
static unsigned int _OS_GetHWTimer_IntPending(void) {
  return (XTtcPs_GetInterruptStatus(&_xTimerInstance) & XTTCPS_IXR_INTERVAL_MASK);
}

/*********************************************************************
*
*       Global functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_InitHW()
*
*  Function description
*    Initialize the hardware (timer) required for embOS to run.
*    May be modified, if an other timer should be used.
*/
void OS_InitHW(void) {
  OS_U32         TickPrio;
  XTtcPs_Config* pxTimerConfig;
  XInterval      usInterval;
  OS_U8          ucPrescale;

  OS_INT_IncDI();
  //
  // Init Caches
  //
  OS_ARM_ICACHE_Enable();                                                                         // Enable instruction cache
  OS_ARM_DCACHE_Enable();                                                                         // Enable data cache
  OS_ARM_CACHE_Sync();                                                                            // Ensure caches are synchronized
  //
  // Init Generic interrupt controller
  //
  _GIC_Init();
  //
  // Configure Triple timer counter 0 (to be used as embOS system tick)
  //
  pxTimerConfig = XTtcPs_LookupConfig(XPAR_PSU_TTC_0_DEVICE_ID);                                  // Retrieve current configuration
  if (XTtcPs_CfgInitialize(&_xTimerInstance, pxTimerConfig, pxTimerConfig->BaseAddress) == XST_DEVICE_IS_STARTED) {
    XTtcPs_Stop(&_xTimerInstance);                                                                // After a software reset, TTC0 still is configured. Re-configuring it then will perpetually halt program execution...
    XTtcPs_CfgInitialize(&_xTimerInstance, pxTimerConfig, pxTimerConfig->BaseAddress);            // ...as a workaround, we therefore stop the timer and re-initialize it afterwards. Would not be necessary if peripherals were to be properly re-set, which unfortunately is not the case.
  }
  XTtcPs_CalcIntervalFromFreq(&_xTimerInstance, OS_TICK_FREQ, &usInterval, &ucPrescale);          // Calculate appropriate timer interval for the current timer frequency
  XTtcPs_SetInterval(&_xTimerInstance, usInterval);                                               // Configure calculated timer interval
  XTtcPs_SetPrescaler(&_xTimerInstance, ucPrescale);                                              // Configure calculated prescaler
  XTtcPs_SetOptions(&_xTimerInstance, XTTCPS_OPTION_INTERVAL_MODE | XTTCPS_OPTION_WAVE_DISABLE);  // Set timer to interval mode, disable wave-generation
  XTtcPs_Start(&_xTimerInstance);                                                                 // Start the timer
  XTtcPs_EnableInterrupts(&_xTimerInstance, XTTCPS_IXR_INTERVAL_MASK);                            // Enable interrupts in the timer
  //
  // Install OS timer ISR Handler and set interrupt priority to lowest priority
  //
  OS_ARM_InstallISRHandler(TTC0_INT_ID, OS_ISR_Tick);                                             // Install embOS tick handler
  OS_ARM_ISRSetPrio(TTC0_INT_ID, 0xFFu);                                                          // Set to lowest level, ALL BITS set
  TickPrio  = OS_ARM_ISRSetPrio(TTC0_INT_ID, 0xFFu);                                              // Read back priority to examine relevant preemption-level-bits
  TickPrio -= 1;                                                                                  // Set to low preemption level, 1 level higher than lowest
  OS_ARM_ISRSetPrio(TTC0_INT_ID, TickPrio);                                                       // Set appropriate ISR priority
  OS_ARM_EnableISR(TTC0_INT_ID);                                                                  // Enable TTC0 interrupt (though it should already be enabled at this point)
  //
  // Setup values for usec precise system time functions
  //
  {
    OS_SYSTIMER_CONFIG SysTimerConfig = {OS_TIMER_FREQ, OS_INT_FREQ, OS_TIMER_UPCOUNTING, _OS_GetHWTimerCycles, _OS_GetHWTimer_IntPending};
    OS_TIME_ConfigSysTimer(&SysTimerConfig);
  }
  //
  // Initialize the optional communication for embOSView
  //
#if (OS_VIEW_IFSELECT != OS_VIEW_DISABLED)
  OS_COM_Init();
#endif
  OS_INT_DecRI();
}

/*********************************************************************
*
*       OS_Idle()
*
*  Function description
*    This is basically the "core" of the idle loop.
*    This core loop can be changed, but:
*    The idle loop does not have a stack of its own, therefore no
*    functionality should be implemented that relies on the stack
*    to be preserved. However, a simple program loop can be programmed
*    (like toggling an output or incrementing a counter)
*/
void OS_Idle(void) {  // Idle loop: No task is ready to execute
  while (1) {         // Nothing to do ... wait for interrupt
    #if ((OS_VIEW_IFSELECT != OS_VIEW_IF_JLINK) && (OS_DEBUG == 0))
                      // Switch CPU into sleep mode
    #endif
  }
}

/*********************************************************************
*
*       Optional communication with embOSView
*
**********************************************************************
*/
#if (OS_VIEW_IFSELECT == OS_VIEW_IF_JLINK)

/*********************************************************************
*
*       _DCC_OnRx()
*/
static void _DCC_OnRx(unsigned Channel, OS_U8 Data) {
  OS_USE_PARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnRx(Data);
}

/*********************************************************************
*
*       _DCC_OnTx()
*/
static void _DCC_OnTx(unsigned Channel) {
  OS_USE_PARA(Channel);  // Not supported by this version, suppress warning
  OS_COM_OnTx();
}

/*********************************************************************
*
*       OS_COM_Init()
*
*  Function description
*    Initialize DCC for embOSView
*/
void OS_COM_Init(void) {
  JLINKDCC_SetpfOnRx(1, _DCC_OnRx);
  JLINKDCC_SetpfOnTx(1, _DCC_OnTx);
}

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Send 1 character via DCC
*/
void OS_COM_Send1(OS_U8 c) {
  JLINKDCC_SendCharOnChannelNB(1, c);
}

#elif (OS_VIEW_IFSELECT == OS_VIEW_DISABLED)

/*********************************************************************
*
*       OS_COM_Send1()
*
*  Function description
*    Dummy routine.
*/
void OS_COM_Send1(OS_U8 c) {
  OS_USE_PARA(c);           // Avoid compiler warning
  OS_COM_ClearTxActive();  // Let embOS know that Tx is not busy
}

#else
  #error "Selected embOSView interface is currently not supported."
#endif

/*********************************************************************
*
*       OS interrupt handler and ISR specific functions
*
**********************************************************************
*/

/*********************************************************************
*
*       OS_irq_handler()
*
*  Function description
*    Detect reason for IRQ and call corresponding service routine.
*    OS_irq_handler is called from the IRQ_handler function
*    found in RTOS.s.
*    IRQ_Handler has to be inserted in the vector table in the
*    startup code.
*/
OS_INTERWORK void OS_irq_handler(void) {
  OS_ISR_HANDLER* pISR;
  volatile OS_U32 Dummy;
  OS_U32          int_id;
  OS_U32          icdipr0_value;

  Dummy  = GICC_HPIR;        // Erratum 801120 [1] - ensure subsequent ICCIAR data will be correct
  OS_USE_PARA(Dummy);        // Avoid warning about "set but not used" variable
  int_id = OS_GIC_AckInt();  // Acknowledge the interrupt and read the interrupt ID
  //
  // Read data of 0x0 (SGI ID0 from CPU0) is possibly affected by 733075
  // Interrupt IDs 0x3FE and 0x3FF are also possibly affected
  //
  if ((int_id == 0u) || (int_id  >= 0x3FEu)) {
    //
    // Erratum 733075 - prevent possible CPU interface deadlock (ARM Errata Notice, PR258-GENC-008719 v6.0).
    // Assumes SGI ID0 from CPU0 is not used as suggested by ARM.
    //
    icdipr0_value  = GICD_IPRIORITY; // One way to get the value for the workaround write is to read it first
    GICD_IPRIORITY = icdipr0_value;
#if ((defined __ICCARM__) && (__VER__ < 6300000))
    __asm("dsb");                   // Ensure write completes before any future ICCHPIR or ICCIAR reads
#else
    __asm volatile("dsb");
#endif
  }
  if ((int_id == 0x3FEu) || (int_id == 0x3FFu)) {  // Check for spurious interrupts
     _SpuriousIrqHandler();
  } else {
    if (int_id < NUM_INT_SOURCES) {
      pISR = _apIsrHandler[int_id]; // Get interrupt vector handler
#if (ALLOW_NESTED_INTERRUPTS == 1)
      OS_INT_EnterNestable();       // Now interrupts may be reenabled. If nesting should be allowed
#else
      OS_INT_Enter();               // Inform OS that interrupt handler is running
#endif
      //
      // Call interrupt service routine
      //
      pISR();
#if (ALLOW_NESTED_INTERRUPTS == 1)
      OS_INT_LeaveNestable();       // Leave nestable interrupt, perform task switch if required
#else
      OS_INT_Leave();               // Leave interrupt, perform task switch if required
#endif
    }
    OS_GIC_EOI(int_id);             // End interrupt
  }
}

/*********************************************************************
*
*       OS_ARM_InstallISRHandler()
*
*  Function description
*    Installs an interrupt handler
*
*  Parameters
*    ISRIndex:    ISR index
*    pISRHandler: Interrupt handler function address
*/
OS_ISR_HANDLER* OS_ARM_InstallISRHandler (int ISRIndex, OS_ISR_HANDLER* pISRHandler) {
  OS_ISR_HANDLER* pOldHandler;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
    return NULL;
  }
#endif
  OS_INT_Disable();
  pOldHandler             = _apIsrHandler[ISRIndex];
  _apIsrHandler[ISRIndex] = pISRHandler;
  OS_INT_EnableConditional();
  return pOldHandler;
}

/*********************************************************************
*
*       OS_ARM_EnableISR()
*
*  Function description
*    Enables an interrupt
*
*  Parameters
*    ISRIndex: ISR index
*/
void OS_ARM_EnableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_GIC_EnableInt(ISRIndex);
}

/*********************************************************************
*
*       OS_ARM_DisableISR()
*
*  Function description
*    Disables an interrupt
*
*  Parameters
*    ISRIndex: ISR index
*/
void OS_ARM_DisableISR(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_GIC_DisableInt(ISRIndex);
}

/*********************************************************************
*
*       OS_ARM_ISRSetPrio()
*
*  Function description
*    Sets an interrupt priority
*
*  Parameters
*    ISRIndex: ISR index
*    Prio    : ISR priority
*/
int OS_ARM_ISRSetPrio(int ISRIndex, int Prio) {
  OS_U32 OldPrio;

#if (OS_DEBUG != 0)
  if ((unsigned)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  OldPrio = OS_GIC_SetPrio(ISRIndex, Prio);
  OS_INT_EnableConditional();
  return (int)OldPrio;
}

/*********************************************************************
*
*       OS_ARM_ClearPendingFlag()
*
*  Function description
*    Clears an interrupt pending flag
*
*  Parameters
*    ISRIndex: ISR index
*/
void OS_ARM_ClearPendingFlag(int ISRIndex) {
#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  OS_GIC_ClrPending(ISRIndex);
  OS_INT_EnableConditional();
}

/*********************************************************************
*
*       OS_ARM_IsPending()
*
*  Function description
*    Checks if an interrupt is pending
*
*  Parameters
*    ISRIndex: ISR index
*
*  Return value
*    0: Interrupt is not pending
*    1: Interrupt is pending
*/
unsigned int OS_ARM_IsPending(int ISRIndex) {
  OS_U32 Result;

#if (OS_DEBUG != 0)
  if ((unsigned int)ISRIndex >= NUM_INT_SOURCES) {
    OS_Error(OS_ERR_ISR_INDEX);
  }
#endif
  OS_INT_Disable();
  Result = OS_GIC_IsPending(ISRIndex);
  OS_INT_EnableConditional();
  return Result;
}

/****** End Of File *************************************************/
